package org.helioviewer.jhv;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.channels.FileChannel;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import org.helioviewer.jhv.gui.ImageViewerGui;

/**
 * A class which provides functions for accessing and working with files.
 * 
 * @author Benjamin Wamsler
 */
public class FileUtils {
	
    /**
     * Method copies a file from src to dst.
     * 
     * @param _src
     *            Source as String
     * @param _dst
     *            Destination as String
     * @param _replace
     *            If true, existing files will be replaced
     * @throws IOException
     */
	public static void copy(String _src, String _dst, boolean _replace) throws IOException {
		JHVGlobals.err("", _src, false);
		JHVGlobals.err("", _dst, false);
		if(!new File(_dst).exists() || _replace) {
			// Create channel on the source
		    FileChannel srcChannel = new FileInputStream(_src).getChannel();
		    
		    // Create channel on the destination
		    FileChannel dstChannel = new FileOutputStream(_dst).getChannel();
		    
		    // Copy file contents from source to destination
		    dstChannel.transferFrom(srcChannel, 0, srcChannel.size());
		    
		    // Close the channels
		    srcChannel.close();
		    dstChannel.close();
		}
	}
	
    /**
     * Method copies a file from src to dst.
     * 
     * @param src
     *            Source file
     * @param dst
     *            Destination file
     * @throws IOException
     */
    public static void copy(File src, File dst) throws IOException {
        InputStream in = new FileInputStream(src);
        OutputStream out = new FileOutputStream(dst);
    
        // Transfer bytes from in to out
        byte[] buf = new byte[1024];
        int len;
        while ((len = in.read(buf)) > 0) {
            out.write(buf, 0, len);
        }
        in.close();
        out.close();
    }
    
    /**
     * Method extracts contents of a jar file.
     * 
     * @param _src
     *            Source file as String
     * @param _dst
     *            Destination folder as String
     */
    public static void extractResourceFromJAR(String _src, String _dst) {
        try {
         String home = ImageViewerGui.getSingletonInstance().getClass().getProtectionDomain().
                       getCodeSource().getLocation().getPath().replaceAll("%20", " ");
         JarFile jar = new JarFile(home);
         ZipEntry entry = jar.getEntry("org/helioviewer/gui/resources/"+_src);
         JHVGlobals.err("", entry.getName(), false);
         File efile = new File(_dst, entry.getName());

         InputStream in = 
            new BufferedInputStream(jar.getInputStream(entry));
         OutputStream out = 
            new BufferedOutputStream(new FileOutputStream(efile));
         
         byte[] buffer = new byte[2048];
         for (;;)  {
           int nBytes = in.read(buffer);
           if (nBytes <= 0) break;
           out.write(buffer, 0, nBytes);
         }
         out.flush();
         out.close();
         in.close();
        }
        catch (Exception e) {
         e.printStackTrace();
       }
     }
    
    /**
     * A method that parses a standard Helioviewer filename
     * 
     * @param filename
     *            the name of the file as String
     * @return result Returns the parsed data as an array of strings: result[0] =
     *         date, result[1] = time, result[2] = observatory, result[3] =
     *         instrument, result[4] = detector, result[5] = measurement,
     * */
    public static String[] parseJHVFilename(String filename)
    {
    	String date;
    	String time;
    	String fields[];
		
		filename = filename.substring(0, filename.lastIndexOf("."));
    	fields = filename.split("_");
				
		for (String s : fields)
		{
			System.out.print(s + " ");
		}
		System.out.println("\n");
		
		date = fields[0]+"/"+fields[1]+"/"+fields[2] + " GMT";
		time = fields[3].substring(0,2)+":" + 
		  fields[3].substring(2,4) + ":" + 
		  fields[3].substring(4,6);
		
		
		String[] result = new String[6];
		
		result[0]=date;
		result[1]=time;
		
		if (fields[4].equals("SOH"))  
			result[2]="SOHO";
		else
			result[2]=fields[4];
		
		if (fields[5].equalsIgnoreCase("LAS"))
			result[3] = "LASCO";
		else
			result[3]=fields[5];
		
		result[4]=fields[6];
		result[5]=fields[7];
		
		return result;
    }
}