package org.helioviewer.jhv;

import java.awt.EventQueue;
import java.io.File;
import java.lang.reflect.Method;
import java.util.Arrays;

import javax.swing.JOptionPane;

/**
 * Intended to be a class for static functions and fields relevant to the
 * application as a whole.
 * 
 * @author caplins
 */
public class JHVGlobals {
	
	/** The maximum amount of memory the JVM will use for the heap. */
	public static final long MAX_JVM_HEAP_SIZE = Runtime.getRuntime().maxMemory(); 
	
	/** The the maximum amount of memory the BufferManager object will use. */
	public static final long MAX_BUFFER_MANAGER_SIZE = (MAX_JVM_HEAP_SIZE*8)/10;

    private final static String[] browsers = { "firefox", "opera", "konqueror", "epiphany",
        "seamonkey", "galeon", "kazehakase", "mozilla", "netscape" };
	
	/** Constructor is private to prevent instantiation. */
	private JHVGlobals(){}
		
    /**
     * Attempts to create the necessary directories if they do not exist. It
     * gets its list of directories to create from the JHVDirectory class.
     * 
     * @throws SecurityException
     */
	public static void createDirs() throws SecurityException{
		JHVDirectory[] dirs = JHVDirectory.values();
		for(JHVDirectory dir : dirs){
			File f = dir.getFile();
			if(!f.exists())
				f.mkdirs();
		}
	}
		
    /**
     * A central error handler. Displays an error message in a JOptionDialog and
     * exits if the flag is set.
     * 
     * @param _title
     *            title of the error message.
     * @param _msg
     *            the message which has to be displayed.
     * @param _exitImmediately
     *            the program exits when the value true will be passed.
     */
	public static void err(final String _title, final Object _msg, final boolean _exitImmediately){
		EventQueue.invokeLater(new Runnable(){
			public void run(){
				JOptionPane.showMessageDialog(
						null,
						((_title == null ? "" : _title+"\n")+(_msg == null ? "No error details available." : _msg.toString())),
						(_exitImmediately ? "Fatal Error!" : "Error!"),
						JOptionPane.ERROR_MESSAGE);
				if(_exitImmediately)
					System.exit(-1);
			}
		});
	}
	
    /**
     * A central warning handler. Displays a warning message in a JOptionDialog.
     * 
     * @param _title
     *            title of the warning message.
     * @param _msg
     *            the message which has to be displayed.
     */
	public static void warn(final String _title, final Object _msg){
		final String msg = _msg.toString();
		
		EventQueue.invokeLater(new Runnable(){
			public void run(){
				JOptionPane.showMessageDialog(
						null,
						((_title == null ? "" : _title+"\n")+(msg == null || msg.equals("")? "No warning details available." : msg)),
						"Warning!",
						JOptionPane.WARNING_MESSAGE);
			}
		});
	}
	
    /**
     * Opens the specified web page in the default web browser
     * 
     * @param url
     *            A web address (URL) of a web page (e.g
     *            "http://www.jhelioviewer.org/")
     */
	public static void openURL(String url) {
		String osName = System.getProperty("os.name");
		try {
			if (osName.startsWith("Mac OS")) {
				
				Class<?> fileMgr = Class.forName("com.apple.eio.FileManager");
				Method openURL = fileMgr.getDeclaredMethod("openURL", new Class[] {String.class});
				openURL.invoke(null, new Object[] {url});
				
			} else if (osName.startsWith("Windows")) {
				Runtime.getRuntime().exec("rundll32 url.dll,FileProtocolHandler " + url);
			} else { //assume Unix or Linux
				boolean found = false;
				for (String browser : browsers) {
					if (!found) {
						found = Runtime.getRuntime().exec(new String[] {"which", browser}).waitFor() == 0;
						if (found) {
							Runtime.getRuntime().exec(new String[] {browser, url});
						}
					}
				}
				if (!found) {
					throw new Exception(Arrays.toString(browsers));
				}
			}
		} catch (Exception e) {
			JOptionPane.showMessageDialog(null, "Error attempting to launch web browser\n" + e.toString());
		}
	}	
}


	

