package org.helioviewer.jhv;

import java.io.IOException;

import javax.swing.UIManager;

import org.helioviewer.jhv.ffmpeg.FFmpegLoader;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.internal_plugins.InternalFilterPlugin;
import org.helioviewer.jhv.internal_plugins.InternalOverlayPlugin;
import org.helioviewer.jhv.io.CommandLineProcessor;
import org.helioviewer.jhv.opengl.GLLoadLibraries;
import org.helioviewer.viewmodel.view.jp2view.JP2Image;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_KduException;
import org.helioviewer.viewmodelplugin.controller.PluginManager;

/**
 * This class starts the applications.
 * 
 * @author caplins
 * @author Benjamin Wamsler
 * @author Markus Langenberg
 * @author Stephan Pagel
 * 
 */
public class JavaHelioViewer {

    /**
     * This method starts the application.
     * 
     * @param args arguments passed to the application.
     */
    public static void main(String[] args) {

        KakaduEngine engine = new KakaduEngine();
        GLLoadLibraries glEngine = new GLLoadLibraries();
        FFmpegLoader ffmpegEngine = new FFmpegLoader();
        
        // display the splash screen
        JHVSplashScreen splash = JHVSplashScreen.getSingletonInstance();
        splash.setProgressSteps(engine.getNumberOfLibs() + glEngine.getNumberOfLibs() + ffmpegEngine.getNumberOfLibs() + 4);
        splash.setProgressText("Initializing JHelioviewer...");
        
        // This attempts to create the necessary directories for the application
        JHVGlobals.createDirs();

        // The following code-block attempts to load/download&load the necessary dynamic libraries
        
        boolean sucess = true;
       
        try {
            engine.loadLibs();            
            splash.setCurrentStep(splash.getCurrentStep() + engine.getNumberOfLibs());
        } catch (UnsatisfiedLinkError ex) {
            sucess = false;
        };
        
        if(!sucess){
            try {
                engine.deleteLibs();
                engine.downloadLibs(splash);
                engine.loadLibs();
            } catch (Throwable t) {
                t.printStackTrace();
                JHVGlobals.err("Error loading Kakadu libraries", t.getMessage(), true);
            }
        }

        // The following code-block attempts to start the native message handling
        try{
            engine.startKduMessageSystem();
        } catch (JHV_KduException ex){
            ex.printStackTrace();
            JHVGlobals.err("Error starting Kakadu message handler", ex.getMessage(), true);
        }
        
        // Load/download OpenGL
        boolean glLoaded = true;
        try {
        	glEngine.loadLibs();
        	splash.setCurrentStep(splash.getCurrentStep() + glEngine.getNumberOfLibs() - 1);
        } catch(UnsatisfiedLinkError ex) {
        	glLoaded = false;
        }
        
        if(!glLoaded) {
        	try {
        		glEngine.deleteLibs();
        		glEngine.downloadLibs(splash);
        		glEngine.loadLibs();
            } catch (Throwable t) {
                t.printStackTrace();
                JHVGlobals.err("Error loading OpenGL libraries", t.getMessage(), true);
            }
        }
        
        try {
			glEngine.checkAndDownloadCgc(splash);
		} catch (IOException e) {
			e.printStackTrace();
			JHVGlobals.err("Error loading Cgc", e.getMessage(), true);
		}
		
		// Load/download ffmpeg
        try {
        	ffmpegEngine.checkAndDownloadFFMPEG(splash);
        	splash.setCurrentStep(splash.getCurrentStep() + ffmpegEngine.getNumberOfLibs() - 1);
        } catch(IOException ex) {
        	ex.printStackTrace();
        	JHVGlobals.err("Error loading FFmpeg", ex.getMessage(), true);
        }

		// Initialize settings and variables
        splash.setProgressText("Loading settings...");
        splash.nextStep();
        Settings.getSingletonInstance().load();

        // If the user has not specified any desired look and feel yet, the system default theme will be used
        if(Settings.getSingletonInstance().getProperty("look.and.feel").length() <= 0)
            Settings.getSingletonInstance().setProperty("look.and.feel", UIManager.getSystemLookAndFeelClassName());

        Settings.getSingletonInstance().update();

        JP2Image.setCachePath(JHVDirectory.CACHE.getFile());
        
        // Load Plug ins
        splash.setProgressText("Loading plug ins...");
        splash.nextStep();
        PluginManager.getSingeltonInstance().loadSettings(JHVDirectory.PLUGINS.getPath());
        
        PluginManager.getSingeltonInstance().addInternalPlugin(new InternalFilterPlugin());
        PluginManager.getSingeltonInstance().addInternalPlugin(new InternalOverlayPlugin());
        
        try {
            PluginManager.getSingeltonInstance().searchForPlugins(JHVDirectory.PLUGINS.getFile(), true);
        } catch (IOException e) {
            String title = "An error occured while loading the plugin files. At least one plugin file is corrupt!";
            String message = "The following files are affected:\n" + e.getMessage();            
            JHVGlobals.warn(title, message);
        }

        // pass the command line arguments to the command line processor
        // so that they are available there
        if ((args.length >= 1) && (!args[0].equals(""))) {
            new CommandLineProcessor().setArguments(args);
        }
        
        // This code updates the ImageViewer
        ImageViewerGui.getSingletonInstance().updateComponents();
        
        // Startup OpenGL and thus show main window
        splash.startUpOpenGL();
        
        // The splash screen will be released in class ImageViewerGUI method packAndShow
    }
};