package org.helioviewer.jhv;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import kdu_jni.KduException;
import kdu_jni.Kdu_global;
import kdu_jni.Kdu_message_formatter;

import org.helioviewer.jhv.io.FileDownloader;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_KduException;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_Kdu_message;

/**
 * This class takes care of loading the Kakadu libs, downloading them
 * if necessary, and setting up the internal Kakadu messaging objects.
 * 
 * @author caplins
 */
public class KakaduEngine {

    /** The JNI shared library as mapped to the system's naming convention. */
    private static final String KDU_JNI_LIB = System.mapLibraryName("kdu_jni");

    /** The aux Kakadu shared library as mapped to the system's naming convention. */
    private static final String KDU_AUX_LIB = System.mapLibraryName("kdu_a63R");

    /** The core Kakadu shared library as mapped to the system's naming convention. */
    private static final String KDU_CORE_LIB = System.mapLibraryName("kdu_v63R");
    
    private static final String SERVER_PATH = "http://www.jhelioviewer.org/libjhv/kdu/";

    /** The operating system. */
    public final String OS;

    /** The operating system version. */
    public final String OS_VERSION;

    /** The system architecture. */
    public final String ARCH;

    /** The platform (enum type) determined via the KakaduEngine constructor. */
    private final Platform PLATFORM;

    /**
     * Static instances of KduSysMessage for both errors and warnings.
     * Although never explicitly used, the references must be maintained since the native
     * code calls back to this.
     */
    private static JHV_Kdu_message warnings, errors;

    /**
     * Static instances of Kdu_message_formatter for both errors and warnings.
     * Although never explicitly used, the references must be maintained since the native
     * code calls back to this.
     */
    private static Kdu_message_formatter warningsFormatter, errorsFormatter;

    /**
     * Default constructor. It determines the Platform.
     */
    public KakaduEngine(){
        OS = System.getProperty("os.name").toUpperCase();
        OS_VERSION = System.getProperty("os.version");
        ARCH = System.getProperty("os.arch").toUpperCase();

        Platform temp = Platform.UNKNOWN;
        if(OS.indexOf("WINDOWS") != -1){
	    	if(ARCH.indexOf("64") != -1)
	    		temp = Platform.WINDOWS_X86_64;
	    	else if(ARCH.indexOf("86") != -1)
	    		temp = Platform.WINDOWS_X86_32;
	    }else if(OS.indexOf("LINUX") != -1){
	    	if(ARCH.indexOf("64") != -1)
	    		temp = Platform.LINUX_X86_64;
	    	else if(ARCH.indexOf("86") != -1)
	    		temp = Platform.LINUX_X86_32;
	    }else if(OS.indexOf("MAC OS X") != -1){
	    	if(ARCH.indexOf("PPC") != -1)
	    		temp = Platform.MAC_PPC;
	    	else if(ARCH.indexOf("64") != -1)
	    		temp = Platform.MAC_X86_64;
	    	else if(ARCH.indexOf("86") != -1)
	    		temp = Platform.MAC_X86_32;
	    } 
        
        if(temp == Platform.UNKNOWN) { 
        	// Fail 'hard'
            JHVGlobals.err(
                    "Cannot detect the OS/ARCH!",
                    "OS - "+OS+"    VERSION - "+OS_VERSION+"\nARCH - "+ARCH,
                    true);
        }
    
        PLATFORM = temp;
    }

    /**
     * Private simple library loading function.
     * 
     * @param lib
     *            name of the library which has to be loaded.
     * @throws UnsatisfiedLinkError
     */
    private void loadLib(String lib) throws UnsatisfiedLinkError{
        String libPath = JHVDirectory.LIBS.getPath() + lib;
        
        try {
			File localLib = new File(libPath);
			URL remoteLib = new URL(SERVER_PATH + PLATFORM.getServerFolder() + lib);
			URLConnection remoteConnection = remoteLib.openConnection();
			
			remoteConnection.setConnectTimeout(2000);
			int remoteLibLength = remoteConnection.getContentLength();
			
			if(localLib.length() != remoteLibLength && remoteLibLength > 0) {
				throw new UnsatisfiedLinkError(libPath + " is outdated or corrupted");
			}
		
        } catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {}
                        
        System.load(libPath);
    }

    /**
     * Attempts to setup the Kakadu message handlers. Honestly, if this fails
     * then there is probably a larger problem and it should error and exit.
     * 
     * @throws JHV_KduException
     */
    public void startKduMessageSystem() throws JHV_KduException{
        try {
            warnings = new JHV_Kdu_message(false);
            errors = new JHV_Kdu_message(true);
            warningsFormatter = new Kdu_message_formatter(warnings, 80);
            errorsFormatter = new Kdu_message_formatter(errors, 80);
            Kdu_global.Kdu_customize_warnings(warningsFormatter);
            Kdu_global.Kdu_customize_errors(errorsFormatter);
        } catch (KduException ex) {
            throw new JHV_KduException("Error initializing Kakadu error handler:\n"+ex.getMessage());
        }
    }

    /**
     * Attempts to delete the old libraries.
     * 
     * @throws SecurityException If the access to delete the files is denied
     */
    public void deleteLibs() throws SecurityException{
        List<String> libs = PLATFORM.getLibNames();
        for(String lib : libs){
            File libFile = new File(JHVDirectory.LIBS.getPath()+lib);
            if(libFile.exists())
                libFile.delete();
        }
    }

    /**
     * Attempts to load the requisite libraries.
     * 
     * @throws UnsatisfiedLinkError If at least one library does not exist
     */
    public void loadLibs() throws UnsatisfiedLinkError{
        List<String> libs = PLATFORM.getLibNames();
        for(String lib : libs)
            loadLib(lib);
    }

    /**
     * Returns the number of required libraries depending on the current
     * platform.
     * 
     * @return number of required libraries.
     */
    public int getNumberOfLibs() {
        
        return PLATFORM.getLibNames().size();
    }
    
    /**
     * Attempts to download the requisite libs from the server.
     * 
     * @param splash
     *            instance of the splash screen to retrieve the progress bar and
     *            status label.
     * @throws IOException
     */
    public void downloadLibs(JHVSplashScreen splash) throws IOException{
    	
    	splash.setProgressText("Installing JPEG2000 Libraries...");
                
        FileDownloader fileDownloader = new FileDownloader();
        List<String> libs = PLATFORM.getLibNames();
        
        for(String lib : libs){
            
            splash.nextStep();
            
            try {
				fileDownloader.get(
				        new URI(SERVER_PATH + PLATFORM.getServerFolder() + lib), 
				        JHVDirectory.LIBS.getFile(), 
				        splash.getProgressBar());
			} catch (URISyntaxException e) {
				throw new IOException(e.getMessage());
			}
        }
    }

    /**
     * The Platform enum type, handles all the Platform dependent stuff.
     * 
     * @author caplins
     */
    static enum Platform {
        WINDOWS_X86_32{
            @Override
            public String getServerFolder() {return "windows/x86-32/";}
            @Override
            public List<String> getLibNames() {return createList(KDU_CORE_LIB, KDU_AUX_LIB, KDU_JNI_LIB);}
        },
        WINDOWS_X86_64{
            @Override
            public String getServerFolder() {return "windows/x86-64/";}
            @Override
            public List<String> getLibNames() {return createList(KDU_CORE_LIB, KDU_AUX_LIB, KDU_JNI_LIB);}
        },
        LINUX_X86_32{
            @Override
            public String getServerFolder() {return "linux/x86-32/";}
        },
        LINUX_X86_64{
            @Override
            public String getServerFolder() {return "linux/x86-64/";}
        },
        MAC_PPC{
            @Override
            public String getServerFolder() {return "macosx/ppc/";}
        },
        MAC_X86_32{
            @Override
            public String getServerFolder() {return "macosx/x86-32/";}
        },
        MAC_X86_64{
            @Override
            public String getServerFolder() {return "macosx/x86-64/";}
        },
        UNKNOWN{
            @Override
            public String getServerFolder() {return "";}
            @Override
            public List<String> getLibNames() {return Collections.emptyList();}
        };

        /**
         * This method must be overridden for each new library.
         * 
         * @return the server folder depending on the platform.
         * */
        abstract public String getServerFolder();

        /** 
         * This must be overridden for any Platform that required more than just the JNI lib loaded.
         * 
         * @return list of library names.
         * */
        public List<String> getLibNames() {
            return createList(KDU_JNI_LIB);
        }

        /**
         * This is a private helper method. Accepts an arbitrary number of args
         * which correspond to the libraries that need to be loaded in the order
         * that they should be loaded. Returns a list version of the arguments.
         * 
         * @param _elements
         * @return a list version of the arguments.
         */
        private static List<String> createList(String ... _elements) {
            List<String> ret = new LinkedList<String>();
            
            for(String element : _elements)	
                ret.add(element);
            
            return ret;
        }
    };
};