package org.helioviewer.jhv;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Properties;

import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.opengl.GLInfo;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_Kdu_cache;

/**
 * A class that stores and reads default values in a settings file.
 * 
 * @author Benjamin Wamsler
 * @author Juan Pablo
 * @author Markus Langenberg
 */
public class Settings {
    /** The sole instance of this class. */
    private static final Settings singletonInstance = new Settings();

    /** The properties object */
    private Properties properties = new Properties();

    /** The current version of the properties */
    private String version = "03/19/10 12:00PM";

    /** The properties file */
    private File propFile = new File(JHVDirectory.HOME.getPath()
            + "defaults.properties");

    /** The original console out stream */
    private PrintStream consoleOut = System.out;

    /** The original console err stream */
    private PrintStream consoleErr = System.err;

    /** Is there any property modified? */
    private boolean modified = false;

    /**
     * The private constructor of this class.
     * */
    private Settings() {
    }

    /**
     * Method initializes the settings and creates settings file if none exists.
     * */
    public void load()
    {
        try {
            properties.clear();

            if(propFile.exists()) {
                FileInputStream fileInput = new FileInputStream(propFile);
                properties.load(fileInput);
                fileInput.close();

                if(!version.equals(getProperty("version"))) {
                    properties.clear();
                    propFile.delete();
                }
            }

            if(!propFile.exists()) {
                setProperty("version", version);

                setProperty("default.save.path", JHVDirectory.EXPORTS.getPath());
                setProperty("default.local.path", JHVDirectory.HOME.getPath());
                setProperty("default.remote.path",
                        "jpip://delphi.nascom.nasa.gov:8090");
                setProperty("default.basicRemote.path",
                        "jpip://delphi.nascom.nasa.gov:8090");
                setProperty("default.httpRemote.path",
                        "http://delphi.nascom.nasa.gov/jp2/");

                setProperty("API.jp2images.path", "http://delphi.nascom.nasa.gov/hv/api/index.php");
                setProperty("API.jp2series.path", "http://delphi.nascom.nasa.gov/hv/api/index.php");
                setProperty("API.event.path",
                        "http://achilles.nascom.nasa.gov/~wamsler/API/");

                setProperty("look.and.feel", "");
                
                setProperty("help.linkingmovies.show", Boolean.toString(true));
                
                setProperty("opengl.enabled", Boolean.toString(true));

                setProperty("debug.log.enabled", Boolean.toString(false));
                setProperty("debug.log.times", Boolean.toString(false));

                setProperty("jpip.cache.size", Double.toString(0));

                FileOutputStream fileOutput = new FileOutputStream(propFile);
                properties.store(fileOutput, null);
                fileOutput.close();
            }

        } catch(Exception ex)
        {
            ex.printStackTrace();
        }
    }

    /**
     * This method saves all the values in the properties file.
     */
    public void save()
    {
        try {
            if(modified) {
                FileOutputStream fileOutput = new FileOutputStream(propFile);
                properties.store(fileOutput, null);
                fileOutput.close();
                modified = false;
            }

        } catch(IOException ex)
        {
            ex.printStackTrace();
        }
    }

    /**
     * The new property values are updated.
     */
    public void update()
    {
        try {
            String val;
            
            val = getProperty("opengl.enabled");
			GLInfo.setGlEnabled(Boolean.parseBoolean(val));

            val = getProperty("debug.log.enabled");

            if(!Boolean.parseBoolean(val)) {
                System.setOut(consoleOut);
                System.setErr(consoleErr);

            } else if(System.out == consoleOut) {
                Date now = new Date();
                SimpleDateFormat formatter = new SimpleDateFormat("yyyyMMdd-HHmmss");
                String logFile = JHVDirectory.LOGS.getPath() + formatter.format(now) + ".log";
                PrintStream logStream = new PrintStream(new FileOutputStream(logFile));

                System.setOut(logStream);
                System.setErr(logStream);
            }

            val = getProperty("look.and.feel");

            setLookAndFeelEverywhere(val);

            val = getProperty("jpip.cache.size");

            double size = 0;

            try {
                size = Double.valueOf(val);
            } catch(NumberFormatException ex) {}

            setProperty("jpip.cache.size", Double.toString(size));

            JHV_Kdu_cache.updateCacheDirectory(JHVDirectory.CACHE.getFile(), size);

        } catch(Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Method sets the value of a specified default field.
     * 
     * @param key
     *            Default field to be written to
     * @param val
     *            Value to be set to
     */
    public void setProperty(String key, String val)
    {
        properties.setProperty(key, val);
        modified = true;
    }

    /**
     * Method that returns the value of the specified default field.
     * 
     * @param key
     *            Default field to read
     */
    public String getProperty(String key)
    {
        return properties.getProperty(key);
    }

    /** 
     * Method returns the sole instance of this class. 
     * 
     * @return returns the sole instance of this class.
     * */
    public static Settings getSingletonInstance()
    {
        return singletonInstance;
    }
    
    /**
     * Sets the look and feel to all windows of the application.
     * 
     * @param lookAndFeel name of the lookandfeel.
     */
    public void setLookAndFeelEverywhere(String lookAndFeel) {
		
		if(!UIManager.getLookAndFeel().getClass().getName().equals(lookAndFeel)) {
    		try {
        		UIManager.setLookAndFeel(lookAndFeel);
        		
        		ImageViewerGui.getSingletonInstance();
        		SwingUtilities.updateComponentTreeUI(ImageViewerGui.getMainFrame());
				
    		} catch(Exception ex)
    		{
    			ex.printStackTrace();
    		}			
		}
		
		
	}

    /**
     * The values are saved to disk only if there have been a modification.
     */
    protected void finalize()
    {
        save();
    }
}
