package org.helioviewer.jhv.ffmpeg;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;

import org.helioviewer.jhv.JHVDirectory;
import org.helioviewer.jhv.JHVGlobals;
import org.helioviewer.jhv.JHVSplashScreen;
import org.helioviewer.jhv.io.FileDownloader;

/** Class responsible for downloading ffmpeg.
 * 
 * <p>Before being able to call ffmpeg, the native ffmpeg command line
 * tool has to be available. Therefore, is has to be downloaded, if
 * it is not available on the machine yet. 
 * 
 * @author Markus Langenberg
 */
public class FFmpegLoader {

	private static final String SERVER_PATH = "http://www.jhelioviewer.org/libjhv/ffmpeg/";
	
	private final String OS;
	private final String ARCH;
    private final Platform PLATFORM;
    
    private static String ffmpegLocal;
    
    /**
     * Default constructor.
     * 
     * Determines operating system, architecture and platform, to decide
     * which libraries are needed.
     */
	public FFmpegLoader() {
		OS = System.getProperty("os.name").toUpperCase();
	    ARCH = System.getProperty("os.arch").toUpperCase();
	    
	    Platform temp = Platform.UNKNOWN;
        if(OS.indexOf("WINDOWS") != -1){
	    	if(ARCH.indexOf("64") != -1)
	    		temp = Platform.WINDOWS_X86_64;
	    	else if(ARCH.indexOf("86") != -1)
	    		temp = Platform.WINDOWS_X86_32;
	    }else if(OS.indexOf("LINUX") != -1){
	    	if(ARCH.indexOf("64") != -1)
	    		temp = Platform.LINUX_X86_64;
	    	else if(ARCH.indexOf("86") != -1)
	    		temp = Platform.LINUX_X86_32;
	    }else if(OS.indexOf("MAC OS X") != -1){
	    	if(ARCH.indexOf("PPC") != -1)
	    		temp = Platform.MAC_PPC;
	    	else if(ARCH.indexOf("64") != -1)
	    		temp = Platform.MAC_X86_64;
	    	else if(ARCH.indexOf("86") != -1)
	    		temp = Platform.MAC_X86_32;
	    }
	    
	    if(temp == Platform.UNKNOWN) {
	        // Fail 'hard'
	        JHVGlobals.err(
	                "Cannot detect the OS/ARCH!",
	                "OS - " + OS + "\nARCH - " + ARCH,
	                true);
	    }
	    PLATFORM = temp;
	}
	
	/**
     * Returns the number of required libraries.
     * 
     * @return number of required libraries.
     */
    public int getNumberOfLibs() {
    	return 1;
    }
	
	/** Checks whether the ffmpeg command line tool is present, downloads it if not.
     * 
     * @param splash
     *            instance of the splash screen to retrieve the progress bar and
     *            status label.
     * @throws IOException If the server is unreachable or the files are not found
     */
    public void checkAndDownloadFFMPEG(JHVSplashScreen splash) throws IOException {
        
        splash.nextStep();
        
    	String ffmpegName = "ffmpeg";
    	if(PLATFORM == Platform.WINDOWS_X86_32 || PLATFORM == Platform.WINDOWS_X86_64)
    		ffmpegName += ".exe";
    	
    	ffmpegLocal = JHVDirectory.LIBS.getPath() + ffmpegName;
    		
    	File ffmpeg = new File(ffmpegLocal);
    	URL remoteffmpeg = new URL(SERVER_PATH + PLATFORM.getServerFolder() + ffmpegName);
    	URLConnection remoteConnection = remoteffmpeg.openConnection();
    	
    	remoteConnection.setConnectTimeout(2000);
		int remoteffmpegLength = remoteConnection.getContentLength();
		
		if(!ffmpeg.exists() || (ffmpeg.length() != remoteffmpegLength && remoteffmpegLength > 0)) {
    		FileDownloader fileDownloader = new FileDownloader();
    		splash.setProgressText("Installing Movie Encoder...");
        
    		try {
				fileDownloader.get(
				        new URI(SERVER_PATH + PLATFORM.getServerFolder() + ffmpegName), 
				        new File(JHVDirectory.LIBS.getPath()), 
				        splash.getProgressBar());
			} catch (URISyntaxException e) {
				throw new IOException(e.getMessage());
			}
    	}
    	
    	if(PLATFORM != Platform.WINDOWS_X86_32 && PLATFORM != Platform.WINDOWS_X86_64) {
    		Runtime.getRuntime().exec("chmod 755 " + ffmpeg);
    	}
    }
    
    /** Returns the path, where to find ffmpeg
     * 
     * @return Path, where to find ffmpeg
     */
    public static String getffmpegPath() {
    	return ffmpegLocal;
    }
    
    static enum Platform {
        WINDOWS_X86_32{
            @Override
            public String getServerFolder() {return "windows/x86-32/";}
        },
        WINDOWS_X86_64{
            @Override
            public String getServerFolder() {return "windows/x86-64/";}
        },
        LINUX_X86_32{
            @Override
            public String getServerFolder() {return "linux/x86-32/";}
        },
        LINUX_X86_64{
            @Override
            public String getServerFolder() {return "linux/x86-64/";}
        },
        MAC_PPC{
            @Override
            public String getServerFolder() {return "macosx/ppc/";}
        },
        MAC_X86_32{
            @Override
            public String getServerFolder() {return "macosx/x86-32/";}
        },
        MAC_X86_64{
            @Override
            public String getServerFolder() {return "macosx/x86-64/";}
        },
        UNKNOWN{
            @Override
            public String getServerFolder() {return "";}
        };

        abstract public String getServerFolder();
    }
}
