package org.helioviewer.jhv.gui.actions;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.IOException;

import javax.swing.AbstractAction;
import javax.swing.JFileChooser;
import javax.swing.KeyStroke;
import javax.swing.filechooser.FileFilter;

import org.helioviewer.jhv.JHVGlobals;
import org.helioviewer.jhv.Settings;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.io.APIRequestManager;

/**
 * Action to open a local file.
 * 
 * <p>Opens a file chooser dialog, opens the selected file.
 * Currently supports the following file extensions: "jpg", 
 * "jpeg", "png", "fts", "fits", "jp2" and "jpx"
 * 
 * @author Markus Langenberg
 */
public class OpenLocalFileAction extends AbstractAction {
	
	private static final long serialVersionUID = 1L;
	
	/**
	 * Default constructor.
	 */
	public OpenLocalFileAction() {
		super("Open...");
		putValue(SHORT_DESCRIPTION, "Open New Image");
		putValue(ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_O, KeyEvent.CTRL_MASK));
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void actionPerformed(ActionEvent e) {		
	    
	    final JFileChooser fileChooser = new JFileChooser(Settings.getSingletonInstance().getProperty("default.local.path"));
	    fileChooser.setAcceptAllFileFilterUsed(false);
        fileChooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
        fileChooser.addChoosableFileFilter(new JP2Filter());
        fileChooser.addChoosableFileFilter(new FitsFilter());
        fileChooser.addChoosableFileFilter(new PNGFilter());
        fileChooser.addChoosableFileFilter(new JPGFilter());
        fileChooser.setFileFilter(new AllSupportedTypesFilter());
        fileChooser.setMultiSelectionEnabled(false);
        
        fileChooser.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent _e) {
                if (_e.getActionCommand().equals(JFileChooser.APPROVE_SELECTION)
                        && fileChooser.getSelectedFile().exists() && fileChooser.getSelectedFile().isFile()) {

                    fileChooser.setVisible(false);
                    
                    // remember the current directory for future
                    Settings.getSingletonInstance().setProperty("default.local.path", fileChooser.getSelectedFile().getParent());
                    Settings.getSingletonInstance().save();
                    
                    ImageViewerGui.getSingletonInstance().getMainImagePanel().setLoading(true);
                     
                    Thread thread = new Thread(new Runnable() {
                        
                        public void run() {
                            try {
                            	APIRequestManager.newLoad(fileChooser.getSelectedFile().toURI());
                            } catch (IOException e) {
                                JHVGlobals.err("An error occured while opening the file!", e.getMessage(), false);
                                fileChooser.setVisible(true);
                            } finally {
                                ImageViewerGui.getSingletonInstance().getMainImagePanel().setLoading(false);
                            }
                        }
                    }, "OpenLocalFile");
                    
                    thread.start();
                }
            }
        });    
	    
        fileChooser.showOpenDialog(ImageViewerGui.getMainFrame());
	}
	
    private class AllSupportedTypesFilter extends FileFilter {

        private final String[] extensions = {"jpg", "jpeg", "png", "fts", "fits", "jp2", "jpx"};
        
        /**
    	 * {@inheritDoc}
    	 */
    	public boolean accept(File f) {
            
            if (f.isDirectory())
                return true;
            
            String testName = f.getName().toLowerCase();
            for (String ext : extensions) {
                if (testName.endsWith(ext))
                    return true;
            }
            
            return false;
        }

    	/**
    	 * {@inheritDoc}
    	 */
    	public String getDescription() {
            return "All supported files (\".jpg\", \".jpeg\", \".png\", \".fts\", \".fits\", \".jp2\", \".jpx\")";
        }
    }
    
    private class JP2Filter extends FileFilter {

        private final String[] extensions = {"jp2", "jpx"};
        
        /**
    	 * {@inheritDoc}
    	 */
    	public boolean accept(File f) {
            
            if (f.isDirectory())
                return true;
            
            String testName = f.getName().toLowerCase();
            for (String ext : extensions) {
                if (testName.endsWith(ext))
                    return true;
            }
            
            return false;
        }

    	/**
    	 * {@inheritDoc}
    	 */
    	public String getDescription() {
            return "JPG2000 files (\".jp2\", \".jpx\")";
        }
    }
    
    private class FitsFilter extends FileFilter {

        private final String[] extensions = {"fts", "fits"};
        
        /**
    	 * {@inheritDoc}
    	 */
    	public boolean accept(File f) {
            
            if (f.isDirectory())
                return true;
            
            String testName = f.getName().toLowerCase();
            for (String ext : extensions) {
                if (testName.endsWith(ext))
                    return true;
            }
            
            return false;
        }

    	/**
    	 * {@inheritDoc}
    	 */
    	public String getDescription() {
            return "FITS files (\".fts\", \".fits\")";
        }
    }
    
    private class PNGFilter extends FileFilter {

        private final String[] extensions = {"png"};
        
        /**
    	 * {@inheritDoc}
    	 */
    	public boolean accept(File f) {
            
            if (f.isDirectory())
                return true;
            
            String testName = f.getName().toLowerCase();
            for (String ext : extensions) {
                if (testName.endsWith(ext))
                    return true;
            }
            
            return false;
        }

    	/**
    	 * {@inheritDoc}
    	 */
    	public String getDescription() {
            return "PNG files (\".png\")";
        }
    }
    
    private class JPGFilter extends FileFilter {

        private final String[] extensions = {"jpg", "jpeg"};
        
        /**
    	 * {@inheritDoc}
    	 */
    	public boolean accept(File f) {
            
            if (f.isDirectory())
                return true;
            
            String testName = f.getName().toLowerCase();
            for (String ext : extensions) {
                if (testName.endsWith(ext))
                    return true;
            }
            
            return false;
        }

    	/**
    	 * {@inheritDoc}
    	 */
    	public String getDescription() {
            return "JPG files (\".jpg\", \".jpeg\")";
        }
    }
}
