package org.helioviewer.jhv.gui.components;

import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Arrays;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JEditorPane;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JWindow;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

import org.helioviewer.jhv.JHVGlobals;
import org.helioviewer.jhv.gui.IconBank;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.gui.IconBank.JHVIcon;
import org.helioviewer.jhv.internal_plugins.overlay.solarevents.SolarEvent;
import org.helioviewer.jhv.internal_plugins.overlay.solarevents.SolarEventCatalogs;
import org.helioviewer.jhv.internal_plugins.overlay.solarevents.SolarEvent.SolarEventException;

/** Popup displaying informations about an solar event.
 * 
 * <p>This panel is a JWindow, so that it can be displayed on
 * top of an GLCanvas, which is not possible for other swing
 * components.
 * 
 * <p>For further informations about solar events, see
 * {@link org.helioviewer.jhv.solarevents}.
 * 
 * @author Markus Langenberg
 *
 */
public class EventInformationWindow extends JWindow implements ActionListener, MouseListener, HyperlinkListener {

	private static final long serialVersionUID = 1L;
	
	private static final String sep = "<br>";
	
	private static final Cursor defaultCursor = Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);
	private static final Cursor clickCursor = Cursor.getPredefinedCursor(Cursor.HAND_CURSOR);
	
	private SolarEvent event;
	
	private JButton closeButton = new JButton(IconBank.getIcon(JHVIcon.EX));
	private JLabel header = new JLabel("Header");
	private JLabel description = new JLabel("Description");
	private JLabel leftColumn = new JLabel("left");
	private JEditorPane rightColumn = new JEditorPane("text/html", "right");

	/**
	 * Default constructor.
	 */
	public EventInformationWindow() {
		super(ImageViewerGui.getMainFrame());
		//setAlwaysOnTop(false);
		setFocusable(false);

		setLayout(new BorderLayout());
		setCursor(clickCursor);
		
		JPanel topContainer = new JPanel(new BorderLayout());
		
		closeButton.addActionListener(this);
		closeButton.addMouseListener(this);
		
		header.setFont(header.getFont().deriveFont(Font.BOLD));
		description.setFont(description.getFont().deriveFont(10.0f));
		leftColumn.setFont(leftColumn.getFont().deriveFont(10.0f).deriveFont(Font.BOLD));
		
		closeButton.setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));
		header.setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));
		description.setBorder(BorderFactory.createEmptyBorder(0, 3, 0, 3));
		leftColumn.setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));
		rightColumn.setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));
		
		leftColumn.setVerticalTextPosition(JLabel.TOP);
		
		rightColumn.setEditable(false);
		rightColumn.setOpaque(false);
		rightColumn.addHyperlinkListener(this);
		
		topContainer.add(header, BorderLayout.WEST);
		topContainer.add(closeButton, BorderLayout.EAST);
		topContainer.add(description, BorderLayout.SOUTH);
		add(topContainer, BorderLayout.PAGE_START);
		add(leftColumn, BorderLayout.WEST);
		add(rightColumn, BorderLayout.EAST);
		
		this.pack();
	}
	
	/** Sets the corresponding solar event.
	 * 
	 * @param newEvent Corresponding solar event
	 */
	public void setEvent(SolarEvent newEvent)
	{
		if(event == newEvent || newEvent == null)
			return;
		
		event = newEvent;
		
		StringBuilder left = new StringBuilder();
		StringBuilder right = new StringBuilder();
			
		try {			
			header.setText(SolarEventCatalogs.getSingletonInstance().getCatalogByID(event.getCatalogID()).getName());
			description.setText(event.getShortInfo());
			
			left.append("<html>");
			right.append("<html><font style=\"font-family: '" + leftColumn.getFont().getFamily() + "'; font-size: 10;\">");
			left.append("Start:" + sep);
			right.append(event.getStartTime() + sep);
			left.append("End:" + sep);
			right.append(event.getEndTime() + sep);
			
			String[] names = event.getPropertyNames();
			Arrays.sort(names);
			for(String s : names) {
				left.append(s + ":" + sep);
				right.append(event.getStringFromCategory("properties", s) + sep);				
			}
			
//			if(!event.isNull("properties")) {
//				JSONObject properties = event.getJSONObject("properties");
//				String[] names = JSONObject.getNames(properties);
//				Arrays.sort(names);
//				for(String s : names) {
//					left.append(s + ":" + sep);
//					right.append(properties.getString(s) + sep);				
//				}
//			}
			
			left.append("Source:");
			right.append("<a href=\"" + event.getSourceURL() + "\">" + event.getSourceURL() + "</a>");
		} catch (SolarEventException e) {
			e.printStackTrace();
		}
		
		left.append("</html>");
		right.append("</font></html>");
		leftColumn.setText(left.toString());
		rightColumn.setText(right.toString());
		
		setSize(Math.max(description.getPreferredSize().width, 
				Math.max(header.getPreferredSize().width + closeButton.getPreferredSize().width,
				leftColumn.getPreferredSize().width + rightColumn.getPreferredSize().width)),
				
				header.getPreferredSize().height + description.getPreferredSize().height + leftColumn.getPreferredSize().height);
	}

	/**
	 * {@inheritDoc}
	 */
	public void actionPerformed(ActionEvent e) {
		setVisible(false);
	}

	/**
	 * {@inheritDoc}
	 */
	public void hyperlinkUpdate(HyperlinkEvent e) {
		if(e.getEventType() == HyperlinkEvent.EventType.ENTERED) {
			setCursor(clickCursor);
		} else if(e.getEventType() == HyperlinkEvent.EventType.EXITED) {
			setCursor(defaultCursor);
		} else {
			JHVGlobals.openURL(e.getURL().toString());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public void mouseClicked(MouseEvent e) {}

	/**
	 * {@inheritDoc}
	 */
	public void mouseEntered(MouseEvent e) {
		setCursor(clickCursor);
	}

	/**
	 * {@inheritDoc}
	 */
	public void mouseExited(MouseEvent e) {
		setCursor(defaultCursor);
	}

	/**
	 * {@inheritDoc}
	 */
	public void mousePressed(MouseEvent e) {}

	/**
	 * {@inheritDoc}
	 */
	public void mouseReleased(MouseEvent e) {}
}
