package org.helioviewer.jhv.gui.components;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.AbstractList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;
import javax.swing.border.Border;

import org.helioviewer.jhv.Settings;
import org.helioviewer.jhv.gui.IconBank;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.gui.ViewListenerDistributor;
import org.helioviewer.jhv.gui.IconBank.JHVIcon;
import org.helioviewer.jhv.gui.components.statusplugins.ActiveLayerChangedReason;
import org.helioviewer.jhv.gui.dialogs.MetaDataDialog;
import org.helioviewer.jhv.io.FileDownloader;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.changeevent.LayerChangedReason;
import org.helioviewer.viewmodel.changeevent.MetaDataChangedReason;
import org.helioviewer.viewmodel.changeevent.LayerChangedReason.LayerChangeType;
import org.helioviewer.viewmodel.metadata.HelioviewerMetaData;
import org.helioviewer.viewmodel.metadata.MetaData;
import org.helioviewer.viewmodel.metadata.ObserverMetaData;
import org.helioviewer.viewmodel.view.ImageInfoView;
import org.helioviewer.viewmodel.view.LayeredView;
import org.helioviewer.viewmodel.view.MetaDataView;
import org.helioviewer.viewmodel.view.MovieView;
import org.helioviewer.viewmodel.view.SynchronizeView;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.ViewListener;
import org.helioviewer.viewmodel.view.jp2view.JHVJP2View;
import org.helioviewer.viewmodel.view.jp2view.JP2Image;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_KduException;

/** Panel for displaying all layers including the layer specific controls.
 * 
 * <p>This panel plays are very central role: It manages all visible layers,
 * including the movie controls and all filter controls. 
 * 
 * @author Markus Langenberg
 */
public class ImageSelectorPanel extends JPanel implements ViewListener{

	private static final long serialVersionUID = 1L;
	private ImageInfoView activeView;
	private LayeredView layeredView;
	private HashMap<ImageInfoView, ImagePanel> associatedPanels = new HashMap<ImageInfoView, ImagePanel>();
	
	/**
	 * Default constructor.
	 */
	public ImageSelectorPanel() {
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		
		ViewListenerDistributor.getSingletonInstance().addViewListener(this);
	}
	
	/** Sets the current LayeredView.
	 * 
	 * This reference is needed to remove layer and toggle
	 * their visibility.
	 * 
	 * @param _layeredView Current LayeredView
	 */
	public void setLayeredView(LayeredView _layeredView) {
		layeredView = _layeredView;
	}
	
	/** Returns the active layer.
	 * 
	 * Many actions and status panel plugins refer to the active
	 * layer.
	 * 
	 * @return Active layer
	 */
	public ImageInfoView getActiveImageInfoView() {
		return activeView;
	}
	
	/** 
	 * Returns the view, which is a direct successor of the
	 * LayeredView and a predecessor of the active view.
	 * 
	 * @return Child of the LayeredView in the same branch as the active view
	 */
	public View getActiveChildFromLayeredView() {
		return findLastViewBeforeLayeredView(activeView);
	}
	
	/** Sets the active view.
	 * 
	 * @param view New active view
	 */
	public void setActiveView(ImageInfoView view) {
		for(Component panel : getComponents()) {
			if(panel instanceof ImagePanel) 
			((ImagePanel) panel).setHighlighted(false);
		}
		
		if(view != null) {
			associatedPanels.get(view).setHighlighted(true);
		}
		
		activeView = view;
		
		ViewListenerDistributor.getSingletonInstance().viewChanged(activeView, new ChangeEvent(new ActiveLayerChangedReason(activeView)));
	}
	
	/** Adds a new layer to the panel.
	 * 
	 * Also sets its corresponding option pane. The option pane is
	 * a tabbed pane containing all the layer specific controls, such
	 * as movie controls and filter controls.
	 * 
	 * @param newView New layer - Does not have to be the ImageInfoView itself
	 * @param optionPane Option pane corresponding to the new layer
	 */
	public void addLayer(View newView, JTabbedPane optionPane) {
		ImageInfoView imageInfoView = newView.getAdapter(ImageInfoView.class);
		ImagePanel newPanel = new ImagePanel(imageInfoView, optionPane);
		
		associatedPanels.put(imageInfoView, newPanel);
		add(newPanel, 1);
		
		setActiveView(imageInfoView);
		
		this.revalidate();
		
		ImageViewerGui.getSingletonInstance().getLeftContentPane().expand(this);
	}
	
	/**
     * Updates the visual behavior of the component.
     */
	public void updateComponent() {
	    
	    // update all headers of image panels
	    for (ImagePanel ip : associatedPanels.values()) {
	        ip.updateName();
	    }
	}

	/**
	 * {@inheritDoc}
	 */
	public void viewChanged(View sender, ChangeEvent aEvent) {
		LayerChangedReason layerReason = aEvent.getLastChangedReasonByType(LayerChangedReason.class);
		
		// if layers were changed, perform same operations on GUI
		if(layerReason != null) {
			
			// If layer was deleted, delete  corresponding panel
			if(layerReason.getLayerChangeType() == LayerChangedReason.LayerChangeType.LAYER_REMOVED) {
				
				View deletedView = layerReason.getSubView().getAdapter(ImageInfoView.class);
				
				if(associatedPanels.containsKey(deletedView)) {
					remove(associatedPanels.get(deletedView));
					associatedPanels.get(deletedView).abolish();
					associatedPanels.remove(deletedView);
					
					if(associatedPanels.size() == 0) {
						setActiveView(null);
						((collapsibleObservation) this.getComponent(0))
                        .explode();
					} else if(activeView == deletedView || activeView == null) {
						setActiveView((ImageInfoView) associatedPanels.keySet().toArray()[0]);
					}
					
					this.revalidate();
				}
			} else if(layerReason.getLayerChangeType() == LayerChangedReason.LayerChangeType.LAYER_VISIBILITY) {
				associatedPanels.get(layerReason.getSubView().getAdapter(ImageInfoView.class)).setLayerVisible(
						layeredView.isVisible(layerReason.getSubView()));
			}
		}
		
		List<MetaDataChangedReason> metaDataReasons = aEvent.getAllChangedReasonsByType(MetaDataChangedReason.class);
		
		// if meta data has changed, update label
		for(MetaDataChangedReason metaDataReason : metaDataReasons) {
			associatedPanels.get(metaDataReason.getView()).updateName();
		}
	}
	
	/** 
	 * Searches the direct successor of the LayeredView being a predecessor
	 * of the given view.
	 * Therefore, this functions traverses recursively through all the view
	 * listeners of the given view.
	 * 
	 * @param aView Starting view for the search
	 * @return View being a direct successor of the LayeredView and a predecessor
	 * of the given view
	 */
	private View findLastViewBeforeLayeredView(View aView) {
		
		AbstractList<ViewListener> viewListeners = aView.getAllViewListener();		
		
		for (ViewListener v: viewListeners) {
			
			if (v instanceof LayeredView)
				return aView;
			else {
				if (v instanceof View){
					View result = findLastViewBeforeLayeredView((View)v);
					
					if (result != null)
						return result;
				}
			}
		}            
		
		return null;		
	}
	
	/** 
	 * Represents a single layer panel
	 */
	private class ImagePanel extends JPanel implements ActionListener, MouseListener {
		
		private static final long serialVersionUID = 1L;
		
		private String baseName = "";
		private ImageInfoView view;
		
		private JLabel lblTitle;
		private JButton cmdDelete;
		private JButton cmdLayerFunctions;
		private JMenuItem miDownload;
		private JMenuItem miVisibility;
		private JMenuItem miInfo;
		private JTabbedPane bottomPane;
		private JPopupMenu popupLayerFunctions;
		
		private final Border normalBorder = BorderFactory.createCompoundBorder(
				BorderFactory.createEmptyBorder(2,1,2,1), 
				BorderFactory.createLineBorder(Color.LIGHT_GRAY, 1));
		private final Border highlightedBorder = BorderFactory.createCompoundBorder(
				BorderFactory.createEmptyBorder(1,0,1,0), 
				BorderFactory.createLineBorder(Color.GRAY, 2));
		
		/** Default constructor.
		 * 
		 * Receives its option pane. The option pane is a tabbed pane 
		 * containing all the layer specific controls, such as movie 
		 * controls and filter controls.
		 * 
		 * @param associatedView ImageInfoView of the new layer
		 * @param optionPane Option pane for the new layer
		 */
		public ImagePanel(ImageInfoView associatedView, JTabbedPane optionPane) {
			super(new BorderLayout());			
			addMouseListener(this);
			
			view = associatedView;

			// Set name
			lblTitle = new JLabel("", IconBank.getIcon(JHVIcon.DOWN2), SwingConstants.LEFT);
			lblTitle.setBorder(BorderFactory.createEmptyBorder(0, 1, 0, 0));
			lblTitle.addMouseListener(this);
			add(lblTitle, BorderLayout.CENTER);
			
			// Add buttons
			JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT,1,1));
			buttonPanel.setBorder(BorderFactory.createEmptyBorder(1, 0, 0, 0));
			add(buttonPanel, BorderLayout.EAST);
			
			cmdDelete = createButton(IconBank.getIcon(JHVIcon.REMOVE_LAYER), "Remove Layer");
			buttonPanel.add(cmdDelete);
			
			if(view instanceof MovieView) {	
				JButton lockMovieButton = MoviePanel.getLinkMovieButton((MovieView) view);
				if(lockMovieButton != null)
					buttonPanel.add(lockMovieButton);
			}
			
			popupLayerFunctions = new JPopupMenu();
			
			if(view.getUri().getScheme().equalsIgnoreCase("jpip")) {
				String title;
				if(view instanceof MovieView) {
					title = "Download Movie";
				} else {
					title = "Download Image";
				}
    			miDownload = new JMenuItem(title, IconBank.getIcon(JHVIcon.DOWNLOAD));
    			miDownload.addActionListener(this);
    			popupLayerFunctions.add(miDownload);
			}
			
			miVisibility = new JMenuItem("Hide Layer", IconBank.getIcon(JHVIcon.VISIBLE));
			miVisibility.addActionListener(this);
            popupLayerFunctions.add(miVisibility);
			
			MetaData metaData = view.getAdapter(MetaDataView.class).getMetaData();
			if(metaData instanceof HelioviewerMetaData) {
			    
			    miInfo = new JMenuItem("Show Layer Information", IconBank.getIcon(JHVIcon.INFO));
			    miInfo.addActionListener(this);
	            popupLayerFunctions.add(miInfo);
				
				baseName = view.getName();
				updateName();
				
			} else {
				lblTitle.setText(view.getName());
			}
			
			lblTitle.setPreferredSize(new Dimension(
					Math.min(lblTitle.getPreferredSize().width + 5, ImageViewerGui.SIDE_PANEL_WIDTH - buttonPanel.getPreferredSize().width - 5),
					lblTitle.getPreferredSize().height));
			
			cmdLayerFunctions = createButton(IconBank.getIcon(JHVIcon.SHOW_MORE), "Show More Layer Functions");
			cmdLayerFunctions.addActionListener(this);
            buttonPanel.add(cmdLayerFunctions);
			
			bottomPane = optionPane;
            bottomPane.setVisible(true);
            bottomPane.addMouseListener(this);
            bottomPane.setOpaque(true);
            addMouseListenerToAllChildren(bottomPane);
            add(bottomPane, BorderLayout.SOUTH);
		}
		
		/** Updates the name of the panel.
		 * 
		 * Usually, this happens during playing an movie, 
		 * when the timestamp changes.
		 */
		public void updateName() {
			if(baseName != "") {
				ObserverMetaData obsMetaData = (ObserverMetaData) view.getAdapter(MetaDataView.class).getMetaData();
				
				String name = baseName;
				
				// get date and convert it to formatted string
				SimpleDateFormat dateFormat;
                try {
                    dateFormat = new SimpleDateFormat(
                            Settings.getSingletonInstance().getProperty("default.date.format"));
                } catch (IllegalArgumentException e1) {
                    dateFormat = new SimpleDateFormat("yyyy/MM/dd");
                } catch (NullPointerException e2) {
                    dateFormat = new SimpleDateFormat("yyyy/MM/dd");
                }
				
				Date date = new GregorianCalendar(
						obsMetaData.getDateTime().getField(Calendar.YEAR), 
						obsMetaData.getDateTime().getField(Calendar.MONTH), 
						obsMetaData.getDateTime().getField(Calendar.DAY_OF_MONTH)).getTime();
	            
	            name += " " + dateFormat.format(date); 
				
	            // add time to name
				name += " " + toTwoLetterString(obsMetaData.getDateTime().getField(Calendar.HOUR_OF_DAY));
				name += ":" + toTwoLetterString(obsMetaData.getDateTime().getField(Calendar.MINUTE));
				name += ":" + toTwoLetterString(obsMetaData.getDateTime().getField(Calendar.SECOND));
				
				// display name
				lblTitle.setText(name);
			}
		}
		
		/** Sets, whether this layer should be highlighted.
		 * 
		 * This happens by changing its border.
		 * 
		 * @param highlighted True, if panel should be highlighted, falso otherwise
		 */
		public void setHighlighted(boolean highlighted) {
			if(highlighted) {
				setBorder(highlightedBorder);
			} else {
				setBorder(normalBorder);
			}
		}
		
		/** Sets, whether the layer is visible or not.
		 * 
		 * This changes the color of the title.
		 * 
		 * @param layerVisible True, if layer is visible, false if not
		 */
		public void setLayerVisible(boolean layerVisible) {
			if(layerVisible) {
				lblTitle.setForeground(Color.BLACK);
			} else {
				lblTitle.setForeground(Color.DARK_GRAY);
			}
		}
		
		/** Adds itself recursively as a mouse listener to its children.
		 * 
		 * After finishing this function, this panel will receive mouse
		 * event from every clickable child.
		 * 
		 * @param parent component to add mouse listener
		 */
		private void addMouseListenerToAllChildren(JComponent parent) {
			for(Component component : parent.getComponents()) {
				if(component.getMouseListeners().length > 0) {
					component.addMouseListener(this);
				}
				if(component instanceof JComponent) {
					addMouseListenerToAllChildren((JComponent) component);
				}
			}
		}
		
		/** Converts an integer to a string, inserting a leading zero if necessary.
		 * 
		 * @param input Integer value to convert
		 * @return String containing at least to letters.
		 */
		private String toTwoLetterString(int input) {
			if(input < 10) {
				return "0" + input;
			}
			return Integer.toString(input);
		}
		
		/** Creates a button for this panel,
		 * 
		 * Sets all common properties.
		 * 
		 * @param icon Icon to use
		 * @param tooltip Tooltip to show
		 * @return Button with all desired properties
		 */
		private JButton createButton(Icon icon, String tooltip) {
			JButton newButton = new JButton(icon);
			newButton.setBorder(BorderFactory.createEtchedBorder());
			newButton.setPreferredSize(new Dimension(icon.getIconWidth()+5,23));
			newButton.setToolTipText(tooltip);
			newButton.addActionListener(this);
			return newButton;
		}

		/**
		 * {@inheritDoc}
		 */
		public void actionPerformed(ActionEvent e) {
			
		    if (e.getSource() instanceof JButton || e.getSource() instanceof JMenuItem)
		        ((ImageSelectorPanel) getParent()).setActiveView(view);
		    
		    if (e.getSource() == cmdDelete) {
		        layeredView.removeLayer(findLastViewBeforeLayeredView(view));
		    } else if (e.getSource() == cmdLayerFunctions) {
		        int x = cmdLayerFunctions.getSize().width - popupLayerFunctions.getSize().width;
		        int y = cmdLayerFunctions.getSize().height;
		        popupLayerFunctions.show(cmdLayerFunctions, x, y);
		        
		    } else if (e.getSource() == miDownload) {
		    	Thread downloadThread = new Thread(new Runnable() {
					public void run() {
						downloadFromJPIP();
					}
				}, "DownloadFromJPIPThread");
		    	downloadThread.start();
		    } else if (e.getSource() == miInfo) {
		        showInfo();
		    } else if (e.getSource() == miVisibility) {
		        
		        if(layeredView.isVisible(findLastViewBeforeLayeredView(view))) {
		            miVisibility.setIcon(IconBank.getIcon(JHVIcon.HIDDEN));
		            miVisibility.setText("Show layer");
		        } else  {
		            miVisibility.setIcon(IconBank.getIcon(JHVIcon.VISIBLE));
		            miVisibility.setText("Hide layer");
                }
                layeredView.toggleVisibility(findLastViewBeforeLayeredView(view));
		    }
		}
		
		/**
		 * {@inheritDoc}
		 */
		public void mouseClicked(MouseEvent e) {
			
			if(e.getSource() == lblTitle) {	
				if(bottomPane.isVisible()) {
					bottomPane.setVisible(false);
					lblTitle.setIcon(IconBank.getIcon(JHVIcon.RIGHT2));
					return;
				} else {
					bottomPane.setVisible(true);
					lblTitle.setIcon(IconBank.getIcon(JHVIcon.DOWN2));
				}
			}
		}

		/**
		 * {@inheritDoc}
		 */
		public void mouseEntered(MouseEvent e) {}

		/**
		 * {@inheritDoc}
		 */
		public void mouseExited(MouseEvent e) {}

		/**
		 * {@inheritDoc}
		 */
		public void mousePressed(MouseEvent e) {
			((ImageSelectorPanel) getParent()).setActiveView(view);
		}

		/**
		 * {@inheritDoc}
		 */
		public void mouseReleased(MouseEvent e) {}
		
		/** Downloads the complete image from the JPIP server.
		 * 
		 * Changes the source of the ImageInfoView afterwards, since
		 * a local file is always faster.
		 */
		private void downloadFromJPIP() {
			
			 FileDownloader fileDownloader = new FileDownloader();
	            URI source = view.getAdapter(ImageInfoView.class).getDownloadURI();

	            // the http server to download the file from is unknown
	            if (view.getAdapter(ImageInfoView.class).getDownloadURI().equals(
	                    view.getAdapter(ImageInfoView.class).getUri())
	                    && !view.getAdapter(ImageInfoView.class).getDownloadURI()
	                            .toString().contains("delphi.nascom.nasa.gov")) {

	                String inputValue = JOptionPane
	                        .showInputDialog(
	                                "To download this file, please specify a concurrent HTTP server address to the JPIP server: ",
	                                view.getAdapter(ImageInfoView.class).getUri());
	               if(inputValue != null) {
	                try {
	                    source = new URI(inputValue);
	                } catch (URISyntaxException e) {
	                }
	               }
	            }

	            File downloadDestination = fileDownloader
	                    .getDefaultDownloadLocation(source);
	            JHVJP2View mainView = view.getAdapter(JHVJP2View.class);
	            JHVJP2View overviewView = (JHVJP2View) ImageViewerGui
	                    .getSingletonInstance().getOverviewView().getAdapter(
	                            SynchronizeView.class).getCorrespondingView(
	                            mainView);
	            try {

	                if (!fileDownloader.get(source, downloadDestination,
	                        "Downloading "
	                                + view.getAdapter(ImageInfoView.class)
	                                        .getName())) {
	                    return;
	                }

	            } catch (IOException e) {
	                e.printStackTrace();
	                return;
	            }

	            try {

	                JP2Image localImage = new JP2Image(downloadDestination.toURI());
	                mainView.setJP2Image(localImage);
	                overviewView.setJP2Image(localImage);

	            } catch (IOException e) {
	                e.printStackTrace();
	            } catch (JHV_KduException e) {
	                e.printStackTrace();
	            }

	            if (view.getAdapter(ImageInfoView.class).getUri().getScheme()
	                    .equalsIgnoreCase("file")) {
	                miDownload.setVisible(false);
	                this.revalidate();
	            }
		}
		
		/** 
		 * Shows a dialog displaying the meta data of the layer.
		 * @see org.helioviewer.jhv.gui.dialogs.MetaDataDialog
		 */
		public void showInfo() {
			MetaDataDialog dialog = new MetaDataDialog();			
			dialog.setMetaData(view.getAdapter(MetaDataView.class));
			dialog.showDialog();
		}
		
		/** 
		 * Destroys the panel.
		 * Calls {@link #removeAllChildrenFromViewListenerDistributor(JComponent)}.
		 */
		public void abolish() {
			removeAllChildrenFromViewListenerDistributor(bottomPane);
		}
		
		/** Removes all components recursively from the ViewListenerDistributor.
		 *  
		 * @param parent component to remove from ViewListenerDistributor
		 * @see org.helioviewer.jhv.gui.ViewListenerDistributor
		 */
		private void removeAllChildrenFromViewListenerDistributor(JComponent parent) {
			ViewListenerDistributor distributor = ViewListenerDistributor.getSingletonInstance();
			
			for(Component component : parent.getComponents()) {
				if(component instanceof ViewListener) {
					((ViewListener) component).viewChanged(view, 
							new ChangeEvent(new LayerChangedReason(view, LayerChangeType.LAYER_REMOVED, view)));
					distributor.removeViewListener((ViewListener) component);
				}
				if(component instanceof JComponent) {
					removeAllChildrenFromViewListenerDistributor((JComponent) component);
				}
			}
		}
	}

	   /**
     * returns the number of Layers in this ImageSelectorPanel
     * 
     * @returnhe number of Layers in this ImageSelectorPanel
     */
    public int getNumLayers() {
        return associatedPanels.size();
    }
}