package org.helioviewer.jhv.gui.components;

import java.awt.Color;
import java.awt.Dimension;

import java.awt.event.ComponentEvent;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.util.AbstractList;
import java.util.LinkedList;
import java.util.Timer;
import java.util.TimerTask;

import org.helioviewer.jhv.gui.IconBank;
import org.helioviewer.jhv.gui.IconBank.JHVIcon;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.changeevent.LayerChangedReason;
import org.helioviewer.viewmodel.renderer.screen.ScreenRenderGraphics;
import org.helioviewer.viewmodel.renderer.screen.ScreenRenderer;
import org.helioviewer.viewmodel.view.ComponentView;
import org.helioviewer.viewmodel.view.LayeredView;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.ViewHelper;

/**
 * This class represents an image component that is used to display the image of
 * all images.
 * 
 * @author caplins
 * @author Alen Agheksanterian
 * @author Benjamin Wamsler
 * @author Stephan Pagel
 * 
 */
public class MainImagePanel extends BasicImagePanel {

	// ///////////////////////////////////////////////////////////////////////////
	// Definitions
	// ///////////////////////////////////////////////////////////////////////////

	// default serialVersionUID
	private static final long serialVersionUID = 1L;

	private NoImagePostRenderer noImagePostRenderer = new NoImagePostRenderer();
	private boolean noImagePostRendererSet = false;
	
	private LoadingPostRenderer loadingPostRenderer = new LoadingPostRenderer();
	private int loadingTasks = 0;
	
	private AbstractList<MouseMotionListener> mouseMotionListeners = new LinkedList<MouseMotionListener>();

	// ///////////////////////////////////////////////////////////////////////////
	// Methods
	// ///////////////////////////////////////////////////////////////////////////

	/** 
	 * The public constructor
	 * */
	public MainImagePanel() {
		
		// call constructor of super class
		super();

		// add post render that no image is loaded
		noImagePostRenderer.setContainerSize(getWidth(), getHeight());
		addPostRenderer(noImagePostRenderer);
		noImagePostRendererSet = true;
		
		loadingPostRenderer.setContainerSize(getWidth(), getHeight());
	}
	
	/** Shows the image loading animation.
	 * 
	 * Manages a counter, so that the animation appears on the
	 * first loading process and disappears on the last.
	 * 
	 * @param isLoading true to start animation, false to stop
	 */
	public void setLoading(boolean isLoading) {
		if(isLoading) {
			if(loadingTasks <= 0) {
				if(noImagePostRendererSet) {
					removePostRenderer(noImagePostRenderer);
                    noImagePostRendererSet = false;
				}
				addPostRenderer(loadingPostRenderer);
				loadingPostRenderer.startAnimation();
			}
			loadingTasks++;
		} else if(loadingTasks > 0){
			loadingTasks--;
			if(loadingTasks == 0) {
				removePostRenderer(loadingPostRenderer);
				loadingPostRenderer.stopAnimation();
			}
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setView(ComponentView newView) {
	    
	    if (renderedImageComponent != null)
	        for (MouseMotionListener l: mouseMotionListeners)
	            renderedImageComponent.removeMouseMotionListener(l);
	    
	    super.setView(newView);
	    
	    if (renderedImageComponent != null)
	        for (MouseMotionListener l: mouseMotionListeners)
	            renderedImageComponent.addMouseMotionListener(l);
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public synchronized void addMouseMotionListener(MouseMotionListener l) {
	    if (l != null)
	        mouseMotionListeners.add(l);
	}
	
	/**
     * {@inheritDoc}
     */
	@Override
	public synchronized void removeMouseMotionListener(MouseMotionListener l) {
	    if (l != null)
	        mouseMotionListeners.remove(l);
	}
	
    /**
     * {@inheritDoc}
     * 
     * Adds and removes a post renderer which displays the no image loaded sign
     * when there is no image available.
     */
	@Override
	public void viewChanged(View sender, ChangeEvent aEvent) {
	    
        // checks if at least one layer is available - if not the post renderer
        // for displaying that no image was selected will be added
	    if (aEvent.reasonOccurred(LayerChangedReason.class)) {	        
	        LayeredView layeredView = ViewHelper.getViewAdapter(sender, LayeredView.class);
	        
	        if (layeredView != null) {
	            if (layeredView.getNumLayers() > 0 || loadingTasks > 0) {
	                //remove
	                if (noImagePostRendererSet) {
	                    removePostRenderer(noImagePostRenderer);
	                    noImagePostRendererSet = false;
	                    
	                    repaint();
	                }
	            } else {
	                //add
	                if (!noImagePostRendererSet) {
	                    addPostRenderer(noImagePostRenderer);
	                    noImagePostRendererSet = true;
	                    
	                    repaint();
	                }
	            }	                
	        }
	    }
	    
	    // call this method in super class
	    super.viewChanged(sender, aEvent);
	}
	
	/**
	 * {@inheritDoc}
	 * 
	 * Centers the no image loaded image when component was resized.
	 */
	@Override
	public void componentResized(ComponentEvent e) {
	    
	    noImagePostRenderer.setContainerSize(getWidth(), getHeight());
	    loadingPostRenderer.setContainerSize(getWidth(), getHeight());
	    
	    super.componentResized(e);
	}
	
    /**
     * A post renderer which displays an image which shows that no image (layer)
     * is loaded.
     * 
     * @author Stephan Pagel
     * */
	private class NoImagePostRenderer implements ScreenRenderer {

	    private BufferedImage image = null;
	    private Dimension size = new Dimension(0, 0);
	    
	    /**
         * Default constructor. 
         */
        public NoImagePostRenderer() {
            try {
                image = IconBank.getImage(JHVIcon.NOIMAGE);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
	    
	    /**
	     * Sets the size of the available image area (viewport).
	     * 
	     * @param width width of the available space of the image.
	     * @param height height of the available space of the image.
	     */
	    public void setContainerSize(int width, int height) {
	        size = new Dimension(width, height);
	    }
	    
	    /**
	     * {@inheritDoc}
	     * 
	     * Draws the no image loaded image.
	     */
        public void render(ScreenRenderGraphics g) {
            if (image != null) {
                g.drawImage(image, (size.width - image.getWidth()) / 2, 
                        (size.height - image.getHeight()) / 2);
            }
        }
	}
	
	/**
	 * A post renderer which indicates that something is being loaded right now.
	 * 
	 * @author Markus Langenberg
	 */
	private class LoadingPostRenderer implements ScreenRenderer {
		
		// track
		private static final int offsetX = -4;
		private static final int offsetY = -27;
		private static final int radiusTrack = 97;
		private final float[] sinPositions;
		
		// pearls
		private static final int numPearlPositions = 32;
		private static final int numPearls = 12;
		private static final int radiusPearl = 6;
		private final Color[] pearlColors;
		private int currentPearlPos = 0;

		// image
		private BufferedImage image;
		
		// container
	    private Dimension size = new Dimension(0, 0);
	    
	    // timer
	    Timer timer;
	    
	    /**
         * Default constructor. 
         */
        public LoadingPostRenderer() {
            sinPositions = new float[numPearlPositions];
            for(int i=0; i<numPearlPositions; i++) {
            	sinPositions[i] = (float) Math.sin(Math.PI * 2 * i / numPearlPositions);
            }
            
        	pearlColors = new Color[numPearls];
            for(int i=0; i<numPearls; i++) {
            	int alpha = 192 - (int) (192 * ((float) i / numPearls));
            	pearlColors[i] = new Color(192, 192, 192, alpha);
            }
            
            try {
                image = IconBank.getImage(JHVIcon.LOADING);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
	    
	    /**
	     * Sets the size of the available image area (viewport).
	     * 
	     * @param width width of the available space of the image.
	     * @param height height of the available space of the image.
	     */
	    public void setContainerSize(int width, int height) {
	        size = new Dimension(width, height);
	    }
	    
	    public void startAnimation() {
	    	currentPearlPos = numPearlPositions / 2;
	    	
	    	if(timer == null) {
    	    	timer = new Timer("Loading Animation", true);
    	    	timer.schedule(new TimerTask() {
    				public void run() {
    					currentPearlPos++;
    					repaint();
    				}
    			}, 0, 100);
	    	}
	    }
	    
	    public void stopAnimation() {
	        if(timer != null) {
	            timer.cancel();
	            timer = null;
	        }
	    }
	    
	    /**
	     * {@inheritDoc}
	     * 
	     * Draws the loading image and its animation.
	     */
        public void render(ScreenRenderGraphics g) {
            if (image != null) {
                g.drawImage(image, (size.width - image.getWidth()) / 2, 
                        (size.height - image.getHeight()) / 2);
            }
            
            int centerX = size.width  / 2 - radiusPearl + offsetX;
            int centerY = size.height / 2 - radiusPearl + offsetY;
            
            for(int i=0; i<numPearls; i++) {
            	g.setColor(pearlColors[i]);
            	g.fillOval(centerX + (int)(radiusTrack * sinPositions[(i - currentPearlPos) & (numPearlPositions - 1)]), 
			               centerY + (int)(radiusTrack * sinPositions[(i - currentPearlPos + (numPearlPositions >> 2)) & (numPearlPositions - 1)]),
			               radiusPearl * 2, radiusPearl * 2);
            }
        }
	}
}