package org.helioviewer.jhv.gui.components;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSeparator;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.border.EmptyBorder;

import org.helioviewer.jhv.JHVGlobals;
import org.helioviewer.jhv.Settings;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.gui.components.calendar.JHVCalendarDatePicker;
import org.helioviewer.jhv.gui.components.calendar.JHVCalendarEvent;
import org.helioviewer.jhv.gui.components.calendar.JHVCalendarListener;
import org.helioviewer.jhv.io.APIRequestManager;

/**
 * This panel provides the main interface to get images or image series from the
 * Helioviewer server.
 * 
 * @author Stephan Pagel
 * */
public class ObservationPanel extends JPanel implements ActionListener {

    // //////////////////////////////////////////////////////////////////////////////
    // Definitions
    // //////////////////////////////////////////////////////////////////////////////
    
    private static final long serialVersionUID = 1L;
    
    private TimeSelectionPanel timeSelectionPanel = new TimeSelectionPanel();
    private CadencePanel cadencePanel = new CadencePanel();
    private InstrumentsPanel instrumentsPanel = new InstrumentsPanel();
    private ButtonPanel buttonPanel = new ButtonPanel();
    
    // //////////////////////////////////////////////////////////////////////////////
    // Methods
    // //////////////////////////////////////////////////////////////////////////////
    
    /**
     * Default constructor.
     * */
    public ObservationPanel() {        
        // set up the visual components (GUI)
        initVisualComponents();
    }
    
    /**
     * Sets up the visual sub components and the component itself.
     * */
    private void initVisualComponents() {

        // set basic layout
        setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
        setFocusable(true);
        
        // add action listener to buttons
        buttonPanel.btnImage.addActionListener(this);
        buttonPanel.btnImages.addActionListener(this);
        
        // add components to panel
        add(timeSelectionPanel);
        add(cadencePanel);
        add(new JSeparator());
        add(instrumentsPanel);
        add(buttonPanel);
    }
    
    /**
     * ActionListener for components of this class.
     * */
    public void actionPerformed(ActionEvent arg0) {
        
        if (arg0.getSource() == buttonPanel.btnImage) {
            
            // load image
            loadImage();
            
        } else if (arg0.getSource() == buttonPanel.btnImages) {
            
            // check if start date is before end date -> if not show message
            if (!timeSelectionPanel.isStartDateBeforeEndDate()) {
                JOptionPane.showMessageDialog(null, "End date is before start date!", "", JOptionPane.ERROR_MESSAGE);
                return;
            }
            ((collapsibleObservation) this.getParent()).collapse();
            // load movie
            loadMovie();        
        }
    }
    
    /**
     * Loads an image from the Helioviewer server and adds a new layer to the
     * GUI which represents the image.
     * */
    private void loadImage() {
        
        // show loading animation
        ImageViewerGui.getSingletonInstance().getMainImagePanel().setLoading(true);
        
        // download and open the requested image in a separated thread and hide
        // loading animation when finished
        Thread thread = new Thread(new Runnable() {
            
            public void run() {
                
                try {
                	APIRequestManager.requestAndOpenRemoteFile( 
                            timeSelectionPanel.getStartDate(), timeSelectionPanel.getStartTime(), 
                            instrumentsPanel.getObservation(), instrumentsPanel.getInstrument(),
                            instrumentsPanel.getDetector(), instrumentsPanel.getMeasurement());
                } catch (IOException e) {
                    JHVGlobals.err("An error occured while opening the remote file!", e.getMessage(), false);
                } finally {
                    ImageViewerGui.getSingletonInstance().getMainImagePanel().setLoading(false);   
                }
            }
        }, "LoadNewImage");
        
        thread.start();
    }
    
    /**
     * Loads an image series from the Helioviewer server and adds a new layer to
     * the GUI which represents the image series.
     * */
    private void loadMovie() {
        
        // show loading animation
        ImageViewerGui.getSingletonInstance().getMainImagePanel().setLoading(true);
        
        // download and open the requested movie in a separated thread and hide
        // loading animation when finished
        Thread thread = new Thread(new Runnable() {
            
            public void run() {
                try {
                	APIRequestManager.requestAndOpenRemoteFile(false, Integer.toString(cadencePanel.getCadence()), 
                            timeSelectionPanel.getStartDate(), timeSelectionPanel.getStartTime(), 
                            timeSelectionPanel.getEndDate(), timeSelectionPanel.getEndTime(), 
                            instrumentsPanel.getObservation(), instrumentsPanel.getInstrument(),
                            instrumentsPanel.getDetector(), instrumentsPanel.getMeasurement());
                } catch (IOException e) {
                    JHVGlobals.err("An error occured while opening the remote file!", e.getMessage(), false);
                } finally {
                    ImageViewerGui.getSingletonInstance().getMainImagePanel().setLoading(false);
                }
            }
        }, "LoadNewMovie");
        
        thread.start();
    }
    
    /**
     * Returns the selected start date.
     * 
     * @return selected start date.
     */
    public Date getStartDate() {
        return timeSelectionPanel.getStartDate();
    }

    /**
     * Returns the selected end date.
     * 
     * @return selected end date.
     */
    public Date getEndDate() {
        return timeSelectionPanel.getEndDate();
    }

    /**
     * Returns the selected start time.
     * 
     * @return selected start time.
     * */
    public String getStartTime() {
        return timeSelectionPanel.getStartTime();
    }
    
    /**
     * Returns the selected end time.
     * 
     * @return seleted end time.
     */
    public String getEndTime() {
        return timeSelectionPanel.getEndTime();
    }
    
    /**
     * Returns the selected cadence.
     * 
     * @return selected cadence.
     */
    public String getCadence() {
        return Integer.toString(cadencePanel.getCadence());
    }
    
    /**
     * Returns the selected observation.
     * 
     * @return selected observation.
     */
    public String getObservation() {
        return instrumentsPanel.getObservation();
    }
    
    /**
     * Returns the selected instrument.
     * 
     * @return selected instrument.
     * */
    public String getInstrument() {
        return instrumentsPanel.getInstrument();
    }
    
    /**
     * Returns the selected detector.
     * 
     * @return selected detector.
     * */
    public String getDetector() {
        return instrumentsPanel.getDetector();
    }
    
    /**
     * Returns the selected measurement.
     * 
     * @return selected measurement.
     * */
    public String getMeasurement() {
        return instrumentsPanel.getMeasurement();
    }
    
    /**
     * Updates the visual behavior of the component.
     */
    public void updateComponent() {
        timeSelectionPanel.updateDateFormat();
    }
    
    /**
     * The panel bundles the components to select the start and end time.
     * 
     * @author Stephan Pagel
     * */
    private class TimeSelectionPanel extends JPanel implements FocusListener, KeyListener, JHVCalendarListener {

        // //////////////////////////////////////////////////////////////////////////
        // Definitions
        // //////////////////////////////////////////////////////////////////////////

        private static final long serialVersionUID = 1L;
      
        private JLabel labelStartDate = new JLabel("Start Date");
        private JLabel labelStartTime = new JLabel("Start Time");
        private JLabel labelEndDate = new JLabel("End Date");
        private JLabel labelEndTime = new JLabel("End Time");
        
        private JTextField textStartTime;
        private JTextField textEndTime;        
        private JHVCalendarDatePicker calendarStartDate;
        private JHVCalendarDatePicker calendarEndDate;
        
     
        // //////////////////////////////////////////////////////////////////////////
        // Methods
        // //////////////////////////////////////////////////////////////////////////
        
        /**
         * Default constructor.
         * */
        public TimeSelectionPanel() {
            // set up the visual components (GUI)
            initVisualComponents();
        }
        
        /**
         * Sets up the visual sub components and the component itself.
         * */
        private void initVisualComponents() {

            // set basic layout
            setLayout(new GridLayout(4, 2));

            // create start date picker            
            calendarStartDate = new JHVCalendarDatePicker();
            calendarStartDate.setDateFormat(Settings.getSingletonInstance().getProperty("default.date.format"));
            calendarStartDate.setDate(new GregorianCalendar(2003, 9, 5).getTime());
            calendarStartDate.addKeyListener(this);
            calendarStartDate.addJHVCalendarListener(this);
            
            // create end date picker              
            calendarEndDate = new JHVCalendarDatePicker();
            calendarEndDate.setDateFormat(Settings.getSingletonInstance().getProperty("default.date.format"));
            calendarEndDate.setDate(new GregorianCalendar(2003, 9, 6).getTime());
            calendarEndDate.addKeyListener(this);
            calendarEndDate.addJHVCalendarListener(this);
            
            // set date format to components
            updateDateFormat();
            
            // create start time field
            textStartTime = new JTextField("00:00:00");             
            textStartTime.addFocusListener(this);
            textStartTime.addKeyListener(this);
          
            // create end time field
            textEndTime = new JTextField("00:00:00");                
            textEndTime.addFocusListener(this);
            textEndTime.addKeyListener(this);

            // add components to panel
            add(labelStartDate);
            add(labelStartTime);
            add(calendarStartDate);
            add(textStartTime);
            add(labelEndDate);
            add(labelEndTime);
            add(calendarEndDate);
            add(textEndTime);
        }
        
        /**
         * Updates the date format to the calendar components.
         */
        public void updateDateFormat() {
           
            String pattern = Settings.getSingletonInstance().getProperty("default.date.format");

            calendarStartDate.setDateFormat(pattern);
            calendarEndDate.setDateFormat(pattern);
            
            calendarStartDate.setDate(calendarStartDate.getDate());
            calendarEndDate.setDate(calendarEndDate.getDate());
        }
        
        /**
         * JHV calendar listener which notices when the user has chosen a date
         * by using the calendar component.
         */
        public void actionPerformed(JHVCalendarEvent e) {
            
            if (e.getSource() == calendarStartDate && !isStartDateBeforeEndDate()) {
                
                Calendar calendar = new GregorianCalendar();
                calendar.setTime(calendarStartDate.getDate());
                calendar.add(Calendar.DATE, 1);                
                calendarEndDate.setDate(calendar.getTime());
            }
            
            if (e.getSource() == calendarEndDate && !isStartDateBeforeEndDate()) {
                
                Calendar calendar = new GregorianCalendar();
                calendar.setTime(calendarEndDate.getDate());
                calendar.add(Calendar.DATE, -1);
                calendarStartDate.setDate(calendar.getTime());
            }
        }
        
        /**
         * Focus gained listener for the components of the panel.
         * */
        public void focusGained(FocusEvent arg0) {}
        
        /**
         * Focus lost listener for the components of the panel. It checks if
         * inputs are made correctly.
         * */
        public void focusLost(FocusEvent arg0) {
            // check when time text field lost the focus if a valid time string was entered.
            // If an invalid time string was entered, replace by default.
            if (arg0.getSource() == textStartTime || arg0.getSource() == textEndTime) {
                if (!APIRequestManager.timeIsValid(((JTextField)arg0.getSource()).getText()))
                    ((JTextField)arg0.getSource()).setText("00:00:00");
            }
        }
        
        /**
         * Checks if the selected start date is before selected end date. The
         * methods checks the entered times when the dates are equal. If the
         * start time is greater or equal than the end time the method will
         * return false.
         * 
         * @return boolean value if selected start date is before selected end
         *         date.
         */
        public boolean isStartDateBeforeEndDate() {
            
            if (calendarEndDate.getDate().before(calendarStartDate.getDate())) {

                // end date is before start date
                return false;
                    
            } else if (calendarStartDate.getDate().equals(calendarEndDate.getDate()) 
                    && toSeconds(textStartTime.getText()) >= toSeconds(textEndTime.getText())) {
                
                // dates are the same but end time is before start time
                return false;
            }
            
            return true;
        }
        
        /**
         * Transforms a valid time entry (separated by ":") into the
         * corresponding number of seconds.
         * 
         * @param time
         *            valid time string.
         * @return corresponding number of seconds.
         */
        private int toSeconds(String time) {
            
            int result = 0;
            String[] parts = time.split(":");
            
            result += new Integer(parts[0]) * 3600; // hours
            result += new Integer(parts[1]) * 60;   // minutes
            result += new Integer(parts[2]);        // seconds
            
            return result;
        }
        
        /**
         * Returns the selected start date.
         * 
         * @return selected start date.
         */
        public Date getStartDate() {
            return calendarStartDate.getDate();
        }
      
        /**
         * Returns the selected end date.
         * 
         * @return selected end date.
         */
        public Date getEndDate() {
            return calendarEndDate.getDate();
        }
      
        /**
         * Returns the selected start time.
         * 
         * @return selected start time.
         * */
        public String getStartTime() {
            return textStartTime.getText();
        }
      
        /**
         * Returns the selected end time.
         * 
         * @return selected end time.
         */
        public String getEndTime() {
            return textEndTime.getText();
        }

        /**
         * Key pressed listener for components of this class. Is needed to
         * handle short cuts correctly and no problems occur when editing
         * entries.
         */
		public void keyPressed(KeyEvent e) {
			if(!e.isActionKey()
					&& e.getKeyCode() != KeyEvent.VK_BACK_SPACE 
					&& e.getKeyCode() != KeyEvent.VK_DELETE) {
				e.consume();
			}
		}

		/**
		 * Key released listener for components of this class.
		 * */
		public void keyReleased(KeyEvent e) {}

		/**
         * Key typed listener for components of this class.
         * */
		public void keyTyped(KeyEvent e) {}
    }
    
    /**
     * The panel bundles the components to select the cadence.
     * 
     * @author Stephan Pagel
     * */
    private class CadencePanel extends JPanel {
     
        // //////////////////////////////////////////////////////////////////////////
        // Definitions
        // //////////////////////////////////////////////////////////////////////////

        private static final long serialVersionUID = 1L;
        
        private final String[] timeStepUnitStrings = { "sec", "min", "hours", "days" };
        
        private JLabel labelTimeStep = new JLabel("Time Step");
        private JSpinner spinnerCadence = new JSpinner();
        private JComboBox comboUnit = new JComboBox(timeStepUnitStrings);
        
        // //////////////////////////////////////////////////////////////////////////
        // Methods
        // //////////////////////////////////////////////////////////////////////////  
        
        /**
         * Default constructor.
         * */
        public CadencePanel() {
            // set up the visual components (GUI)
            initVisualComponents();
        }
        
        /**
         * Sets up the visual sub components and the component itself.
         * */
        private void initVisualComponents() {
            
            // set basic layout
            setLayout(new GridLayout(1, 2));
            setBorder(new EmptyBorder(3, 0, 0, 0));
            
            spinnerCadence.setPreferredSize(new Dimension(50, 25));
            spinnerCadence.setModel(new SpinnerNumberModel(1, 1, 1000000, 1));
            
            comboUnit.setSelectedIndex(2);
            
            JPanel panel = new JPanel();
            panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));            
            panel.add(spinnerCadence);
            panel.add(comboUnit);
            
            // add components to panel
            add(labelTimeStep);
            add(panel);
        }
        
        /**
         * Returns the number of seconds of the selected cadence.
         * 
         * @return number of seconds of the selected cadence.
         * */
        public int getCadence() {
            
            int value = ((SpinnerNumberModel)spinnerCadence.getModel()).getNumber().intValue();
            
            switch (comboUnit.getSelectedIndex()) {
            case 1: // min
                value *= 60;
                break;
            case 2: // hour
                value *= 3600;
                break;
            case 3: // day
                value *= 86400;
                break;
            }
            
            return value;
        }
    }
    
    /**
     * The panel bundles the components to select the instrument etc.
     * 
     * @author Stephan Pagel
     * */
    private class InstrumentsPanel extends JPanel implements ActionListener {

        // //////////////////////////////////////////////////////////////////////////
        // Definitions
        // //////////////////////////////////////////////////////////////////////////

        private static final long serialVersionUID = 1L;
        
        private final char angstrom = '\u212B';
        
        private final String[] observationStrings = { "SOHO" };
        private final String[] instrumentsStrings = { "EIT", "MDI", "LASCO" };            
        private final String[] detectorMeasurementStrings_EIT = { "171 " + angstrom, "195 " + angstrom, "284 " + angstrom, "304 " + angstrom };
        private final String[] detectorMeasurementStrings_MDI = { "magnetogram", "continuum" };
        private final String[] detectorMeasurementStrings_LASCO = { "C2", "C3"};
        
        private JLabel labelObservation = new JLabel("Observatory");
        private JLabel labelInstrument = new JLabel("Instrument");
        private JLabel labelDetectorMeasurement = new JLabel("Detector/Measurement");
      
        private JComboBox comboObservation = new JComboBox(observationStrings);
        private JComboBox comboInstrument = new JComboBox(instrumentsStrings);
        private JComboBox comboDetectorMeasurement = new JComboBox(detectorMeasurementStrings_EIT);
        
        // //////////////////////////////////////////////////////////////////////////
        // Methods
        // //////////////////////////////////////////////////////////////////////////  
        
        /**
         * Default constructor.
         * */
        public InstrumentsPanel() {
            // set up the visual components (GUI)
            initVisualComponents();
        }
        
        /**
         * Sets up the visual sub components and the component itself.
         * */
        private void initVisualComponents() {
            setLayout(new GridLayout(3, 2));
            
            comboInstrument.addActionListener(this);
            comboDetectorMeasurement.setSelectedIndex(1);
            
            add(labelObservation);
            add(comboObservation);
            add(labelInstrument);
            add(comboInstrument);
            add(labelDetectorMeasurement);
            add(comboDetectorMeasurement);
        }
        
        /**
         * ActionListener for components of this class.
         * */
        public void actionPerformed(ActionEvent e) {

            Object source = e.getSource();

            if (source == comboInstrument) {
                instrumentChange();
            }
        }

        /**
         * Sets the right values to the detector and measurement combobox
         * depending on the selected instrument.
         * */
        private void instrumentChange() {
            comboDetectorMeasurement.removeAllItems();

            switch (comboInstrument.getSelectedIndex()) {
            case 0:
                fillComboBox(detectorMeasurementStrings_EIT, comboDetectorMeasurement);
                comboDetectorMeasurement.setSelectedIndex(1);
                break;
            case 1:
                fillComboBox(detectorMeasurementStrings_MDI, comboDetectorMeasurement);
                break;
            case 2:
                fillComboBox(detectorMeasurementStrings_LASCO, comboDetectorMeasurement);
                break;
            }
        }

        /**
         * Adds all string entries of an array as items to to given combobox.
         * 
         * @param items
         *            string array which contains the names for the items of the
         *            combobox.
         * @param container
         *            combobox where to add the items.
         */
        private void fillComboBox(String[] items, JComboBox container) {
            for (int i = 0; i < items.length; i++) {
                container.addItem(items[i]);
            }
        }
        
        /**
         * Returns the selected observation.
         * 
         * @return selected observation.
         * */
        public String getObservation() {
            return comboObservation.getSelectedItem().toString();
        }

        /**
         * Returns the selected instrument.
         * 
         * @return selected instrument.
         * */
        public String getInstrument() {
            return comboInstrument.getSelectedItem().toString();
        }

        /**
         * Returns the selected detector.
         * 
         * @return selected detector.
         * */
        public String getDetector() {
            switch (comboInstrument.getSelectedIndex()) {
            case 0:
                return comboInstrument.getSelectedItem().toString();
            case 1:
                return comboInstrument.getSelectedItem().toString();
            case 2:
                return comboDetectorMeasurement.getSelectedItem().toString();
            }

            return null;
        }

        /**
         * Returns the selected measurement.
         * 
         * @return selected measurement.
         * */
        public String getMeasurement() {

            switch (comboInstrument.getSelectedIndex()) {
            case 0:
                return comboDetectorMeasurement.getSelectedItem().toString().substring(0, 3);
            case 1:
                return comboDetectorMeasurement.getSelectedItem().toString();
            case 2:
                return "white light";
            }

            return null;
        }
    }
    
    /**
     * The panel bundles the buttons to confirm the entries made in the
     * observation panel.
     * 
     * @author Stephan Pagel
     * */
    private class ButtonPanel extends JPanel {

        // //////////////////////////////////////////////////////////////////////////
        // Definitions
        // //////////////////////////////////////////////////////////////////////////

        private static final long serialVersionUID = 1L;
        
        public JButton btnImage = new JButton("Get 1st Image");
        public JButton btnImages = new JButton("Get Images");
        
        // //////////////////////////////////////////////////////////////////////////
        // Methods
        // //////////////////////////////////////////////////////////////////////////  
        
        /**
         * Default constructor.
         * */
        public ButtonPanel() {
            // set up the visual components (GUI)
            initVisualComponents();
        }
        
        /**
         * Sets up the visual sub components and the component itself.
         * */
        private void initVisualComponents() {
            
            // set basic layout
            setLayout(new FlowLayout(FlowLayout.RIGHT, 0, 2));
            
            // set up the buttons
            int btnWidth = Math.max(btnImage.getPreferredSize().getSize().width, btnImages.getPreferredSize().getSize().width);
            btnImage.setPreferredSize(new Dimension(btnWidth, 25));
            btnImages.setPreferredSize(new Dimension(btnWidth, 25));

            // create spacer between buttons
            JPanel spacer = new JPanel();
            spacer.setPreferredSize(new Dimension(2, 2));
            
            // add components to panel
            //add(btnImage);
            add(spacer);
            add(btnImages);            
        }
    }
}