package org.helioviewer.jhv.gui.components;

import java.awt.Color;
import java.awt.Rectangle;

import org.helioviewer.base.math.Vector2dInt;
import org.helioviewer.jhv.gui.controller.OverviewImagePanelMousePanController;
import org.helioviewer.jhv.gui.interfaces.ImagePanelInputController;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.changeevent.LayerChangedReason;
import org.helioviewer.viewmodel.changeevent.SynchronizedROIChangedReason;
import org.helioviewer.viewmodel.metadata.MetaData;
import org.helioviewer.viewmodel.region.Region;
import org.helioviewer.viewmodel.renderer.screen.ScreenRenderGraphics;
import org.helioviewer.viewmodel.renderer.screen.ScreenRenderer;
import org.helioviewer.viewmodel.view.LayeredView;
import org.helioviewer.viewmodel.view.MetaDataView;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.ViewHelper;
import org.helioviewer.viewmodel.viewportimagesize.ViewportImageSize;

/**
 * This class represents an image component that is used to display the overview
 * of an image.
 * 
 * @author Stephan Pagel
 * 
 */
public class OverviewImagePanel extends BasicImagePanel {

	// ///////////////////////////////////////////////////////////////////////////
	// Definitions
	// ///////////////////////////////////////////////////////////////////////////

	// default serialVersionUID
	private static final long serialVersionUID = 1L;
	
	private boolean drawRect = true;
	
	// buffered data
	private ViewportImageSize viewportImageSize;
	private Vector2dInt offset;
	private MetaData metaData;
	
	private OverviewImagePanelPostRenderer postRenderer = new OverviewImagePanelPostRenderer();
	private OverviewImagePanelMousePanController inputControllerPanning = new OverviewImagePanelMousePanController(); 

	// ///////////////////////////////////////////////////////////////////////////
	// Methods
	// ///////////////////////////////////////////////////////////////////////////

	/** 
	 * Default constructor.
	 * */
	public OverviewImagePanel() {
		
		// call constructor of super class
		super();
		
		// add post renderer
		addPostRenderer(postRenderer);
		
		// add a mouse input controller to the panel
		setInputController(inputControllerPanning);
	}
	
    /**
     * {@inheritDoc}
     * 
     * This method is overridden to protect the internal input controller. Input
     * controller for this class are handled by this class by its own.
     */
	@Override
	public void setInputController(ImagePanelInputController newInputController) {
	    
	    if (newInputController == inputControllerPanning)
	        super.setInputController(newInputController);	    
	}
	
	/** 
	 * {@inheritDoc}
	 * 
	 * Computes the ROI rectangle for the overview image.
	 * */
	@Override
	public void viewChanged(View sender, ChangeEvent aEvent) {

		// Show rendered image centered
		if (renderedImageComponent != null) {
			
			boolean layerChanged = aEvent.reasonOccurred(LayerChangedReason.class);
		    
            if (layerChanged) {                
                drawRect = sender.getAdapter(LayeredView.class).getNumLayers() > 0
                        && sender.getAdapter(LayeredView.class).getNumberOfVisibleLayer() > 0;
            }
            
		    if (drawRect) {
		    	
		    	
		    	// get size of view port of overview image
		    	ViewportImageSize tempViewportImageSize = ViewHelper.calculateViewportImageSize(componentView);
	            
	            if (tempViewportImageSize != null) {
	            	viewportImageSize = tempViewportImageSize;
	            	
                    // compute offset, so image will be displayed centered                
                    offset = computeOffset(viewportImageSize);
                    
                    componentView.setOffset(offset);    
                    
                    inputControllerPanning.setImageArea(new Rectangle(
                            offset.getX(), offset.getY(), viewportImageSize.getWidth(), 
                            viewportImageSize.getHeight()));

                    // get meta data of the image, which includes e.g. total size
                    metaData = sender.getAdapter(MetaDataView.class).getMetaData();
	            }
                    
                SynchronizedROIChangedReason reason = aEvent.getLastChangedReasonByType(SynchronizedROIChangedReason.class);

                // compute rectangle of current ROI
                if (metaData != null && reason != null) {
                    Rectangle roi = calculateRectangle(metaData, viewportImageSize, reason.getRegionToSynchronize());
                    roi.x = Math.abs(roi.x) + offset.getX();
                    roi.y = Math.abs(roi.y) + offset.getY();
                    
                    roi.width--;
                    roi.height--;                        
                    roi.width = roi.width < 1 ? 1 : roi.width;
                    roi.height = roi.height < 1 ? 1 : roi.height;
                    
                    postRenderer.setROIRectangle(roi);
                    inputControllerPanning.setROIArea(roi);
                }
		    } else {
		        // define "dummy" rectangle
		        Rectangle rect = new Rectangle(0, 0);
		        
		        // "remove" region of interest
		        postRenderer.setROIRectangle(rect);
                inputControllerPanning.setROIArea(rect);
                
                // "remove" image region
                inputControllerPanning.setImageArea(rect);
		    }
		}
		
		// repaint the image
		repaint();
	}
	
    /**
     * Computes the offset that image will be displayed centered.
     * 
     * @param viewportImageSize
     *            image size
     * @return offset that image will be displayed centered.
     */
	private Vector2dInt computeOffset(ViewportImageSize viewportImageSize) { 
        int x = (getViewport().getWidth() - viewportImageSize.getWidth()) / 2;
        x = x >= 0 ? x : 0;
        
        int y = (getViewport().getHeight() - viewportImageSize.getHeight()) / 2;
        y = y >= 0 ? y : 0;
        
        return new Vector2dInt(x,y);
	}
	
    /**
     * Computes the size and position of the rectangle which indicates the
     * current ROI.
     * 
     * @param metaData
     *            Meta data of the image
     * @param vieportImageSize
     *            Size of the view port of the overview image
     * @param regionMainImage
     *            The region which is currently shown in the main image
     * @return position and size of the rectangle which indicates the current
     *         ROI.
     * */
    private Rectangle calculateRectangle(MetaData metaData, ViewportImageSize viewportImageSize, Region regionMainImage) {
        
        Rectangle rect = new Rectangle();
        
        rect.x = (int) Math.round(((metaData.getPhysicalUpperLeft().getX() - regionMainImage.getUpperLeftCorner().getX()) / metaData.getPhysicalImageWidth()) * viewportImageSize.getWidth());
        rect.y = (int) Math.round(((metaData.getPhysicalUpperLeft().getY() - regionMainImage.getUpperLeftCorner().getY()) / metaData.getPhysicalImageHeight()) * viewportImageSize.getHeight());
        
        rect.width = (int) Math.round((regionMainImage.getWidth() / metaData.getPhysicalImageWidth()) * viewportImageSize.getWidth());
        rect.height = (int) Math.round((regionMainImage.getHeight() / metaData.getPhysicalImageHeight()) * viewportImageSize.getHeight());
        
        return rect;
    }
	
	/**
	 * Class represents the post renderer for the overview image panel.
	 * Contains the drawing management for the ROI rectangle.
	 * 
	 * @author Stephan Pagel
	 * */
	private class OverviewImagePanelPostRenderer implements ScreenRenderer {
		
		// ///////////////////////////////////////////////////////////////////////
		// Definitions
		// ///////////////////////////////////////////////////////////////////////
		
		private Rectangle rect = new Rectangle();
		private Color color = Color.YELLOW;
		
		// ///////////////////////////////////////////////////////////////////////
		// Methods
		// ///////////////////////////////////////////////////////////////////////
		
        /**
         * Sets the last position and size of the rectangle.
         * 
         * @param rect
         *            size and position of the rectangle.
         */
	    public void setROIRectangle(Rectangle rect) {
	        this.rect = rect;
	    }
		
        /**
         * Draws the rectangle.
         * 
         * @param g
         *            Graphics object where to draw the rectangle.
         */
		public void render(ScreenRenderGraphics g) {
			
			if (rect != null) {
				g.setColor(color);
				g.drawRectangle(rect.x, rect.y, rect.width, rect.height);
			}			
		}		
	}
}
