package org.helioviewer.jhv.gui.components.statusplugins;

import java.awt.Dimension;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.JWindow;

import org.helioviewer.jhv.gui.IconBank;
import org.helioviewer.jhv.gui.ViewListenerDistributor;
import org.helioviewer.jhv.gui.IconBank.JHVIcon;
import org.helioviewer.jhv.gui.interfaces.StatusPanelPlugin;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.ViewHelper;
import org.helioviewer.viewmodel.view.ViewListener;
import org.helioviewer.viewmodel.view.jp2view.JHVJP2View;

/** Status panel for displaying the status of the JPIP connection.
 * 
 * <p>If the status changes, a small popup will appear.
 * 
 * <p>The information of this panel is always shown for the
 * active layer.
 * 
 * <p>This panel is not visible, if the active layer is not
 * an remote JPIP image.
 * 
 * @author Markus Langenberg
 */
public class JPIPStatusPanel extends JLabel implements ViewListener, StatusPanelPlugin {

	private static final long serialVersionUID = 1L;
	private JHVJP2View activeView;
	private boolean lastStatus = true;
	private StatusChangedWindow statusChangedWindow = new StatusChangedWindow();
	
	private static final Icon connectedIcon = IconBank.getIcon(JHVIcon.CONNECTED);
	private static final Icon disconnectedIcon = IconBank.getIcon(JHVIcon.DISCONNECTED);

	/**
	 * Default constructor.
	 */
    public JPIPStatusPanel() {
        setBorder(BorderFactory.createEtchedBorder());

        setPreferredSize(new Dimension(55,20));
        setText("JPIP:");
        setIcon(disconnectedIcon);
        setVerticalTextPosition(JLabel.CENTER);
        setHorizontalTextPosition(JLabel.LEFT);
        
        ViewListenerDistributor.getSingletonInstance().addViewListener(this);
    }

    /**
     * Updates the JPIP Status, shows a popup if the status has changed.
     */
    public void updateJPIPStatus()
    {
    	if(activeView != null) {
    		if(activeView.isConnectedToJPIP()) {
	    		setIcon(connectedIcon);
	    	} else {
	    		setIcon(disconnectedIcon);
	    	}
    	
	    	if(lastStatus != activeView.isConnectedToJPIP()) {
	    		lastStatus = activeView.isConnectedToJPIP();
	    		statusChangedWindow.popUp(lastStatus);
	    	}
    	}
    }

    /**
     * {@inheritDoc}
     */
	public void viewChanged(View sender, ChangeEvent aEvent) {
		if(aEvent.reasonOccurred(ActiveLayerChangedReason.class)) {
    		activeView = ViewHelper.getViewAdapter(sender, JHVJP2View.class);
    		
    		if(activeView != null && activeView.isRemote()) {
    			setVisible(true);
    			updateJPIPStatus();
    		} else {
    			setVisible(false);
    		}
    	} else if(isVisible()) {
    		if(activeView != null && activeView.isRemote())
    			updateJPIPStatus();
    		else
    			setVisible(false);
    	}
	}
	
	/** Internal function for accessing this pane from within the popup.
	 * 
	 * @return Reference to this
	 */
	private JLabel getLabel() {
		return this;
	}
	
	/** 
	 * Popup indicating that the status has changed.
	 */
	private class StatusChangedWindow extends JWindow implements Runnable{
		
		private static final long serialVersionUID = 1L;
		private JLabel label = new JLabel("", JLabel.CENTER);
		
		/** Shows the popup.
		 * 
		 * @param newStatus New status to show
		 */
		public void popUp(boolean newStatus) {
			remove(label);
			if(newStatus) {
				label = new JLabel("JPIP Server is online");
			} else {
				label = new JLabel("JPIP Server is offline");
			}
			label.setPreferredSize(new Dimension(label.getPreferredSize().width + 10, label.getPreferredSize().height + 10));
			add(label);
			setPreferredSize(label.getPreferredSize());
			pack();
			setLocation(getLabel().getLocationOnScreen().x - 40, getLabel().getLocationOnScreen().y - 40);
			setVisible(true);
			
			Thread closeThread = new Thread(this, "Close JPIP Popup");
			closeThread.start();
		}
		
		/**
		 * Closes the popup after three seconds
		 */
		public void run() {
			try {
				Thread.sleep(3000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			this.setVisible(false);
		}
	}
}
