package org.helioviewer.jhv.gui.controller;

import java.awt.Cursor;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import org.helioviewer.base.math.Vector2dInt;
import org.helioviewer.base.physics.Constants;
import org.helioviewer.jhv.gui.components.BasicImagePanel;
import org.helioviewer.jhv.gui.components.EventInformationWindow;
import org.helioviewer.jhv.gui.interfaces.ImagePanelPlugin;
import org.helioviewer.jhv.internal_plugins.overlay.solarevents.SolarEvent;
import org.helioviewer.jhv.internal_plugins.overlay.solarevents.SolarEventCatalog;
import org.helioviewer.jhv.internal_plugins.overlay.solarevents.SolarEventCatalogs;
import org.helioviewer.jhv.internal_plugins.overlay.solarevents.SolarEvent.SolarEventException;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.changeevent.MetaDataChangedReason;
import org.helioviewer.viewmodel.changeevent.RegionChangedReason;
import org.helioviewer.viewmodel.changeevent.ViewportChangedReason;
import org.helioviewer.viewmodel.view.RegionView;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.ViewHelper;
import org.helioviewer.viewmodel.view.ViewListener;
import org.helioviewer.viewmodel.view.ViewportView;
import org.helioviewer.viewmodel.viewportimagesize.ViewportImageSize;

/** Implementation of ImagePanelPlugin for showing event popups.
 * 
 * <p>This plugin provides the capability to open an event popup
 * when clicking on an event icon within the main image. Apart
 * from that, it changes the mouse pointer when hovering over an
 * event icon to indicate that it is clickable.
 * 
 * @author Markus Langenberg
 *
 */
public class ImagePanelEventPopupController implements ImagePanelPlugin, MouseListener, MouseMotionListener, ViewListener {

    // ///////////////////////////////////////////////////////////////////////////
    // Definitions
    // ///////////////////////////////////////////////////////////////////////////

    private static final Cursor helpCursor = Cursor.getPredefinedCursor(Cursor.HAND_CURSOR);
    private static final int xOffset = 0;
    private static final int yOffset = 16;

    private View view;
    private ViewportView viewportView;
    private RegionView regionView;

    private BasicImagePanel imagePanel;

    private SolarEvent mouseOverSolarEvent = null;
    private Cursor lastCursor;
    private EventInformationWindow popUp;
	
    // ///////////////////////////////////////////////////////////////////////////
	// Methods
	// ///////////////////////////////////////////////////////////////////////////
    
    /**
     * {@inheritDoc}
     */
    public View getView() {
        return view;
    }

    /**
     * {@inheritDoc}
     */
    public void setView(View newView) {
        view =  newView;

        viewportView = view.getAdapter(ViewportView.class);
        regionView = view.getAdapter(RegionView.class);
    }

    /**
     * {@inheritDoc}
     */
    public void setImagePanel(BasicImagePanel newImagePanel) {
        imagePanel = newImagePanel;
        imagePanel.addMouseListener(this);
        imagePanel.addMouseMotionListener(this);

        if(imagePanel.getView() != null)
            imagePanel.getView().addViewListener(this);
    }

    /**
     * {@inheritDoc}
     */
    public BasicImagePanel getImagePanel() {
        return imagePanel;
    }
    
    /** Converts physical coordinate to screen coordinates
     * 
     * @param x Physical x-coordinate
     * @param y Physical y-coordinate
     * @return Corresponding screen coordinate
     */
    private Vector2dInt convertPhysicalToScreen(double x, double y) {
        ViewportImageSize viewportImageSize = ViewHelper.calculateViewportImageSize(viewportView.getViewport(), regionView.getRegion());

        Vector2dInt offset = ViewHelper.convertImageToScreenDisplacement(
                -regionView.getRegion().getUpperLeftCorner().getX(),
                regionView.getRegion().getUpperLeftCorner().getY(),
                regionView.getRegion(), viewportImageSize);

        return ViewHelper.convertImageToScreenDisplacement(x, y, regionView.getRegion(), viewportImageSize).add(offset);
    }

    /**
     * {@inheritDoc}
     */
    public void mouseClicked(MouseEvent e) {
        if(mouseOverSolarEvent != null) {

            if(popUp == null) {
                popUp = new EventInformationWindow();
            }

            popUp.setEvent(mouseOverSolarEvent);

            int xCoord = 0;
            if(e.getPoint().x + popUp.getSize().width < imagePanel.getSize().width) {
                xCoord = e.getPoint().x + imagePanel.getLocationOnScreen().x + xOffset;
            } else {
                xCoord = e.getPoint().x + imagePanel.getLocationOnScreen().x - popUp.getSize().width;
            }

            int yCoord = 0;
            if(e.getPoint().y + popUp.getSize().height < imagePanel.getSize().height) {
                yCoord = e.getPoint().y + imagePanel.getLocationOnScreen().y + yOffset;
            } else {
                yCoord = e.getPoint().y + imagePanel.getLocationOnScreen().y - popUp.getSize().height;
            }

            popUp.setLocation(xCoord, yCoord);
            popUp.setVisible(true);
            imagePanel.setCursor(helpCursor);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void mouseEntered(MouseEvent e) {}

    /**
     * {@inheritDoc}
     */
    public void mouseExited(MouseEvent e) {}

    /**
     * {@inheritDoc}
     */
    public void mousePressed(MouseEvent e) {}

    /**
     * {@inheritDoc}
     */
    public void mouseReleased(MouseEvent e) {}

    /**
     * {@inheritDoc}
     */
    public void mouseDragged(MouseEvent e) {}

    /**
     * {@inheritDoc}
     */
    public void mouseMoved(MouseEvent e) {

        SolarEvent lastSolarEvent = mouseOverSolarEvent;
        mouseOverSolarEvent = null;

        for (SolarEventCatalog catalog : SolarEventCatalogs.getSingletonInstance().getVisibleCatalogs()) {
            for (SolarEvent event : catalog.getEvents()) {

                try {
                    Vector2dInt eventPos = convertPhysicalToScreen(
                            Constants.SunRadius * event.getDouble("sunX"), // Offset of event in x-axis starting from the center of the sun
                            Constants.SunRadius * event.getDouble("sunY")); // Offset of event in y-axis starting from the center of the sun (down=positive)


                    if (e.getPoint().getX() >= eventPos.getX() - 8 && e.getPoint().getX() <= eventPos.getX() + 8
                            && e.getPoint().getY() >= eventPos.getY() - 8 && e.getPoint().getY() <= eventPos.getY() + 8) {
                        mouseOverSolarEvent = event;
                    }
                } catch (SolarEventException see) {}

            }
        }

        if(lastSolarEvent == null && mouseOverSolarEvent != null) {
            lastCursor = imagePanel.getCursor();
            imagePanel.setCursor(helpCursor);
        } else if(lastSolarEvent != null && mouseOverSolarEvent == null) {
            imagePanel.setCursor(lastCursor);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void viewChanged(View sender, ChangeEvent aEvent) {
        if(popUp != null) {
            if(aEvent.reasonOccurred(RegionChangedReason.class)
                    || aEvent.reasonOccurred(ViewportChangedReason.class)
                    || aEvent.reasonOccurred(MetaDataChangedReason.class)) {
                popUp.setVisible(false);
            }
        }
    }

    public void detach() { }

}
