package org.helioviewer.jhv.gui.controller;

import java.util.Date;
import java.util.GregorianCalendar;

import org.helioviewer.base.math.Vector2dDouble;
import org.helioviewer.base.physics.Astronomy;
import org.helioviewer.base.physics.Constants;
import org.helioviewer.base.physics.DifferentialRotation;
import org.helioviewer.base.physics.HeliocentricCartesianCoordinatesFromEarth;
import org.helioviewer.base.physics.StonyhurstHeliographicCoordinates;
import org.helioviewer.jhv.gui.ViewListenerDistributor;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.changeevent.MetaDataChangedReason;
import org.helioviewer.viewmodel.metadata.ObserverMetaData;
import org.helioviewer.viewmodel.region.Region;
import org.helioviewer.viewmodel.region.StaticRegion;
import org.helioviewer.viewmodel.view.LinkedMovieManager;
import org.helioviewer.viewmodel.view.TimedMovieView;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.ViewListener;

/** Implementation of ImagePanelInputController for the main image panel
 * using focus selection mode.
 * 
 * <p>The main feature about this mode is, that is uses the differential
 * rotation tracking to keep the center of the screen over a certain
 * point of the surface of the sun.
 * 
 * <p>Apart from that, the user can move the region of interest of
 * the main image panel by dragging the image around. Also, by using the
 * mouse wheel or double-clicking into the image, zooming is possible.
 * 
 * <p>Also, see {@link org.helioviewer.jhv.gui.components.MainImagePanel}.
 *
 * @author Ludwig Schmidt
 */
public class MainImagePanelMouseFocusController extends
MainImagePanelMousePanController implements ViewListener {

    private Date lastFrameDate = null;

    /**
     * {@inheritDoc}
     */
    public synchronized void setView(View v) {
        super.setView(v);
        if(v != null) {
        	ViewListenerDistributor.getSingletonInstance().addViewListener(this);
        }
    }

    /**
     * {@inheritDoc}
     */
    public synchronized void viewChanged(View sender, ChangeEvent aEvent) {

        MetaDataChangedReason metaDataReason = aEvent.getLastChangedReasonByType(MetaDataChangedReason.class);
        if(metaDataReason != null 
        		&& metaDataReason.getNewMetaData() instanceof ObserverMetaData
        		&& metaDataReason.getView() instanceof TimedMovieView 
        		&& LinkedMovieManager.getActiveInstance().isMaster((TimedMovieView) metaDataReason.getView())) {
        	
            Date currentDate = ((ObserverMetaData) metaDataReason.getNewMetaData()).getDateTime().getTime();
            GregorianCalendar currentCalendarDate = new GregorianCalendar();
            currentCalendarDate.setTime(currentDate);
            
            if(lastFrameDate != null) {
                long timeDiff = (currentDate.getTime() - lastFrameDate.getTime()) / 1000;
                
                if(timeDiff == 0) {
                    return;
                }
                Region currentRegion = regionView.getRegion();
                Vector2dDouble center = currentRegion.getLowerLeftCorner().add(currentRegion.getSize().scale(0.5));
                // don't rotate anything if the center is not in the sun
                if(center.lengthSq() >= Constants.SunRadius * Constants.SunRadius) {
                    return;
                }
                
                HeliocentricCartesianCoordinatesFromEarth hcc = new HeliocentricCartesianCoordinatesFromEarth(center, currentCalendarDate);
                StonyhurstHeliographicCoordinates currentPosition = new StonyhurstHeliographicCoordinates(hcc);
                StonyhurstHeliographicCoordinates nextPosition = DifferentialRotation.calculateNextPosition(currentPosition, timeDiff);
                HeliocentricCartesianCoordinatesFromEarth nextHCC = new HeliocentricCartesianCoordinatesFromEarth(nextPosition, currentCalendarDate);
                
                /*
                System.out.println("Current center in Stonyhurst: " + currentPosition.toString());
                System.out.println("Next center in Stonyhuyrst: " + nextPosition.toString());
                System.out.println("Next center in HCC: " + nextHCC.toString());
                System.out.println("B0: " + Astronomy.getB0InDegree(currentCalendarDate));
                System.out.println("timediff: " + timeDiff + "    center: (" + center.getX() + ", " + center.getY() + ")     nextPosition: (" + nextPosition.getX() + ", " + nextPosition.getY() + ")    movement: (" + movement.getX() + ", " + movement.getY() + ")");
                */
                
                Vector2dDouble nextCenter = nextHCC.getCartesianCoordinatesOnDisc();
                Vector2dDouble movement = nextCenter.subtract(center);
                Region newRegion = StaticRegion.createAdaptedRegion(regionView.getRegion().getLowerLeftCorner().add(movement), regionView.getRegion().getSize());
                regionView.setRegion(newRegion, new ChangeEvent());
            }
            lastFrameDate = currentDate;
        }
    }

    /**
     * {@inheritDoc}
     */
    public void detach() {
        ViewListenerDistributor.getSingletonInstance().removeViewListener(this);
        lastFrameDate = null;
    }
}
