package org.helioviewer.jhv.gui.controller;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;

import org.helioviewer.base.math.Vector2dDouble;
import org.helioviewer.jhv.gui.IconBank;
import org.helioviewer.jhv.gui.IconBank.JHVIcon;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.metadata.MetaData;
import org.helioviewer.viewmodel.region.Region;
import org.helioviewer.viewmodel.region.StaticRegion;
import org.helioviewer.viewmodel.view.RegionView;
import org.helioviewer.viewmodel.view.SynchronizeView;
import org.helioviewer.viewmodel.view.ViewHelper;
import org.helioviewer.viewmodel.view.ViewportView;
import org.helioviewer.viewmodel.viewport.Viewport;

/** Implementation of ImagePanelInputController for the overview image panel.
 * 
 * <p>By using this controller, the user can move the region of interest of
 * the main image panel. This can happen on two ways: Either by dragging the
 * current region or by clicking into the area outside the current region,
 * which will center the region on that point.
 * 
 * <p>Also, see {@link org.helioviewer.jhv.gui.components.OverviewImagePanel}.
 * 
 * @author Stephan Pagel
 * 
 * */
public class OverviewImagePanelMousePanController extends AbstractImagePanelMouseController {

	// ///////////////////////////////////////////////////////////////////////////
	// Definitions
	// ///////////////////////////////////////////////////////////////////////////
	
	// Cursor image when mouse button is pressed
	private static final Cursor closedHandCursor = Toolkit.getDefaultToolkit()
    .createCustomCursor(IconBank.getIcon(JHVIcon.CLOSED_HAND).getImage(),
        new Point(9, 9),
        IconBank.getIcon(JHVIcon.CLOSED_HAND).toString());

	// Cursor image when cursor is over the image display area
	private static final Cursor openHandCursor = Toolkit.getDefaultToolkit()
    .createCustomCursor(IconBank.getIcon(JHVIcon.OPEN_HAND).getImage(),
        new Point(9, 9),
        IconBank.getIcon(JHVIcon.OPEN_HAND).toString());

	// position and size of ROI rectangle
	private Rectangle roiArea = new Rectangle();
	// position and size of available image data in overview image area
	private Rectangle imageArea = new Rectangle();
	
	// flag if mouse cursor is inside ROI area
	private boolean insideROI = false;
	// flag if rectangle for ROI was dragged
	private boolean insideROIDragged = false;
	// point where the mouse was last
	private Point lastMouseLocation = new Point(0,0);
	
	// memorize position where user clicked in ROI area relative to 
	// lower left corner of ROI rectangle
	private Dimension dimensionFromBL = null;
	
	// Zoom controller
	private ZoomController zoomController = new ZoomController();
	
	// ///////////////////////////////////////////////////////////////////////////
    // Methods
    // ///////////////////////////////////////////////////////////////////////////
	
    /**
     * Sets the area where image data is available. (ViewportImageSize + Offset)
     * 
     * @param imageArea
     *            Position and size of the area where image data is available.
     * */
    public void setImageArea(Rectangle imageArea) {
        this.imageArea = imageArea;
    }

    /**
     * Sets the area where the ROI in main image currently is.
     * 
     * @param roiArea
     *            Position and size of the ROI of the main image adopted to the
     *            overview.
     * */
    public void setROIArea(Rectangle roiArea) {
        this.roiArea = roiArea;
        if(!insideROIDragged) {
        	setCursorDependingOnRegion(lastMouseLocation);
        }
    }
	
    /**
     * Checks if mouse cursor is placed over area where image data is available.
     * 
     * @param point
     *            Coordinates of the mouse cursor.
     * @returns Boolean value; true if mouse cursor is located over this area.
     * */
    private boolean isCursorInImageArea(Point point) {
        return point.x >= imageArea.x
                && point.x <= imageArea.x + imageArea.width
                && point.y >= imageArea.y
                && point.y <= imageArea.y + imageArea.height;
    }
    
    /**
     * Checks if mouse cursor is placed over ROI.
     * 
     * @param point
     *            Coordinates of the mouse cursor.
     * @returns Boolean value; true if mouse cursor is located over this area.
     * */
    private boolean isCursorInROIArea(Point point) {
        return point.x >= roiArea.x
                && point.x <= roiArea.x + roiArea.width
                && point.y >= roiArea.y
                && point.y <= roiArea.y + roiArea.height;
    }
    
    /**
     * Sets the image of the mouse cursor depending on its position.
     * 
     * @param point
     *            Coordinates of the mouse cursor.
     * */
    private void setCursorDependingOnRegion(Point point) {
        // check if mouse is inside image area
        if (isCursorInImageArea(point)) {
            
            // check if mouse is inside image area
            if (isCursorInROIArea(point)) {
             
                // Cursor is located in ROI area
                imagePanel.setCursor(openHandCursor);
                insideROI = true;
            } else {
                // Cursor is located in area where image data is available but outside of ROI area
                imagePanel.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
                insideROI = false;
            }
        } else {
            // Cursor is located in area where no image data is available            
            imagePanel.setCursor(Cursor.getDefaultCursor());
            insideROI = false;
        }
    }
    
    /**
     * Sets the region in main image depending on the mouse position in overview
     * image.
     * 
     * @param coordinate
     *            Coordinate of cursor in overview image.
     * */
    private void updateRegion(Point coordinate) {
    	
    	SynchronizeView synchronizeView = view.getAdapter(SynchronizeView.class);
    	
    	if(synchronizeView == null || synchronizeView.getObservedView() == null) {
    		return;
    	}

        // get needed information and check values
        RegionView mainRegionView = synchronizeView.getObservedView().getAdapter(RegionView.class);
        ViewportView mainViewportView = synchronizeView.getObservedView().getAdapter(ViewportView.class);
        
        if (mainRegionView == null || mainViewportView == null || coordinate == null) {            
            return;
        }
        
        Region regionMainImage = mainRegionView.getRegion();
        Viewport viewportMainImage = mainViewportView.getViewport();
        MetaData metaData = metaDataView.getMetaData();
        
        if (regionMainImage == null || viewportMainImage == null || metaData == null) {            
            return;
        }
        
        // compute coordinates relativized to view port image size
        Point point = coordinatesRelativeToViewportSize(coordinate);
        
        // calculate new position of lower left corner of rectangle
        if (insideROIDragged) {
            // ... when rectangle was dragged
            point.x = point.x - dimensionFromBL.width;
            point.y = point.y + dimensionFromBL.height;
        } else {
            // ... when rectangle was placed by mouse click
            point.x = point.x - (roiArea.width / 2);
            point.y = point.y + (roiArea.height / 2);
        }
        
        // compute new position of lower left corner of region in physical image 
        double x = point.x / (double) imageArea.getWidth() * metaData.getPhysicalImageWidth() + metaData.getPhysicalLowerLeft().getX();
        double y = point.y / (double) imageArea.getHeight() * metaData.getPhysicalImageHeight() - metaData.getPhysicalUpperLeft().getY();
        Vector2dDouble newCorner = new Vector2dDouble(x, -y);
        
        // set new region in main image
        Region newRegion = ViewHelper.cropRegionToImage(StaticRegion.createAdaptedRegion(newCorner, regionMainImage.getSize()), metaDataView.getMetaData());
        mainRegionView.setRegion(newRegion, new ChangeEvent());       
    }
    
    /** Updates the view of the zoom controller.
     * 
     * It is always set to the observed view of the synchronize view
     */
    private void updateZoomController() {
    	SynchronizeView synchronizeView = view.getAdapter(SynchronizeView.class);
    	
    	if(synchronizeView == null) {
    		return;
    	}
    	
    	zoomController.setView(synchronizeView.getObservedView());
    }
    
    /**
     * Computes the cursor coordinate in relation to view port size only.
     * 
     * @param coordinates
     *            Cursor coordinates in overview image area.
     * @return Coordinates relativized to view port size.
     * */
    private Point coordinatesRelativeToViewportSize(Point coordinates) {
        
        int x = coordinates.x - imageArea.x;
        x = x < 0 ? 0 : (x > imageArea.width ? imageArea.width : x);
        
        int y = coordinates.y - imageArea.y;
        y = y < 0 ? 0 : (y > imageArea.height ? imageArea.height : y);
        
        return new Point(x,y);
        
    }
	
	// ///////////////////////////////////////////////////////////////////////////
	// Mouse events
	// ///////////////////////////////////////////////////////////////////////////
	
    /**
     * {@inheritDoc}
     */
    public void mouseClicked(MouseEvent e) {
    	if(e.getClickCount() > 1) {
    		updateZoomController();
    		
    		if(e.getButton() == MouseEvent.BUTTON1) {
    			zoomController.zoomIn();
    		} else if(e.getButton() == MouseEvent.BUTTON3) {
    			zoomController.zoomOut();
    		}	
    	}
    }

    /**
     * {@inheritDoc}
     */
    public void mouseEntered(MouseEvent e) {}

    /**
     * {@inheritDoc}
     */
    public void mouseExited(MouseEvent e) {}

    /**
     * {@inheritDoc}
     */
    public void mousePressed(MouseEvent e) {
        if (insideROI) {
            // memorize position where user clicked inside rectangle
            dimensionFromBL = new Dimension(e.getPoint().x - roiArea.x,
                    roiArea.y + roiArea.height - e.getPoint().y);

            // update cursor
            imagePanel.setCursor(closedHandCursor);
        }
    }

    /**
     * {@inheritDoc}
     */
    public void mouseReleased(MouseEvent e) {
	    // update region
	    updateRegion(e.getPoint());
	    
	    // set flag that ROI rectangle will not be dragged anymore
	    insideROIDragged = false;
	    
	    // update cursor
	    setCursorDependingOnRegion(e.getPoint());
	}

    /**
     * {@inheritDoc}
     */
    public void mouseDragged(MouseEvent e) {
    	lastMouseLocation = e.getPoint();
    	
	    if (insideROI) {
	        // set flag for dragging the ROI rectangle
	        insideROIDragged = true;
	        
	        // update region
	        updateRegion(e.getPoint());
	    }
	}

    /**
     * {@inheritDoc}
     */
    public void mouseMoved(MouseEvent e) {
    	lastMouseLocation = e.getPoint();
	    // update cursor
	    setCursorDependingOnRegion(e.getPoint());
	}

    /**
     * {@inheritDoc}
     */
    public void mouseWheelMoved(MouseWheelEvent e) {
    	if(zoomController != null) {
    		updateZoomController();
    		zoomController.zoomSteps((int) (-Math.max(1, e.getUnitsToScroll() / 3) * Math.signum(e.getUnitsToScroll())));
    	}
    }
}
