package org.helioviewer.jhv.gui.controller;

import java.awt.Dimension;

import org.helioviewer.base.math.Vector2dDouble;
import org.helioviewer.base.math.Vector2dInt;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.metadata.ImageSizeMetaData;
import org.helioviewer.viewmodel.region.Region;
import org.helioviewer.viewmodel.region.StaticRegion;
import org.helioviewer.viewmodel.view.ImageInfoView;
import org.helioviewer.viewmodel.view.MetaDataView;
import org.helioviewer.viewmodel.view.RegionView;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.ViewHelper;

/** Collection of several zooming functions.
 *
 * <p>This class provides several zooming functions. The controller
 * is used by several classes, such as
 * {@link org.helioviewer.jhv.gui.actions.ZoomInAction},
 * {@link org.helioviewer.jhv.gui.actions.ZoomOutAction},
 * {@link org.helioviewer.jhv.gui.actions.ZoomFitAction},
 * {@link org.helioviewer.jhv.gui.actions.Zoom1to1Action} and
 * {@link org.helioviewer.jhv.gui.controller.MainImagePanelMouseController}.
 */
public class ZoomController {

    private View view;
    private RegionView regionView;
    private MetaDataView metaDataView;

    private static final double zoomFactorStep = Math.sqrt(2);

    
    /** Returns the topmost view of the view chain  associated
     * with this controller.
     * 
     * @return Topmost view of the view chain
     * @see #setView(View)
     */
    public View getView() {
        return view;
    }

    /** Sets the topmost view of the view chain.
	 * 
	 * That way the controller can access the whole view chain.
	 * 
	 * @param newView Topmost view of the view chain 
	 * @see #getView()
	 */
    public void setView(View newView) {
        view = newView;
        regionView = view == null ? null : view.getAdapter(RegionView.class);
        metaDataView = view == null ? null : view.getAdapter(MetaDataView.class);
    }

    /**
     * Zooms in one step.
     * A step mean scaling the current region of interest by the
     * square root of two
     */
    public void zoomIn() {
        zoom(zoomFactorStep);
    }

    /**
     * Zooms out one step.
     * A step mean scaling the current region of interest by the
     * square root of two
     */
    public void zoomOut() {
        zoom(1.0 / zoomFactorStep);
    }

    /**
     * Zooms in or out the desired number of steps.
     * A step mean scaling the current region of interest by the
     * square root of two. To zoom in, steps has to be greater
     * than zero, to zoom out it has to be lesser than zero.
     * 
     * @param steps Number of steps to zoom, the sign defines the direction.
     */
    public void zoomSteps(int steps) {
        zoom(Math.pow(zoomFactorStep, steps));
    }

    /** Zooms by scaling the current region by the given zoom factor.
     *  
     * @param zoomFactor zoom factor to scale the current region with
     */
    public void zoom(double zoomFactor) {
        if(regionView != null && metaDataView != null && metaDataView.getMetaData() != null) {
            Region oldRegion = regionView.getRegion();
            if(oldRegion == null) {
                return;
            }
            Vector2dDouble newSizeVector = Vector2dDouble.scale(oldRegion.getSize(), 1.0 / zoomFactor);
            Vector2dDouble newCorner = Vector2dDouble.add(oldRegion.getLowerLeftCorner(),
                    Vector2dDouble.scale(Vector2dDouble.subtract(oldRegion.getSize(), newSizeVector), 0.5));
            Region zoomedRegion = ViewHelper.cropRegionToImage(StaticRegion.createAdaptedRegion(newCorner, newSizeVector), metaDataView.getMetaData());
            regionView.setRegion(zoomedRegion, new ChangeEvent());
        }
    }
    
    /** 
     * Zooms the image in such a way, that the active layer is displayed in its
     * native resolution.
     */
    public void zoom1to1() {    	
    	ImageInfoView view = ImageViewerGui.getSingletonInstance().getImageSelectorPanel().getActiveImageInfoView();
    	
    	if(view == null)
    		return;

		Dimension panelSize = ImageViewerGui.getSingletonInstance().getMainImagePanel().getSize();
		ImageSizeMetaData imageSizeMetaData = (ImageSizeMetaData) view.getAdapter(MetaDataView.class).getMetaData();
		Vector2dInt maxResolution = imageSizeMetaData.getResolution();
		RegionView oldView = ImageViewerGui.getSingletonInstance().getMainView().getAdapter(RegionView.class);

		int newWidth = panelSize.width;
		int newHeight = panelSize.height;

		if(panelSize.width > maxResolution.getX()) {
			newWidth = maxResolution.getX();
		}

		if(panelSize.height > maxResolution.getY()) {
			newHeight = maxResolution.getY();
		}
		
		Vector2dDouble newSize = new Vector2dDouble(newWidth, newHeight).scale(imageSizeMetaData.getUnitsPerPixel());
		
		Region oldRegion = oldView.getRegion();
		Vector2dDouble newCenter = oldRegion.getLowerLeftCorner().add(oldRegion.getSize().scale(0.5));
        
        Region newRegion = ViewHelper.cropRegionToImage(
            StaticRegion.createAdaptedRegion(
            		newCenter.subtract(newSize.scale(0.5)),
            		newSize),
            oldView.getAdapter(MetaDataView.class).getMetaData());
		
        oldView.setRegion(newRegion, new ChangeEvent());
    }
    
    /** 
     * Zooms the image in such a way, that the active layer fits exactly into
     * the viewport.
     */
    public void zoomFit() {
    	ImageInfoView view = ImageViewerGui.getSingletonInstance().getImageSelectorPanel().getActiveImageInfoView();
    	
    	if(view == null)
    		return;
    	
		Region newRegion = view.getAdapter(MetaDataView.class).getMetaData().getPhysicalRegion();
		ImageViewerGui.getSingletonInstance().getMainView().getAdapter(RegionView.class).setRegion(newRegion, new ChangeEvent());
    }

}
