package org.helioviewer.jhv.gui.dialogs;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.helioviewer.jhv.JHVDirectory;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.gui.interfaces.ShowableDialog;
import org.helioviewer.viewmodel.metadata.HelioviewerMetaData;
import org.helioviewer.viewmodel.metadata.MetaData;
import org.helioviewer.viewmodel.view.ImageInfoView;
import org.helioviewer.viewmodel.view.MetaDataView;
import org.helioviewer.viewmodel.view.fitsview.JHVFITSView;
import org.helioviewer.viewmodel.view.jp2view.JHVJP2View;
import org.helioviewer.viewmodel.view.jp2view.JHVJPXView;
import org.helioviewer.viewmodel.view.jp2view.JP2Image;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_KduException;
import org.helioviewer.viewmodel.view.jp2view.kakadu.KakaduUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/** Dialog that is used to display meta data for an image.
 * 
 * @author Alen Agheksanterian
 * @author Stephan Pagel
 */
public class MetaDataDialog extends JDialog implements ActionListener, ShowableDialog {
	
	private static final long serialVersionUID = 1L;

	private final JButton closeButton = new JButton("Close");
	private final JButton exportFitsButton = new JButton("Export FITS Header as XML");
	//private final JButton exportButton = new JButton("Export as XML");

	private Vector<String> infoList = new Vector<String>();
	private JList listBox = new JList();
	private Document xmlDoc = null;
	private boolean metaDataOK;
	private String outFileName;

	
	/**
	 * The private constructor that sets the fields and the dialog.
	 */
	public MetaDataDialog() {
		super(ImageViewerGui.getMainFrame(), "Image Information");
		setAlwaysOnTop(true);
		setLayout(new BorderLayout());
		setResizable(false);
		
		listBox.setFont(new Font("Courier",Font.PLAIN, 12));
				
		JPanel bottomPanel = new JPanel();
		bottomPanel.setLayout(new FlowLayout(FlowLayout.RIGHT));
		bottomPanel.add(exportFitsButton);
		//bottomPanel.add(exportButton);
		bottomPanel.add(closeButton);
		
		JScrollPane listScroller = new JScrollPane(listBox);
		this.add(listScroller, BorderLayout.CENTER);
		this.add(bottomPanel, BorderLayout.PAGE_END);
		
		// add action listeners to the buttons
		closeButton.addActionListener(this);
		exportFitsButton.addActionListener(this);
		//exportButton.addActionListener(this);
	}

	/** 
	 * Resets the list.
	 */
	public void resetData()
	{
		this.infoList.removeAllElements();
		
		// update the listBox
		this.listBox.setListData(this.infoList); 
		
		// set the status of export button
		if (!metaDataOK)
		{
			exportFitsButton.setEnabled(false);
		}
		else
		{
			exportFitsButton.setEnabled(true);
		}
	}
	
	/** 
	 * Adds a data item to the list 
	 * 
	 * @param _item New item to add
	 * @see #setMetaData(MetaDataView)
	 */
	public void addDataItem(String _item)
	{
		this.infoList.add(_item);
		
		// update the listBox
		this.listBox.setListData(this.infoList); 
	}
	
	/** 
	 * {@inheritDoc}
	 */
	public void showDialog(){
		pack();
		if (!metaDataOK)
			setSize(450, 200);
		else
			setSize(450, 600);

		setLocationRelativeTo(ImageViewerGui.getMainFrame());
		setVisible(true);
	}
	
	/** Sets the full document which can be found reading the given MetaDataView.
	 * 
	 * @param metaDataView Source to read
	 * @see #addDataItem(String)
	 */
	public void setMetaData(MetaDataView metaDataView) {
	    
	    if (metaDataView == null)
	        return;
	    
	    MetaData metaData = metaDataView.getMetaData();
	    
	    if (!(metaData instanceof HelioviewerMetaData)) {
            setMetadataStatus(false);
            resetData();
            addDataItem("No metadata is available.");
        } else {
            HelioviewerMetaData m = (HelioviewerMetaData) metaData;
            setMetadataStatus(true);
            resetData();
            addDataItem("-------------------------------");
            addDataItem("       Basic Information       ");
            addDataItem("-------------------------------");
            addDataItem("\tObservatory : " + m.getObservatory());
            addDataItem("\tInstrument  : " + m.getInstrument());
            addDataItem("\tDetector    : " + m.getDetector());
            addDataItem("\tMeasurement : " + m.getMeasurement());
            addDataItem("\tDate        : " + m.getDateTime().getFormattedDate());
            addDataItem("\tTime        : " + m.getDateTime().getFormattedTime());

            String xmlText = null;
            if (metaDataView instanceof JHVJP2View) {
                JP2Image img = ((JHVJP2View) metaDataView).getJP2Image();
                
                int boxNumber = 1;
                if(metaDataView instanceof JHVJPXView) {
                    boxNumber = ((JHVJPXView) metaDataView).getCurrentFrameNumber() + 1;
                }

                try {
                    xmlText = KakaduUtils.getXml(img.getFamilySrc(), boxNumber);
                } catch (JHV_KduException e) {
                    e.printStackTrace();
                }
            } else if(metaDataView instanceof JHVFITSView) {
            	xmlText = ((JHVFITSView) metaDataView).getHeaderAsXML();
            }

            if (xmlText != null) {
                String xml = xmlText.trim().replace("$OBS", "");
                
                InputStream in = null;
                try {
                    in = new ByteArrayInputStream(xml.getBytes("UTF-8"));
                } catch (Exception ex) {
                    ex.printStackTrace();
                }
                try {
                    DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
                    Document doc = builder.parse(in);

                    // Send xml data to meta data dialog box
                    Node root = doc.getDocumentElement().getElementsByTagName("fits").item(0);
                    writeXMLData(root, 0);
                    
                    // set the xml data for the MetaDataDialog
                    setXMLData(doc);

                    ImageInfoView iIV = metaDataView.getAdapter(ImageInfoView.class);

                    // set the export file name for
                    // MetaDataDialog
                    String filename = iIV.getUri().getPath();
                    filename = filename.substring(filename.lastIndexOf('/') + 1, filename.lastIndexOf('.'));

                    filename = JHVDirectory.EXPORTS.getPath() + filename;
                    setOutFileName(filename);

                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
	}
	
	/**
     * A method that writes the xml box specified by its root node to the list
     * box in image info dialog box.
     * 
     * @param node Node to write
     * @param indent Number of tabstops to insert
     */
    private void writeXMLData(Node node, int indent) {
        // get element name and value
        String nodeName = node.getNodeName();
        String nodeValue = getElementValue(node);

        if (nodeName.equals("fits")) {
            addDataItem("-------------------------------");
            addDataItem("          FITS Header");
            addDataItem("-------------------------------");
        } else {
            
            String tab = "";
            for (int i = 0; i < indent; i++) {
                tab = tab + "\t";
            }
            
            addDataItem(tab + nodeName + ": " + nodeValue);
        }

        // write the child nodes recursively
        NodeList children = node.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            Node child = children.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE) {
                writeXMLData(child, indent + 1);
            }
        }
    }
    
    /**
     * A method that gets the value of a node element.
     * 
     * If the node itself has children and no text value, an empty string
     * is returned. This is maybe an overkill for our purposes now, 
     * but takes into account the possibility of nested tags.
     * 
     * @param node Node to read
     */
    private final String getElementValue(Node elem) {
        Node child;
        if (elem != null) {
            if (elem.hasChildNodes()) {
                for (child = elem.getFirstChild(); child != null; child = child
                .getNextSibling()) {
                    if (child.getNodeType() == Node.TEXT_NODE) {
                        return child.getNodeValue();
                    }
                }
            }
        }
        return "";
    }
	
	/** Sets the XML data for saving later
	 * 
	 * @param _doc New XML document to remember
	 */
	private void setXMLData(Document _doc)
	{
		this.xmlDoc = _doc;
	}
	
	/** sets the file name where FITS data will be exported */
	private void setOutFileName(String _value)
	{
		this.outFileName = _value;
	}
	
	/** Sets the status of meta data 
	 * 
	 * @param _value True, if meta data is available, false otherwise
	 */
	private void setMetadataStatus(boolean _value)
	{
		this.metaDataOK = _value;
	}
	
	/**
	 * {@inheritDoc}
	 */ 
	public void actionPerformed(ActionEvent _a) {
		if (_a.getSource() == this.closeButton) {
			this.xmlDoc = null;
			this.infoList.removeAllElements();
			this.resetData();
			this.dispose();
			
		} else if (_a.getSource() == this.exportFitsButton) {
		    DOMSource source = new DOMSource(this.xmlDoc.getDocumentElement().getElementsByTagName("fits").item(0));
		    String filename = this.outFileName + ".fits.xml";
		    
			boolean saveSuccessful = saveXMLDocument(source, filename);
			if (saveSuccessful)
				JOptionPane.showMessageDialog(this, "Fits data saved to " + filename);
		}
/*	
		{
		    DOMSource source = new DOMSource(this.xmlDoc);
		    String filename = this.outFileName + ".xml";

			boolean saveSuccessful = saveXMLDocument(source, filename);
			if (saveSuccessful)
				JOptionPane.showMessageDialog(null, "XML data saved to " + filename);
		}
*/		
	}
    
    /** This routine saves the fits data into an XML file.
     * 
     * @param source XML document to save
     * @param filename XML file name
     */    
    public boolean saveXMLDocument(DOMSource source, String filename) 
    {        
        // open the output stream where XML Document will be saved
        File xmlOutFile = new File(filename);
        FileOutputStream fos;
        Transformer transformer;
        try {
            fos = new FileOutputStream(xmlOutFile);
        }
        catch (FileNotFoundException e) {
            e.printStackTrace();
            return false;
        }
        
        // Use a Transformer for the purpose of output
        TransformerFactory transformerFactory = TransformerFactory.newInstance();
        try {
            transformer = transformerFactory.newTransformer();
        }
        catch (TransformerConfigurationException e) {
            e.printStackTrace();
            return false;
        }
        
        // The source is the fits header
        
        // The destination for output
        StreamResult result = new StreamResult(fos);
        
        // transform source into result will do a file save
        try {
            transformer.transform(source, result);
        }
        catch (TransformerException e) {
            e.printStackTrace();
        }
        return true;
    }
}

