package org.helioviewer.jhv.gui.dialogs;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.util.LinkedList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UIManager.LookAndFeelInfo;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;
import javax.swing.table.TableModel;

import org.helioviewer.jhv.JHVDirectory;
import org.helioviewer.jhv.JHVGlobals;
import org.helioviewer.jhv.Settings;
import org.helioviewer.jhv.gui.IconBank;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.gui.IconBank.JHVIcon;
import org.helioviewer.jhv.gui.interfaces.ShowableDialog;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_Kdu_cache;


/** Dialog that allows the user to change default preferences and settings.
 * 
 * @author Desmond Amadigwe
 * @author Benjamin Wamsler
 * @author Juan Pablo
 * @author Markus Langenberg
 */
public class PreferencesDialog extends JDialog implements ShowableDialog {
	
	private static final long serialVersionUID = 1L;
	
	private final String defaultDateFormat = "yyyy/MM/dd";
	
	private JComboBox lafCombo;
	private JPanel paramsPanel;
	private JCheckBox limitMaxSize = new JCheckBox("Limit size");
	private JLabel occupiedSizeLabel = new JLabel();
	private JTextField maxCacheBox = new JTextField("0.0");
	private JLabel maxCacheBoxLabel = new JLabel(" Mbytes");
	private JRadioButton openGLItemOff;
	private JRadioButton openGLItemOn;
    private JRadioButton debugItemOff;
    private JRadioButton debugItemOn;
    private JRadioButton timerItemOff;
	private JRadioButton timerItemOn;
	private DefaultsSelectionPanel defaultsPanel;
	private JTextField dateFormatField;
	private JButton dateFormatInfo;
	
	private Settings settings = Settings.getSingletonInstance();
	
	/* This array will contain look and feels that are not allowed when using JHV */
	private String[] disallowedLafs = {"CDE/Motif"};
	
	/** 
	 * The private constructor that sets the fields and the dialog. 
	 */
	public PreferencesDialog() {
		
		super(ImageViewerGui.getMainFrame(), "Preferences", true);
		setResizable(false);
		
		JPanel mainPanel = new JPanel(new BorderLayout());
		
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());
		
		JPanel paramsSubPanel = new JPanel(new BorderLayout());
		paramsSubPanel.setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));
		paramsSubPanel.add(createParametersPanel(), BorderLayout.CENTER);
		
		JPanel defaultsSubPanel = new JPanel(new BorderLayout());
		defaultsSubPanel.setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));
		defaultsSubPanel.add(createDefaultSaveDirPanel(), BorderLayout.CENTER);
		
		JPanel jpipSupPanel = new JPanel(new BorderLayout());
		jpipSupPanel.setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));
		jpipSupPanel.add(createJPIPCachePanel(), BorderLayout.CENTER);
		
		panel.add(paramsSubPanel, BorderLayout.NORTH);
		panel.add(defaultsSubPanel, BorderLayout.CENTER);
		panel.add(jpipSupPanel, BorderLayout.SOUTH);
		
		mainPanel.add(panel, BorderLayout.CENTER);
		
		JPanel btnPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		btnPanel.setBorder(BorderFactory.createEmptyBorder(10, 0, 0, 0));
		
		JButton acceptBtn = new JButton(" Accept ");
		JButton cancelBtn = new JButton(" Cancel ");
		JButton resetBtn = new JButton(" Reset ");
		
		acceptBtn.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
			    if (!isDateFormatValid(dateFormatField.getText())) {
			        JHVGlobals.err("Syntax error", "The entered date pattern contains illegal signs!\nAll suppported signs are listed in the associated information dialog.", false);
			        return;
			    }
			    
			    try {
                    if (limitMaxSize.isSelected() && !(Double.parseDouble(maxCacheBox.getText()) > 0.0)) {
                        JHVGlobals.err("Invalid value", "The value for the maximal cache size must be greater than 0.", false);
                        return;
                    }
                } catch (NumberFormatException ex) {
                    JHVGlobals.err("Invalid value", "The value for the maximal cache size is not a number.", false);
                    return;
                }
			    
				saveSettings();				
				ImageViewerGui.getSingletonInstance().updateComponents();				
				dispose();
			}
		});
		
		cancelBtn.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e)
			{
				setLookAndFeelCombo(settings.getProperty("look.and.feel"));
				dispose();
			}
		});
		
		resetBtn.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {	
				
				if(JOptionPane.showConfirmDialog(null, 
						"Do you really want to reset the setting values?", "Attention", 
						JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
				
					defaultsPanel.resetSettings();
					
					setLookAndFeelCombo("javax.swing.plaf.metal.MetalLookAndFeel");
					
					openGLItemOn.setSelected(true);
					openGLItemOn.doClick();
					
					debugItemOff.setSelected(true);
					debugItemOff.doClick();

					maxCacheBox.setText("0.0");	
					
					dateFormatField.setText(defaultDateFormat);
				}
			}
		});
		
		if(System.getProperty("os.name").toUpperCase().contains("WIN")) {
			btnPanel.add(acceptBtn);
			btnPanel.add(resetBtn);
			btnPanel.add(cancelBtn);
		} else {
			btnPanel.add(resetBtn);
			btnPanel.add(cancelBtn);
			btnPanel.add(acceptBtn);
		}
		
		mainPanel.add(btnPanel, BorderLayout.SOUTH);
		
		getContentPane().add(mainPanel);
		pack();
	}

    /**
     * Checks the passed pattern if it is a supported date pattern. The pattern
     * could contain defined letters and special characters. The method checks
     * valid signs only!
     * 
     * @param format
     *            pattern to check.
     * @return boolean value if pattern is supported.
     */
	private boolean isDateFormatValid(String format) {
	    
	    // go through all signs of pattern
	    for (int i = 0; i < format.length(); i++) {
	        char sign = format.charAt(i);
	        int ascii = (int) sign;
	        
	        // if it is a number or letter, check it if it is supported
	        if ((ascii >= 48 && ascii <= 57) || (ascii >=65 && ascii <=90) || (ascii >= 97 && ascii <= 122)) {
	            if (sign != 'y' && sign != 'M' && sign != 'd' && sign != 'w' &&
	                    sign != 'D' && sign != 'E'){
	                
	                return false;
	            }
	        }	            
	    }    
	    
	    return true;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void showDialog() {
		
		loadSettings();
		
		pack();
		setSize(getPreferredSize());
		setLocationRelativeTo(ImageViewerGui.getMainFrame());
		setVisible(true);
	}
	
	/** Method that returns allowed look and feels for JHV
	 * 
	 * @return Array containing allowed look and feels
	 */
	private UIManager.LookAndFeelInfo[] getAllowedLookAndFeels() {
		
		UIManager.LookAndFeelInfo[] lafs = UIManager.getInstalledLookAndFeels();	
		
		// erase disallowed look and feels:
	    List<LookAndFeelInfo> result = new LinkedList<LookAndFeelInfo>();
	    for (int i = 0; i < this.disallowedLafs.length; i++) {
	    	for(UIManager.LookAndFeelInfo item : lafs) {
	    		if(!disallowedLafs[i].equals(item.getName())) {
	    			result.add(item);
	    		}
	    	}
	    }

	    return result.toArray(new UIManager.LookAndFeelInfo[0]);
	}
	
	/** Updates the look and feel combobox.
	 * 
	 * @param lafClassName Entry to select
	 */
	private void setLookAndFeelCombo(String lafClassName) {
	    // UIManager.LookAndFeelInfo[] lafs = UIManager.getInstalledLookAndFeels();	
		UIManager.LookAndFeelInfo[] lafs = getAllowedLookAndFeels();	
				
		for(int i = 0; i < lafs.length; i++) {
			if(lafs[i].getClassName().equals(lafClassName)) {
				lafCombo.setSelectedIndex(i);
				break;
			}
		}	
		
		Settings.getSingletonInstance().setLookAndFeelEverywhere(lafClassName);
	}
	
	/** Loads the settings.
	 * 
	 * Reads the informations from
	 * {@link org.helioviewer.jhv.Settings} and sets all gui elements
	 * according to them.
	 */
	private void loadSettings() {
		
		// In principle the settings have been previously loaded
		// settings.load();
		
		// The current cache size
		occupiedSizeLabel.setText(getCacheSizeText());
		
		// Look and feel
		setLookAndFeelCombo(settings.getProperty("look.and.feel"));
		
		// OpenGL options
		if(Boolean.parseBoolean(settings.getProperty("opengl.enabled"))) {
			openGLItemOn.setSelected(true);
			openGLItemOn.doClick();
		} else {
			openGLItemOff.setSelected(true);
			openGLItemOff.doClick();
		}
		
		// Debug options
		debugItemOn.setSelected(Boolean.parseBoolean(settings.getProperty("debug.log.enabled")));
		timerItemOn.setSelected(Boolean.parseBoolean(settings.getProperty("debug.log.times")));
		
		if(debugItemOn.isSelected()) debugItemOn.doClick();
		else debugItemOff.doClick();
		
		// Default date format
		String fmt = settings.getProperty("default.date.format");
		
		if (fmt == null)
		    dateFormatField.setText(defaultDateFormat);
		else
		    dateFormatField.setText(fmt);
		
		// Default values
		defaultsPanel.loadSettings();
		
		// Maximum JPIP cache size
		maxCacheBox.setText(settings.getProperty("jpip.cache.size"));
		
		try {
		    limitMaxSize.setSelected(Double.parseDouble(maxCacheBox.getText()) > 0);
		} catch (NumberFormatException e) {
		    maxCacheBox.setText("0.0");
		    limitMaxSize.setSelected(false);
		}
		
		maxCacheBox.setVisible(limitMaxSize.isSelected());
		maxCacheBoxLabel.setVisible(limitMaxSize.isSelected());
	}
	
	/** Saves the settings.
	 * 
	 * Writes the informations to {@link org.helioviewer.jhv.Settings}.
	 */
	private void saveSettings() {
		
		// Look and feel
		//UIManager.LookAndFeelInfo[] lafs = UIManager.getInstalledLookAndFeels();	
		UIManager.LookAndFeelInfo[] lafs = getAllowedLookAndFeels();	
		settings.setProperty("look.and.feel", lafs[lafCombo.getSelectedIndex()].getClassName());
		
		// OpenGL options
		settings.setProperty("opengl.enabled", Boolean.toString(openGLItemOn.isSelected()));
		
		// Debug options
		settings.setProperty("debug.log.enabled", Boolean.toString(debugItemOn.isSelected()));
		settings.setProperty("debug.log.times", Boolean.toString(timerItemOn.isSelected()));
		
		// Default date format
		settings.setProperty("default.date.format", dateFormatField.getText());
		
		// Default values
		defaultsPanel.saveSettings();
		
		// Maximum JPIP cache size
		if (limitMaxSize.isSelected())
		    settings.setProperty("jpip.cache.size", maxCacheBox.getText());
		else
		    settings.setProperty("jpip.cache.size", "0.0");
		
		// Update and save settings
		settings.update();
		settings.save();
	}
	
	/** Builds the string showing the size of the cache currently used.
	 * 
	 * @return String showing the size of the cache currently used
	 */
	private String getCacheSizeText() {
		
		long len = 0;
		File[] list = JHV_Kdu_cache.getCacheFiles(JHVDirectory.CACHE.getFile());
		
		for(File f: list) 
			len += f.length();
		
		if (len < 1024) 
		    return ("Occupied size: " + len + " bytes");
		else if (len < 1048576) 
		    return ("Occupied size: " + ((double)Math.round(((double)len / 1024.0) * 100.0) / 100.0) + " Kbytes");
		else 
		    return ("Occupied size: " + ((double)Math.round(((double)len / 1048576.0) * 100.0) / 100.0) + " Mbytes");
	}
	
	/** Creates the JPIP cache panel.
	 * 
	 * @return JPIP cache panel
	 */
	private JPanel createJPIPCachePanel() {
	    
	    JPanel cachePanel = new JPanel(new GridLayout(0, 1));
		cachePanel.setBorder(BorderFactory.createTitledBorder(" JPIP Cache "));
				
		// cache location
		JPanel row1 = new JPanel(new FlowLayout(FlowLayout.LEADING));
		row1.add(new JLabel("Cache location: " + JHVDirectory.CACHE.getFile().getAbsolutePath()));
		
		// maximum cache size
		maxCacheBox.setPreferredSize(new Dimension(80, maxCacheBox.getPreferredSize().height));
		
		limitMaxSize.addActionListener(new ActionListener() {
            
            public void actionPerformed(ActionEvent e) {                
                maxCacheBox.setVisible(limitMaxSize.isSelected());
                maxCacheBoxLabel.setVisible(limitMaxSize.isSelected());
            }
        });
		
		JPanel row2 = new JPanel(new FlowLayout(FlowLayout.LEADING));
		row2.add(new JLabel("Cache size: "));
		row2.add(limitMaxSize);
		row2.add(maxCacheBox);
		row2.add(maxCacheBoxLabel);
		
		// used cache size
		JPanel row3 = new JPanel(new FlowLayout(FlowLayout.LEADING));
		row3.add(occupiedSizeLabel);
		
		cachePanel.add(row1);
		cachePanel.add(row2);
		cachePanel.add(row3);
		
		return cachePanel;
	}	

	/** Creates the general parameters panel.
	 * 
	 * @return General parameters panel
	 */
	private JPanel createParametersPanel() {
		paramsPanel = new JPanel();
		
		paramsPanel.setBorder(BorderFactory.createTitledBorder(" Configuration "));
		paramsPanel.setLayout(new GridLayout(0, 1));
		
		//UIManager.LookAndFeelInfo[] lafs = UIManager.getInstalledLookAndFeels();
		UIManager.LookAndFeelInfo[] lafs = getAllowedLookAndFeels();	

		String[] lafNames = new String[lafs.length];
		for(int i = 0; i < lafs.length; i++) {
			lafNames[i] = lafs[i].getName();
		}
		lafCombo = new JComboBox(lafNames);
		
		final PreferencesDialog parent = this;
		lafCombo.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				JComboBox source = (JComboBox)e.getSource();
				//UIManager.LookAndFeelInfo[] lafs = UIManager.getInstalledLookAndFeels();
				UIManager.LookAndFeelInfo[] lafs = getAllowedLookAndFeels();	
        		UIManager.LookAndFeelInfo selectedLaf = lafs[source.getSelectedIndex()];
        		
        		Settings.getSingletonInstance().setLookAndFeelEverywhere(selectedLaf.getClassName());
        		SwingUtilities.updateComponentTreeUI(parent);
			}
		});
		
		JPanel row1 = new JPanel(new FlowLayout(FlowLayout.LEADING));
		row1.add(new JLabel("Look and Feel:  "));
		row1.add(lafCombo);
		
		dateFormatField = new JTextField();
        dateFormatField.setPreferredSize(new Dimension(150, 23));
        
        JPanel row2 = new JPanel(new FlowLayout(FlowLayout.LEADING));
        row2.add(new JLabel("Default date format:  "));        
        row2.add(dateFormatField);
        
        Icon infoIcon = IconBank.getIcon(JHVIcon.INFO);
        
        dateFormatInfo = new JButton(infoIcon);
        dateFormatInfo.setBorder(BorderFactory.createEtchedBorder());
        dateFormatInfo.setPreferredSize(new Dimension(infoIcon.getIconWidth()+5,23));
        dateFormatInfo.setToolTipText("Show possible date format information");
        dateFormatInfo.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                DateFormatInfoDialog dialog = new DateFormatInfoDialog();
                dialog.showDialog();
            }
        });
        
        row2.add(dateFormatInfo);
		
		JPanel row3 = new JPanel(new FlowLayout(FlowLayout.LEADING));
		row3.add(new JLabel("Use OpenGL, if available: "));
		
		openGLItemOff = new JRadioButton("Off", false);
		openGLItemOn = new JRadioButton("On", true);
		
		ButtonGroup openGlGroupRadio = new ButtonGroup();
		openGlGroupRadio.add(openGLItemOff);
		openGlGroupRadio.add(openGLItemOn);
		
		row3.add(openGLItemOff);
		row3.add(openGLItemOn);

	    debugItemOff = new JRadioButton("Off", true);
		debugItemOn = new JRadioButton("On", false);
	    timerItemOff = new JRadioButton("Off", true);
		timerItemOn = new JRadioButton("On", false);
		
		JPanel row4 = new JPanel(new FlowLayout(FlowLayout.LEADING));
		row4.add(new JLabel("Write log files: "));
		
		ButtonGroup debugGroupRadio = new ButtonGroup();
		debugGroupRadio.add(debugItemOff);
		debugGroupRadio.add(debugItemOn);
		
		row4.add(debugItemOff);
		row4.add(debugItemOn);
		
		final JPanel row5 = new JPanel(new FlowLayout(FlowLayout.LEADING));
		row5.add(new JLabel("Timing values: "));
		
		ButtonGroup timerGroupRadio = new ButtonGroup();
		timerGroupRadio.add(timerItemOff);
		timerGroupRadio.add(timerItemOn);
		
		row5.add(timerItemOff);
		row5.add(timerItemOn);
		
		paramsPanel.add(row1);
		paramsPanel.add(row2);
		paramsPanel.add(row3);
		paramsPanel.add(row4);
		
		debugItemOn.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				paramsPanel.add(row5);
				pack();
				repaint();
			}		
		});
		
		debugItemOff.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				timerItemOn.setSelected(false);
				timerItemOff.setSelected(true);
				
				paramsPanel.remove(row5);
			}		
		});
		
		return paramsPanel;
	}	

	/** Creates the default save directories panel.
	 * 
	 * @return Default save directories panel
	 */
	private JPanel createDefaultSaveDirPanel() {
		
		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(BorderFactory.createTitledBorder(" Defaults "));
		
		defaultsPanel = new DefaultsSelectionPanel();
		defaultsPanel.setPreferredSize(new Dimension(450, 100));		
		defaultsPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		
		panel.add(defaultsPanel, BorderLayout.CENTER);
		
		return panel;
	}
	
	private class DefaultsSelectionPanel extends JPanel {
		
		private static final long serialVersionUID = 1L;
		
		private JTable table = null;
		private Object[][] tableData = null;
			
		
		public DefaultsSelectionPanel() {
			
			super(new BorderLayout());		
			setPreferredSize(new Dimension(150, 180));
			
			Settings settings = Settings.getSingletonInstance();
			
			tableData = new Object[][] {
					{"Default save directory", settings.getProperty("default.save.path")},
					{"Default local path", settings.getProperty("default.local.path")},
					{"Default remote path", settings.getProperty("default.remote.path")}
			};
			
			table = new JTable(new DefaultTableModel(tableData, new String[] {"Description", "Value"}) {  
				private static final long serialVersionUID = 1L;

				public boolean isCellEditable(int row, int column) {  
			        return ((row == 2) && (column == 1));
			    }  
			});
			
			table.setRowHeight(20);
			JScrollPane scrollPane = new JScrollPane(table);
			table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			//table.setFillsViewportHeight(true);
			
			table.addMouseListener(new MouseAdapter() {
				public void mouseClicked(MouseEvent e) {
					if(e.getClickCount() != 2) return;
					
					int row = table.getSelectedRow();
					if(row >= 2) return;
					
					JFileChooser 	chooser = new JFileChooser((String) table.getModel().getValueAt(row, 1));
					chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
						
					if(chooser.showDialog(null, "Select") == JFileChooser.APPROVE_OPTION) 
						table.getModel().setValueAt(chooser.getSelectedFile().toString(), row, 1);
				}
			});
			
			TableColumn col = table.getColumnModel().getColumn(0);
			col.setMaxWidth(150);
			col.setMinWidth(150);
			    
			add(scrollPane, BorderLayout.CENTER);
		}
		
		public void loadSettings() {
			
	        TableModel model = table.getModel();

			Settings settings = Settings.getSingletonInstance();
			
			model.setValueAt(settings.getProperty("default.save.path"), 0, 1);
			model.setValueAt(settings.getProperty("default.local.path"), 1, 1);
			model.setValueAt(settings.getProperty("default.remote.path"), 2, 1);			
		}
		
		public void saveSettings() {
			
	        TableModel model = table.getModel();

			Settings settings = Settings.getSingletonInstance();
			
			settings.setProperty("default.save.path", model.getValueAt(0, 1).toString());
			settings.setProperty("default.local.path", model.getValueAt(1, 1).toString());								
			settings.setProperty("default.remote.path", model.getValueAt(2, 1).toString());			
		}
		
		public void resetSettings() {
			
	        TableModel model = table.getModel();
			
			model.setValueAt(JHVDirectory.EXPORTS.getPath(), 0, 1);
			model.setValueAt(JHVDirectory.HOME.getPath(), 1, 1);
			model.setValueAt("jpip://delphi.nascom.nasa.gov:8090", 2, 1);				
		}
	}
}
