package org.helioviewer.jhv.internal_plugins.filter.SOHOLUTFilterPlugin;

import java.nio.IntBuffer;
import java.util.Collection;
import java.util.LinkedList;

import javax.media.opengl.GL;

import org.helioviewer.viewmodel.filter.FilterListener;
import org.helioviewer.viewmodel.filter.GLFragmentShaderFilter;
import org.helioviewer.viewmodel.filter.ObservableFilter;
import org.helioviewer.viewmodel.filter.StandardFilter;
import org.helioviewer.viewmodel.imagedata.ARGBInt32ImageData;
import org.helioviewer.viewmodel.imagedata.ImageData;
import org.helioviewer.viewmodel.imageformat.SingleChannelImageFormat;
import org.helioviewer.viewmodel.imagetransport.Byte8ImageTransport;
import org.helioviewer.viewmodel.imagetransport.Short16ImageTransport;
import org.helioviewer.viewmodel.view.opengl.GLTextureHelper;
import org.helioviewer.viewmodel.view.opengl.shader.GLShaderBuilder;
import org.helioviewer.viewmodel.view.opengl.shader.GLSingleChannelLookupFragmentShaderProgram;

/** Filter for applying a color table to a single channel image.
 * 
 * <p>If the input image is not a single channel image, the filter
 * does nothing and returns the input data.
 * 
 * <p>This filter supports software rendering as well as rendering
 * in OpenGL.
 *
 */
public class SOHOLUTFilter implements StandardFilter, GLFragmentShaderFilter, ObservableFilter {

	// /////////////////////////
	//         GENERAL        //
	// /////////////////////////
    private SOHOLUT.LUT lut;
    private Collection<FilterListener> listeners = new LinkedList<FilterListener>();
    
    /**
     * {@inheritDoc}
     * 
     * <p>This filter is a major filter.
     */
    public boolean isMajorFilter() {
        return true;
    }
    
    /** Default constructor.
     * 
     * Sets the color table to GRAY.
     */
    public SOHOLUTFilter() {
    	lut = SOHOLUT.LUT.GRAY;
    }
    
    /** Constructor setting the color table.
     * 
     * @param startWithLut Color table to apply to the image
     */
    public SOHOLUTFilter(SOHOLUT.LUT startWithLut) {
    	lut = startWithLut;
    }

    /**
     * {@inheritDoc}
     */
    public void addFilterListener(FilterListener l) {
    	lastLut = null;
        listeners.add(l);
    }

    /**
     * {@inheritDoc}
     */
    public void removeFilterListener(FilterListener l) {
        listeners.remove(l);
    }

    /** 
     * Notifies all registered listeners, that something has changed.
     */
    private void notifyAllListeners() {
        for (FilterListener f : listeners) {
            f.filterChanged(this);
        }
    }

    /** Returns the color table currently used.
     * 
     * @return Color table currently used
     */
    public SOHOLUT.LUT getLUT() {
        return lut;
    }

    /** Sets a new color table to use from now on.
     * 
     * @param newLUT New color table
     */
    public void setLUT(SOHOLUT.LUT newLUT, boolean invert) {
        if (lut == newLUT && lut.isInverted() == invert) {
            return;
        }
        lut = newLUT;
        
        if(lut.isInverted() != invert) {
        	lut.invertLUT();
        }
        
        notifyAllListeners();
    }

    // /////////////////////////
	//         STANDARD       //
	// /////////////////////////
    
    /**
     * {@inheritDoc}
     */
    public ImageData apply(ImageData data) {
    	if (data == null
    	|| !(data.getImageFormat() instanceof SingleChannelImageFormat)
    	|| lut == SOHOLUT.LUT.GRAY) {
            return data;
        }
        
        if (data.getImageTransport() instanceof Byte8ImageTransport) {
            byte[] pixelData = ((Byte8ImageTransport) data.getImageTransport())
                    .getByte8PixelData();
            int[] resultPixelData = new int[pixelData.length];
            SOHOLUT.lookup(pixelData, resultPixelData, lut);
            return new ARGBInt32ImageData(data.getWidth(), data.getHeight(),
                    resultPixelData);
            
        } else if (data.getImageTransport() instanceof Short16ImageTransport) {

            short[] pixelData = ((Short16ImageTransport) data
                    .getImageTransport()).getShort16PixelData();
            int[] resultPixelData = new int[pixelData.length];
            SOHOLUT16.lookup16(pixelData, resultPixelData, lut);
            pixelData = new short[1];
            data = new ARGBInt32ImageData(data.getWidth(), data.getHeight(),
                    resultPixelData);
            return data;
        }
        
        return null;
    }

    // /////////////////////////
	//         OPENGL         //
	// /////////////////////////
    private GLSingleChannelLookupFragmentShaderProgram shader = new GLSingleChannelLookupFragmentShaderProgram();
    private int lookupTex = 0;
    private SOHOLUT.LUT lastLut = null;
    private boolean lastInverted = false;
    
    /**
     * {@inheritDoc}
     */
    public GLShaderBuilder buildFragmentShader(GLShaderBuilder shaderBuilder) {
		shader.build(shaderBuilder);
		
		if(lastLut == null) {
			GLTextureHelper textureHelper = new GLTextureHelper();
			textureHelper.delTextureID(shaderBuilder.getGL(), lookupTex);
			lookupTex = textureHelper.genTextureID(shaderBuilder.getGL());
		}
		
		return shaderBuilder;
	}
    
    /**
     * {@inheritDoc}
     * 
     * In this case, also updates the color table, if necessary.
     */
    public void applyGL(GL gl) {
		
		shader.bind(gl);
		shader.activateLutTexture(gl);
		
		// Note: The lookup table will always be power of two,
		// so we won't get any problems here.
		
		gl.glBindTexture(GL.GL_TEXTURE_1D, lookupTex);
		
		if(lastLut != lut || lut.isInverted() != lastInverted) {
			IntBuffer buffer = IntBuffer.wrap(lut.getLUT());
			
			gl.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 4);
			
			gl.glTexImage1D(GL.GL_TEXTURE_1D, 0, GL.GL_RGBA, lut.getLUT().length, 0, GL.GL_BGRA, GL.GL_UNSIGNED_INT_8_8_8_8_REV, buffer);
			gl.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST);
	        gl.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST);
	        gl.glTexParameteri(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP);
			
			lastLut = lut;
			lastInverted = lut.isInverted();
		}
	}
	
	protected void finalize() {
		if(lookupTex != 0) {
			GLTextureHelper textureHelper = new GLTextureHelper();
			textureHelper.delTextureID(null, lookupTex);
		}
	}
}
