package org.helioviewer.jhv.internal_plugins.filter.opacity;

import java.awt.AlphaComposite;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.util.LinkedList;

import javax.media.opengl.GL;

import org.helioviewer.viewmodel.filter.FilterListener;
import org.helioviewer.viewmodel.filter.GLFragmentShaderFilter;
import org.helioviewer.viewmodel.filter.ObservableFilter;
import org.helioviewer.viewmodel.filter.StandardFilter;
import org.helioviewer.viewmodel.imagedata.ARGBInt32ImageData;
import org.helioviewer.viewmodel.imagedata.ImageData;
import org.helioviewer.viewmodel.imagedata.JavaBufferedImageData;
import org.helioviewer.viewmodel.view.opengl.shader.GLFragmentShaderProgram;
import org.helioviewer.viewmodel.view.opengl.shader.GLShaderBuilder;
import org.helioviewer.viewmodel.view.opengl.shader.GLTextureCoordinate;
import org.helioviewer.viewmodel.view.opengl.shader.GLShaderBuilder.GLBuildShaderException;

/** Filter for changing the opacity of an image.
 * 
 * <p>The output of the filter always is an ARGB image,
 * since that is currently the only format supporting
 * an alpha channel. Thus, this filter should be applied
 * as late as possible.
 * 
 * <p>This filter supports software rendering as well as
 * rendering in OpenGL.
 * 
 * @author Markus Langenberg
 *
 */
public class OpacityFilter implements StandardFilter, GLFragmentShaderFilter, ObservableFilter {
	
    // ////////////////////////////////////////////////////////////////
    // Definitions
    // ////////////////////////////////////////////////////////////////
    
	private LinkedList<FilterListener> listeners = new LinkedList<FilterListener>();
	private float opacity;
	private OpacityShader shader = new OpacityShader();
	private OpacityPanel panel;

	// ////////////////////////////////////////////////////////////////
    // Methods
    // ////////////////////////////////////////////////////////////////
	
	public OpacityFilter(float initialOpacity) {
		opacity = initialOpacity;
	}
	
    /**
     * Sets the corresponding opacity panel.
     * 
     * @param panel
     *            Corresponding panel.
     */
	public void setOpacityPanel(OpacityPanel panel) {
	    this.panel = panel;
	    panel.setSliderValue((int) (opacity * 100));
	}
	
	/**
     * {@inheritDoc}
     * 
     * <p>This filter is a major filter.
     */
    public boolean isMajorFilter() {
	    return true;
	}
	
    /** Sets the opacity.
     * 
     * This function also updates the slider and is supposed to be
     * called from external sources.
     * 
     * @param newOpacity New opacity, value has to be within [0, 1]
     */
	public void setOpacityExternal(float newOpacity) {
	    opacity = newOpacity;
	    
	    panel.setSliderValue((int) (opacity * 100));
	}
	
	/** Sets the opacity.
     * 
     * This function does not the slider, thus should only be called
     * by the slider itself. Otherwise, use {@link #setOpacityExternal(float)}.
     * 
     * @param newOpacity New opacity, value has to be within [0, 1]
     */
	void setOpacity(float newOpacity) {
		opacity = newOpacity;
		notifyAllListeners();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public ImageData apply(ImageData data) {
		if (data == null) {
            return null;
        }
		
		if(opacity > 0.999f)
			return data;
		
		if(data instanceof JavaBufferedImageData) {
			BufferedImage source = ((JavaBufferedImageData) data).getBufferedImage();
			BufferedImage target = new BufferedImage(source.getWidth(), source.getHeight(), BufferedImage.TYPE_INT_ARGB);
			
			Graphics2D g = target.createGraphics();
			g.setComposite(AlphaComposite.Clear);
			g.fillRect(0, 0, data.getWidth(), data.getHeight());
			g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC, opacity));
			g.drawImage(source, 0, 0, null);
			g.dispose();
			
			return new ARGBInt32ImageData(target);
		}
		return null;
	}
	
	/** 
	 * Fragment shader setting the opacity.
	 */
	private class OpacityShader extends GLFragmentShaderProgram {
		private GLTextureCoordinate alphaParam;
		
		/** Sets the new alpha value.
		 * 
		 * @param gl Valid reference to the current gl object
		 * @param alpha Alpha value
		 */
		private void setAlpha(GL gl, float alpha) {
			if(alphaParam != null) {
				alphaParam.setValue(gl, alpha);
			}
		}
		
		/**
		 * {@inheritDoc}
		 */
		protected void buildImpl(GLShaderBuilder shaderBuilder) {
			try {
				alphaParam = shaderBuilder.addTexCoordParameter(1);
				String program = "\toutput.a = output.a * alpha;";
				program = program.replace("output", shaderBuilder.useOutputValue("float4", "COLOR"));
				program = program.replace("alpha", alphaParam.getIdentifier());
				shaderBuilder.addMainFragment(program);
			} catch (GLBuildShaderException e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public GLShaderBuilder buildFragmentShader(GLShaderBuilder shaderBuilder) {
		shader.build(shaderBuilder);
		return shaderBuilder;
	}

	/**
	 * {@inheritDoc}
	 */
	public void applyGL(GL gl) {
		shader.bind(gl);
		shader.setAlpha(gl, opacity);
	}

	/**
	 * {@inheritDoc}
	 */
	public void addFilterListener(FilterListener l) {
        listeners.add(l);
    }

	/**
	 * {@inheritDoc}
	 */
	public void removeFilterListener(FilterListener l) {
        listeners.remove(l);
    }

	/** 
     * Notifies all registered listeners, that something has changed.
     */
    private void notifyAllListeners() {
        for (FilterListener f : listeners) {
            f.filterChanged(this);
        }
    }
}
