package org.helioviewer.jhv.internal_plugins.overlay.solarevents;

import org.json.JSONException;
import org.json.JSONObject;

/**
 * The class represents a solar event and manages all the associated
 * information.
 * 
 * <p>For further informations, also see {@link SolarEventCatalog}.
 * 
 * @author Stephan Pagel
 * */
public class SolarEvent {

	// ///////////////////////////////////////////////////////////////////////////
	// Definitions
	// ///////////////////////////////////////////////////////////////////////////
		
	private JSONObject eventObject;
		
	// ///////////////////////////////////////////////////////////////////////////
	// Methods
	// ///////////////////////////////////////////////////////////////////////////
	
    /***
     * Default Constructor.
     * 
     * @param obj
     *            JSONObject which contains the information of this solar event.
     */
	public SolarEvent(JSONObject obj) {
	    eventObject = obj;
    }
	
    /**
     * Returns the beginning of the solar event period.
     * 
     * @return time stamp of the beginning of the solar event period.
     * @throws SolarEventException
     */
	public String getStartTime() throws SolarEventException {
		return getStringFromCategory("time", "startTime");
	}
	
    /**
     * Returns the end of the solar event period.
     * 
     * @return time stamp of the end of the solar event period.
     * @throws SolarEventException
     */
	public String getEndTime() throws SolarEventException {
		return getStringFromCategory("time", "endTime");
	}
	
	/**
	 * Returns the id of the associated solar event catalog.
	 * 
	 * @return id of the associated solar event catalog.
	 * @throws SolarEventException
	 */
	public String getCatalogID() throws SolarEventException {
		return getString("catalogId");
	}
	
	/**
	 * Returns the short information about the solar event.
	 * 
	 * @return short information about the solar event.
	 * @throws SolarEventException
	 */
	public String getShortInfo() throws SolarEventException {
		return getString("shortInfo");
	}
	
	/**
	 * Returns the source URL of the solar event.
	 * 
	 * @return source URL of the solar event.
	 * @throws SolarEventException
	 */
	public String getSourceURL() throws SolarEventException {
		return getString("sourceUrl");
	}
	
    /**
     * Returns an array which contains all available property names of the solar
     * event.
     * 
     * @return array with all available property names of the solar event.
     * @throws SolarEventException
     */
	public String[] getPropertyNames() throws SolarEventException {
		return getNamesOfCategory("properties");
	}
	
    /**
     * Returns an array which contains all available category names of the solar
     * event.
     * 
     * @return array with all available category names of the solar event.
     */
	public String[] getCategories() {
		return getNames();
	}
	
    /**
     * Returns the value of a property as String.
     * 
     * @param name
     *            read the value from this property.
     * @return value of passed property.
     * @throws SolarEventException
     */
	public String getString(String name) throws SolarEventException {
		
		try {
			return eventObject.getString(name);
		} catch (JSONException e) {
			throw new SolarEventException();
		}
	}
	
    /**
     * Returns the value of a property as Double.
     * 
     * @param name
     *            read the value from this property.
     * @return value of passed property.
     * @throws SolarEventException
     */
	public double getDouble(String name) throws SolarEventException {
		
		try {
			return eventObject.getDouble(name);
		} catch (JSONException e) {
			throw new SolarEventException();
		}
	}

    /**
     * Returns the value of a property as String which is integrated in a
     * category.
     * 
     * @param category
     *            category where property is integrated.
     * @param name
     *            read the value from this property.
     * @return value of passed property.
     * @throws SolarEventException
     */
	public String getStringFromCategory(String category, String name) throws SolarEventException {
		
		try {
			return eventObject.getJSONObject(category).getString(name);
		} catch (JSONException e) {
			throw new SolarEventException();
		}
	}
	
	/**
	 * Returns all property names.
	 * 
	 * @return property names.
	 */
	private String[] getNames() {
		return JSONObject.getNames(eventObject);
	}
	
    /**
     * Returns all property names of the given category.
     * 
     * @param category
     *            get property names of this category.
     * @return all property names of the given category.
     * @throws SolarEventException
     */
	private String[] getNamesOfCategory(String category) throws SolarEventException {
	    
		try {
			if (!eventObject.isNull(category)) {
				return JSONObject.getNames(eventObject.getJSONObject(category));
			}
		} catch (JSONException e) {
			throw new SolarEventException();
		}
		
		return null;
	}
	
    /**
     * Exception class for exceptions which can occur inside the solar event
     * class. Exceptions which occurred from the internal structure of the solar
     * event class should be mapped to this exception class.
     * 
     * @author Stephan Pagel
     */
	public class SolarEventException extends Exception {

        private static final long serialVersionUID = 1L;
	} 
}
