package org.helioviewer.jhv.internal_plugins.overlay.solarevents;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.text.SimpleDateFormat;
import java.util.AbstractList;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.TimeZone;

import org.helioviewer.jhv.Settings;
import org.helioviewer.jhv.gui.components.StatusPanel;
import org.helioviewer.viewmodel.view.ImageInfoView;
import org.helioviewer.viewmodel.view.MovieView;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.jp2view.datetime.ImmutableDateTime;
import org.json.JSONArray;
import org.json.JSONException;

/**
 * This class represents a solar event catalog which provides information about
 * the catalog and manages all associated solar events.
 * <p>
 * When solar event are downloaded they will be stored in an internal cache so
 * they do not have to be downloaded again (for each image info layer!). Because
 * of the current definition of the server API it is not possible to download
 * solar event information for a period in one step. For each date a new request
 * has to be send to the server.
 * <p>
 * The concept of the internal cache and the way solar event information are
 * downloaded has to be revise when the API supports a possibility to download
 * solar event information in one step!
 * 
 * @author Stephan Pagel
 * */
public class SolarEventCatalog {
	
	// ///////////////////////////////////////////////////////////////////////////
	// Definitions
	// ///////////////////////////////////////////////////////////////////////////
	
    /*
     * The map offers the possibility for caching solar events for a faster
     * access. The map provides the connection between the image info view and
     * all solar events against there time stamps. The value of the map provides
     * internally the connection between the time stamp and the solar event
     * information.
     */
	private HashMap<ImageInfoView, CacheEventValue> cacheEvents = new HashMap<ImageInfoView, CacheEventValue>();
	
	// list of all solar events which belong to the catalog
	private AbstractList<SolarEvent> solarEvents = new LinkedList<SolarEvent>();
	
	// The coordinate system used by the event catalog
	private String coordinateSystem;
	
	// A description of the event catalog
	private String description;
	
	// The event type of the catalog
	private String eventType;
	
	// The ID of the catalog
	private String id;
	
	// The name of the catalog
	private String name;
	
	// Date of creation of this descriptor
	private Date creationDate;
	
	// Flag if solar events of event catalog should be displayed or not
	private boolean eventsVisible = false;
	
	// ///////////////////////////////////////////////////////////////////////////
	// Methods
	// ///////////////////////////////////////////////////////////////////////////
	
	/**
	 * Default constructor.
	 * 
	 * @param coordinateSystem Coordinate system for solar event catalog.
	 * @param description Description for solar event catalog.
	 * @param eventType Event type for solare event catalog.
	 * @param id ID of solar event catalog.
	 * @param name Name for solar event catalog.
	 * */
	public SolarEventCatalog(String coordinateSystem, String description, String eventType, String id, String name) {
	    this.coordinateSystem = coordinateSystem;
	    this.description = description;
	    this.eventType = eventType;
	    this.id = id;
	    this.name = name;
    }
	
    /**
     * Returns a date as String how it is defined by the API on the server.
     * 
     * @param date
     *            Date to format.
     * @return formatted date.
     * */
	private String getDateStringForAPI(Date date) {
		SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
		sdf.setTimeZone(TimeZone.getTimeZone("GMT"));
		return sdf.format(date);
	}
	
    /**
     * Load all events of the catalog for a specific date and image info view.
     * If the solar events are not in the internal cache they will be downloaded
     * from the server.
     * 
     * @param view
     *            Corresponding image info view.
     * @param date
     *            Date for which solar events are needed.
     */
	public void loadCatalogEvents(ImageInfoView view, ImmutableDateTime date) {
		
		// check if events are available in cache for specific date
		boolean containsView = cacheEvents.containsKey(view);
		boolean containsDate = false;
		
		if (containsView)
			containsDate = cacheEvents.get(view).solarEventsForDateInCache(date);
		
		if (containsView && containsDate) {
		    // solar events are available in cache
			loadCatalogEventsFromCache(view, date);
		} else {
		    
            // TODO SP: the following if block is just temporary. It prevents
            // that solar events are loaded for each frame when a movie is
            // running.
		    if (view instanceof MovieView && ((MovieView) view).isMoviePlaying()) {
	            
	            solarEvents = new LinkedList<SolarEvent>();
	            return;
	        }
		    
			// download solar catalog events
			StatusPanel.setStatusInfoText("Loading solar events");			
			loadCatalogEventsRemote(date.getTime());			
			StatusPanel.setStatusInfoText("");
			
			// add loaded catalog events to cache
			addCatalogEventsToCache(view, date, solarEvents);
		}
	}
	
    /**
     * Tries to download the solar events for a given date from the server.
     * 
     * @param date
     *            Date for which solar events are requested.
     */
	private void loadCatalogEventsRemote(Date date) {
		// reinitialize event list
		solarEvents = new LinkedList<SolarEvent>();
		
		// initialize local variables
		String response = null;
		
		try {
	        // Construct data
	        String data = URLEncoder.encode("action", "UTF-8") + "=" + "getEvents";
	        data += "&" + URLEncoder.encode("date", "UTF-8") + "=" + getDateStringForAPI(date);
	        data += "&" + URLEncoder.encode("windowSize", "UTF-8") + "=" + (id.equals("VSOService::noaa") || id.equals("EGSOService::halpha_flares_event") ? "86400" : "43200");
	        data += "&" + URLEncoder.encode("catalogs", "UTF-8") + "=" + id;

	        // Send data
	        URL url = new URL(Settings.getSingletonInstance().getProperty("API.event.path"));
	        URLConnection conn = url.openConnection();
	        conn.setDoOutput(true);
	        OutputStreamWriter wr = new OutputStreamWriter(conn.getOutputStream());
	        wr.write(data);
	        wr.flush();
	    
	        // Get the response
	        BufferedReader rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
	        response = rd.readLine();
	        wr.close();
	        rd.close();
	    } catch (Exception e) {
	    	e.printStackTrace();
	    }
	    
	    if(response != null && response != "") {
		    try {
		    	JSONArray events = new JSONArray(response);
		    	
		    	for (int i = 0; i < events.length(); ++i) {
		    		solarEvents.add(new SolarEvent(events.getJSONObject(i)));
		    	}
		    	
		    } catch (JSONException e) {
		    	e.printStackTrace();
		    }
		}
	}
	
	/**
	 * Gets the solar events for a given date from the internal cache.
	 * 
     * @param view
     *            Corresponding image info view.
     * @param date
     *            Date for which solar events are needed.
	 */
	private void loadCatalogEventsFromCache(View view, ImmutableDateTime date) {
		solarEvents = cacheEvents.get(view).getSolarEvents(date);
	}
	
    /**
     * Adds solar event objects for a given date and image info view to the
     * internal cache.
     * 
     * @param view
     *            Corresponding image info view.
     * @param date
     *            Associated date for solar events.
     * @param events
     *            List of events which have to be stored in the internal cache.
     * */
	private void addCatalogEventsToCache(ImageInfoView view, ImmutableDateTime date, AbstractList<SolarEvent> events) {
		
		if (cacheEvents.containsKey(view)) {
			// if key already exists in cache remember events with associated time
			cacheEvents.get(view).add(date, events);
		} else {
			// if key doesn't already exist in cache add it to the hash
			CacheEventValue value = new CacheEventValue();
			value.add(date, events);
			
			cacheEvents.put(view, value);
		}
	}
	
    /**
     * Removes all solar events which belong to the given image info view.
     * 
     * @param view
     *            image remove all solar events from internal cache for this
     *            image info view.
     */
	public void removeCatalogEventsFromCache(ImageInfoView view) {
		
		if (cacheEvents.containsKey(view))
			cacheEvents.remove(view);
	}
	
	/**
	 * Clears the hole internal cache.
	 */
	public void clearCatalogEventsCache() {
		cacheEvents.clear();
	}
	
	/**
	 * Returns a list of all available solar event objects.
	 * 
	 * @return list of all available solar event objects.
	 * */
	public AbstractList<SolarEvent> getEvents() {
		return solarEvents;
	}
	
    /**
     * Overridden toString method which returns the name of the solar event
     * catalog.
     * 
     * @return name of the solar event catalog.
     * */
	@Override
	public String toString() {
	    return name;
	}
	
    /**
     * Sets the visibility flag of the solar event catalog. The visible is used
     * to identify which solar event catalogs are currently displayed in the
     * solar event catalog panel.
     * 
     * @param visible
     *            boolean value if catalog is visible or not.
     * */
	public void setVisibility(boolean visible) {
		eventsVisible = visible;
	}
	
	/**
	 * Returns the value of the visibility flag.
	 * 
	 * @return value of the visibility flag.
	 * @see #setVisibility(boolean)
	 * */
	public boolean isVisible() {
		return eventsVisible;
	}
	
	/**
	 * Returns the coordinate system for solar event catalog.
	 * 
	 * @return coordinate system for solar event catalog.
	 * */
	public String getCoordinateSystem() {
		return coordinateSystem;
	}

	/**
     * Returns the description for solar event catalog.
     * 
     * @return description for solar event catalog.
     * */
	public String getDescription() {
		return description;
	}

	/**
     * Returns the event type for solar event catalog.
     * 
     * @return event type for solar event catalog.
     * */
	public String getEventType() {
		return eventType;
	}	
	
	/**
     * Returns the id for solar event catalog.
     * 
     * @return id for solar event catalog.
     * */
	public String getID() {
		return id;
	}
	
	/**
     * Returns the name for solar event catalog.
     * 
     * @return name for solar event catalog.
     * */
	public String getName() {
		return name;
	}

	/**
     * Returns the creation date for solar event catalog.
     * 
     * @return creation date for solar event catalog.
     * */
	public Date getCreationDate() {
		return creationDate;
	}
	
    /**
     * Instances of this class are used as values for the internal cache of the
     * solar events. This class provides the connection between a time stamp and
     * the solar event.
     * 
     * @author Stephan Pagel
     * */
	private class CacheEventValue {

		// ///////////////////////////////////////////////////////////////////////
		// Definitions
		// ///////////////////////////////////////////////////////////////////////		
		
		private HashMap<Long, AbstractList<SolarEvent>> timeHash;
				
		// ///////////////////////////////////////////////////////////////////////
		// Methods
		// ///////////////////////////////////////////////////////////////////////
		
		/**
		 * Default constructor.
		 * */
		public CacheEventValue() {
			timeHash = new HashMap<Long, AbstractList<SolarEvent>>();
		}
		
        /**
         * Adds a new connection of a time stamp and solar event to the cache.
         * 
         * @param date
         *            time stamp for the solar event.
         * @param events
         *            solar event object which provides the information.
         */
		public void add(ImmutableDateTime date, AbstractList<SolarEvent> events) {
			timeHash.put(new Long(date.getMillis()), events);
		}
		
        /**
         * Returns a boolean value which indicates if a solar event is available
         * for a given date.
         * 
         * @param date
         *            search for a solar event at this date.
         * @return boolean value if a solar event is available for this date.
         */
		public boolean solarEventsForDateInCache(ImmutableDateTime date) {
			return timeHash.containsKey(new Long(date.getMillis()));
		}
		
        /**
         * Returns a list with all available solar events at the given date.
         * 
         * @param date
         *            get the solar events for this date.
         * @return list of solar events at the given date.
         * */
		public AbstractList<SolarEvent> getSolarEvents(ImmutableDateTime date) {
			
			Long key = new Long(date.getMillis());
			
			if (timeHash.containsKey(key)) {
				return timeHash.get(key);
			}
			
			return new LinkedList<SolarEvent>();
		}
	}
}
