package org.helioviewer.jhv.internal_plugins.overlay.solarevents;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.net.UnknownHostException;
import java.util.AbstractList;
import java.util.LinkedList;

import org.helioviewer.jhv.JHVGlobals;
import org.helioviewer.jhv.Settings;
import org.helioviewer.jhv.gui.components.StatusPanel;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

/**
 * This class acts as a container for all available solar event catalogs and
 * provides basic methods to manage these.
 * 
 * All solar events are bundled in different solar catalogs. The catalogs are
 * managed inside this class.
 * <p>
 * The class follows the singleton pattern.
 * 
 * @author Stephan Pagel
 * */
public class SolarEventCatalogs {

	// ///////////////////////////////////////////////////////////////////////////
	// Definitions
	// ///////////////////////////////////////////////////////////////////////////
	
	// the sole instance of this class
	private static final SolarEventCatalogs singletonInstance = new SolarEventCatalogs();
	
	// list of all solar event catalogs
	private AbstractList<SolarEventCatalog> solarEventCatalogs = new LinkedList<SolarEventCatalog>();
	
	// ///////////////////////////////////////////////////////////////////////////
	// Methods
	// ///////////////////////////////////////////////////////////////////////////
	
	/** 
	 * The private constructor to support the singleton pattern. 
	 * */
	private SolarEventCatalogs() {}
		
	/** 
	 * Method returns the sole instance of this class.
	 * 
	 * @return the only instance of this class.
	 * */
	public static SolarEventCatalogs getSingletonInstance() {
		return singletonInstance;
	}
	
    /**
     * Searches for and returns a solar event catalog which will be identified
     * by its id.
     * 
     * @param catalogID
     *            id which identifies the solar event catalog.
     * @return the corresponding solar event catalog of the passed id or null if
     *         no solar event catalog exists with the passed id.
     */
	public SolarEventCatalog getCatalogByID(String catalogID) {
		
		for (SolarEventCatalog catalog : solarEventCatalogs) {
			if (catalog.getID().equals(catalogID))
				return catalog;
		}
		
		return null;
	}
	
    /**
     * Returns a list with all available solar event catalog objects.
     * 
     * @return list with all available solar event catalog objects.
     */
	public AbstractList<SolarEventCatalog> getAllCatalogs() {
		return solarEventCatalogs;
	}
	
    /**
     * Returns a list with all solar event catalog objects where the visible
     * flag is set. The visible flag will be used to identify which solar event
     * catalogs are currently displayed in the solar event catalog panel.
     * 
     * @return list with all visible solar event catalog objects.
     */
	public AbstractList<SolarEventCatalog> getVisibleCatalogs() {
		
		AbstractList<SolarEventCatalog> result = new LinkedList<SolarEventCatalog>();
		
		for (SolarEventCatalog catalog : solarEventCatalogs) {
			if (catalog.isVisible()) {
				result.add(catalog);
			}
		}
		
		return result;
	}
	
    /**
     * Downloads all available solar event catalog information from the
     * specified server. Information about the contained solar events of a solar
     * event catalog will not be downloaded!
     * <p>
     * To download the solar event information see
     * {@link SolarEventCatalog#loadCatalogEvents(org.helioviewer.viewmodel.view.ImageInfoView, org.helioviewer.viewmodel.view.jp2view.datetime.ImmutableDateTime)}
     * */
	public void loadCatalogs() {
		
		StatusPanel.setStatusInfoText("Loading solar event catalogs");
		
		String response = null;
		
		try {
			// Construct data
			String data = URLEncoder.encode("action", "UTF-8") + "=" + URLEncoder.encode("getEventCatalogs", "UTF-8");

			// Send data
	        URL url = new URL(Settings.getSingletonInstance().getProperty("API.event.path"));
			URLConnection conn = url.openConnection();
			conn.setDoOutput(true);
			OutputStreamWriter wr = new OutputStreamWriter(conn.getOutputStream());
			wr.write(data);
			wr.flush();
			
			// Get the response
			BufferedReader rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
			response = rd.readLine();
			wr.close();
			rd.close();
		} catch (Exception e) {
			String message = e.getMessage();
			
			if(e instanceof UnknownHostException) {
				message = "Unknown Host: " + e.getMessage();
			}
			
			JHVGlobals.err("An error occured while loading the event catalogs!", message, false);
			return;
		}	    
			    
		try {	 
		    // analyze response and add new solar event catalog objects to the container. 
			JSONArray jsonarray = new JSONArray(response);
			
			if (jsonarray != null) {
				for(int i = 0; i < jsonarray.length(); i++) {
					
					JSONObject obj = jsonarray.getJSONObject(i);
					
					solarEventCatalogs.add(new SolarEventCatalog(
							obj.get("coordinateSystem").toString(),
							obj.get("description").toString(),
							obj.get("eventType").toString(),
							obj.get("id").toString(),
							obj.get("name").toString()));
				}
			}
		} catch (JSONException e) {
			e.printStackTrace();
		}
		
		StatusPanel.setStatusInfoText("");
	}
}
