package org.helioviewer.jhv.internal_plugins.overlay.solarevents;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.AbstractList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;

import org.helioviewer.jhv.gui.IconBank;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.gui.ViewListenerDistributor;
import org.helioviewer.jhv.gui.IconBank.JHVIcon;
import org.helioviewer.jhv.gui.components.statusplugins.ActiveLayerChangedReason;
import org.helioviewer.jhv.internal_plugins.overlay.solarevents.SolarEventCatalog;
import org.helioviewer.jhv.internal_plugins.overlay.solarevents.SolarEventCatalogs;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.changeevent.LayerChangedReason;
import org.helioviewer.viewmodel.changeevent.MetaDataChangedReason;
import org.helioviewer.viewmodel.changeevent.SubImageDataChangedReason;
import org.helioviewer.viewmodel.changeevent.ViewChainChangedReason;
import org.helioviewer.viewmodel.changeevent.LayerChangedReason.LayerChangeType;
import org.helioviewer.viewmodel.metadata.MetaData;
import org.helioviewer.viewmodel.metadata.ObserverMetaData;
import org.helioviewer.viewmodel.view.ImageInfoView;
import org.helioviewer.viewmodel.view.LayeredView;
import org.helioviewer.viewmodel.view.MetaDataView;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.ViewListener;
import org.helioviewer.viewmodel.view.jp2view.datetime.ImmutableDateTime;
import org.helioviewer.viewmodelplugin.overlay.OverlayPanel;

/**
 * Represents a component which manages the available solar event catalogs and
 * associated solar events.
 * 
 * @see SolarEventCatalogs
 * @author Stephan Pagel
 * */
public class SolarEventCatalogsPanel extends OverlayPanel implements ActionListener, ViewListener {

	// ///////////////////////////////////////////////////////////////////////////
	// Definitions
	// ///////////////////////////////////////////////////////////////////////////
	
	private static final long serialVersionUID = 1L;
	
	private HashMap<SolarEventCatalog, EventPanel> associatedPanels = new HashMap<SolarEventCatalog, EventPanel>();	
	private ImmutableDateTime currentDateOfActiveLayer = null;
	private ImageInfoView currentViewOfActiveLayer = null;
	private LayeredView layeredView;
	
	private JComboBox cbAvailableEvents = new JComboBox();
	private JButton cmdAddEvent = new JButton("Add");
	
	// ///////////////////////////////////////////////////////////////////////////
	// Methods
	// ///////////////////////////////////////////////////////////////////////////

	/**
     * Default constructor
     * */
	public SolarEventCatalogsPanel() {
		
		// set up visual components
		initVisualComponents();
		
		// register as view listener
		ViewListenerDistributor.getSingletonInstance().addViewListener(this);
		
		// initialize global variables
		ImageInfoView imageView = ImageViewerGui.getSingletonInstance().getImageSelectorPanel().getActiveImageInfoView();
		
		if (imageView != null) {
		    MetaData metaData = imageView.getAdapter(MetaDataView.class).getMetaData();
		    
		    if (metaData != null && imageView != null && metaData instanceof ObserverMetaData) {
		        currentViewOfActiveLayer = imageView;
                currentDateOfActiveLayer = ((ObserverMetaData)metaData).getDateTime();
		    }
		}
	}
	
	/**
     * Sets up the visual sub components and the visual part of the component itself.
     * */
	private void initVisualComponents() {

		// set general appearance
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		
		// add control components
		JPanel controlPanel = new JPanel(new BorderLayout());
		
		cmdAddEvent.addActionListener(this);
		
		controlPanel.add(cbAvailableEvents, BorderLayout.CENTER);
		controlPanel.add(cmdAddEvent, BorderLayout.EAST);
		
		add(controlPanel);

		// load component contents
		updateCatalogs();
	}
	
	/**
	 * Creates a layer for a selected event catalog and removes catalog from the
	 * list of available catalogs.
	 * 
	 * @param catalog
	 *            Catalog object that should be displayed in a new layer and
	 *            which will be removed from list of available catalogs.
	 * */
	private void addEventLayer(SolarEventCatalog catalog) {
		
		// create new layer
		EventPanel newPanel = new EventPanel(catalog);
		add(newPanel);
		
		// remember connection between layer and catalog object		
		associatedPanels.put(catalog, newPanel);
		
		// set visibility flag of the corresponding solar events
		catalog.setVisibility(true);
				
		// display everything correctly
		revalidate();
		
		if (layeredView != null && layeredView instanceof ViewListener) {
			((ViewListener)layeredView).viewChanged(null, new ChangeEvent(new SubImageDataChangedReason(null)));
		}
	}
	
	/**
	 * Removes the layer of a selected event catalog and adds the catalog to the
	 * list of available event catalogs.
	 * 
	 * @param catalog
	 *            Catalog object of the layer which should be removed and the
	 *            catalog object will be added to the list of available event
	 *            catalogs.
	 * */
	private void removeEventLayer(SolarEventCatalog catalog) {
		
		// remove associated layer of catalog
		remove(associatedPanels.remove(catalog));
		
		// add catalog to combobox
		cbAvailableEvents.addItem(catalog);
		
		// set visibility flag of the corresponding solar events
		catalog.setVisibility(false);
		
		// clean up the associated solar event cache
		catalog.clearCatalogEventsCache();
		
		// display components correctly
		cmdAddEvent.setEnabled(true);
		
		revalidate();
		
		if (layeredView != null && layeredView instanceof ViewListener) {
			((ViewListener)layeredView).viewChanged(null, new ChangeEvent(new SubImageDataChangedReason(null)));
		}
	};
	
	/**
	 * Updates components.
	 * */
	public void updateComponents() {
		
		// update the content of the combobox with available event catalogs
		updateCatalogs();
	}
	
	/**
     * Update the content of the combobox with available event catalogs.
     * */
	private void updateCatalogs() {
		// clear combo box
		cbAvailableEvents.removeAllItems();
		
		// add all event catalogs to combo box
		AbstractList<SolarEventCatalog> catalogs = SolarEventCatalogs.getSingletonInstance().getAllCatalogs();
		
		for (SolarEventCatalog sec : catalogs) {
		    if (!associatedPanels.containsKey(sec))
		        cbAvailableEvents.addItem(sec);
		}
		
		// set add button depending on number of available catalogs
		cmdAddEvent.setEnabled(cbAvailableEvents.getItemCount() > 0);
	}
	
	/**
	 * Updates the solar events for each available solar event catalog.
	 * 
	 * @param view
	 *            Associated view
	 * @param date
	 *            Date for solar events.
	 * @param redrawRequest
	 *            If true a change event will be fired which represents a
	 *            request that the main image will be redrawn
	 * */
	private void updateEvents(ImageInfoView view, ImmutableDateTime date, boolean redrawRequest) {
	
		// update solar events of each visible catalog
		for (Map.Entry<SolarEventCatalog, EventPanel> entry : associatedPanels.entrySet()) {
			
			SolarEventCatalog catalog = entry.getKey(); 
			
			if (catalog.isVisible())
				catalog.loadCatalogEvents(view, date);
		}
		
		// update solar event icons in display area
		if (redrawRequest && layeredView != null && layeredView instanceof ViewListener) {
			((ViewListener)layeredView).viewChanged(null, new ChangeEvent(new SubImageDataChangedReason(null)));
		}
	}
	
	/**
     * ActionListener for components of the class.
     * */
    public void actionPerformed(ActionEvent e) {
	    
    	if (e.getSource() == cmdAddEvent) {
	    
    		// get selected event catalog from combobox
	    	SolarEventCatalog catalog = (SolarEventCatalog)cbAvailableEvents.getSelectedItem();
	    	
	    	catalog.loadCatalogEvents(currentViewOfActiveLayer, currentDateOfActiveLayer);
	    	
	    	// add new event catalog layer and remove catalog from combobox
	    	addEventLayer(catalog);
	    	cbAvailableEvents.removeItem(catalog);
	    	
	    	// check if more catalogs available at combobox, if not disable add button
	    	if (cbAvailableEvents.getItemCount() == 0)
	    		cmdAddEvent.setEnabled(false);
	    }	    
    }
    
	/**
	 * Reacts on changes to the view chain or in certain circumstances to the GUI.
	 * 
	 * @param sender
	 * @param aEvent
	 * */
    public void viewChanged(View sender, ChangeEvent aEvent) {
    	
    	// check if the dominant image layer has changed
    	ActiveLayerChangedReason activeLayerReason = aEvent.getLastChangedReasonByType(ActiveLayerChangedReason.class);
    	
    	if (activeLayerReason != null) {
    		
    		// the layer has changed -> update events (show corresponding events to selected image layer)
    		if (sender != null) {
    			MetaData metaData = sender.getAdapter(MetaDataView.class).getMetaData();
    			ImageInfoView view = sender.getAdapter(ImageInfoView.class);
        		
        		if (metaData != null && view != null && metaData instanceof ObserverMetaData) {
        			ImmutableDateTime newDate = ((ObserverMetaData)metaData).getDateTime();
        			
        			if (!view.equals(currentViewOfActiveLayer)) {
        				currentViewOfActiveLayer = view;
        				currentDateOfActiveLayer = newDate;
                		updateEvents(view, newDate, true);	
        			}
        		}	
    		}
    	}
    	
    	// check if the meta data (frame in a movie) has changed
    	List<MetaDataChangedReason> metaDataReasons = aEvent.getAllChangedReasonsByType(MetaDataChangedReason.class);
    	
    	for(MetaDataChangedReason metaDataReason : metaDataReasons) {
			MetaData metaData = metaDataReason.getNewMetaData();
			ImageInfoView view = metaDataReason.getView().getAdapter(ImageInfoView.class);
        		
    		if (metaData != null && view != null && metaData instanceof ObserverMetaData) {
    			ImmutableDateTime newDate = ((ObserverMetaData)metaData).getDateTime();
    			
    			if (!newDate.equals(currentDateOfActiveLayer)) {
    				currentViewOfActiveLayer = view;
    				currentDateOfActiveLayer = newDate;
            		updateEvents(view, newDate, false);	
    			}
    		}	
    	}
    	
    	// check if an image layer has been removed
    	LayerChangedReason layerReason = aEvent.getLastChangedReasonByType(LayerChangedReason.class);
    	
    	if (layerReason != null && layerReason.getLayerChangeType() == LayerChangeType.LAYER_REMOVED) {
    		
    		// clear cache for corresponding times
    		View subView = layerReason.getSubView();
    		
    		if (subView != null) {
    			clearCacheWhenImageLayerRemoved(subView.getAdapter(ImageInfoView.class));	
    		}
    	}
    	
    	// check if something changed with the view chain
    	ViewChainChangedReason viewChainReason = aEvent.getLastChangedReasonByType(ViewChainChangedReason.class);
    	
    	if (viewChainReason != null) {
    		
    		// a change occurred to the view chain -> get reference of the current layered view 
    		layeredView = ImageViewerGui.getSingletonInstance().getMainView().getAdapter(LayeredView.class);
    	}
    }
    
    /**
     * Clears the associated solar event cache for the layer containing the
     * removed image.
     * 
     * @param imageInfoView
     *            image info view which was removed and specifies the layer.
     */
    private void clearCacheWhenImageLayerRemoved(ImageInfoView imageInfoView) {
    	
    	if (imageInfoView == null)
    		return; 
    	
    	for (Map.Entry<SolarEventCatalog, EventPanel> entry : associatedPanels.entrySet()) {
			entry.getKey().removeCatalogEventsFromCache(imageInfoView); 
		}
    }
	
    /**
     * Represents a layer of an event catalog.
     * 
     * @author Stephan Pagel
     * */
	private class EventPanel extends JPanel implements ActionListener {
		// ///////////////////////////////////////////////////////////////////////
		// Definitions
		// ///////////////////////////////////////////////////////////////////////
		
		private static final long serialVersionUID = 1L;
		
		private SolarEventCatalog catalog;
		
		private JButton cmdVisibility;
		private JButton cmdRemove;
		
		// ///////////////////////////////////////////////////////////////////////
		// Methods
		// ///////////////////////////////////////////////////////////////////////
		
		/**
		 * Default constructor.
		 * 
		 * @param catalog
		 *            Associated solar event catalog of the layer.
		 * */
		public EventPanel(SolarEventCatalog catalog) {
			
			// memorize solar event catalog
	        this.catalog = catalog;
	        
	        // set up components
	        initVisualComponents();
        }
		
		/**
         * Sets up the visual sub components and the visual part of the component itself.
         * */
		private void initVisualComponents() {
			
			// Set up basic appearance
			setLayout(new BorderLayout());
			setBorder(BorderFactory.createEtchedBorder(null, null));
				
			// Add icon and name of catalog
			add(new SolarEventCatalogImagePanel("resources/images/EventIcons/" + catalog.getID().replace(':', '_') + ".png"), BorderLayout.WEST);
			add(new JLabel("  " + catalog.getName()), BorderLayout.CENTER);			
			
			// Add Buttons
			JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT,1,1));
			
			cmdRemove = createButton(IconBank.getIcon(JHVIcon.REMOVE_LAYER), "Remove catalog");
			buttonPanel.add(cmdRemove);
			
			cmdVisibility = createButton(IconBank.getIcon(JHVIcon.VISIBLE), "Hide/Show events");
			buttonPanel.add(cmdVisibility);
			
			add(buttonPanel, BorderLayout.EAST);
		}
		
		/**
		 * Adds a control button to the layer.
		 * 
		 * @param icon
		 *            Icon which should be displayed on the button
		 * @param tooltip
		 *            Tooltip which should be displayed when mouse cursor will
		 *            be hold over button for longer time.
		 * */
		private JButton createButton(Icon icon, String tooltip) {
			
			JButton newButton = new JButton(icon);
			
			newButton.setBorder(BorderFactory.createEtchedBorder());
			newButton.setPreferredSize(new Dimension(icon.getIconWidth()+5,23));
			newButton.setToolTipText(tooltip);
			newButton.addActionListener(this);
			
			return newButton;
		}

		/**
	     * ActionListener for components of the class.
	     * */
        public void actionPerformed(ActionEvent arg0) {
	        
        	if (arg0.getSource() == cmdRemove) {
        		// remove layer 
        		removeButtonClicked();
        	} else if (arg0.getSource() == cmdVisibility) {
        		// change visibility of events which are belong to the solar event catalog of the layer
        		visibilityButtonClicked();	        
        	}
        }
        
        /**
         * Arranges for this layer will be removed.
         * */
        private void removeButtonClicked() {
        	removeEventLayer(catalog);
        }
        
        /**
         * Arranges for changes of visibility of solar events which belongs to this layer.
         * */
        private void visibilityButtonClicked() {
        	
        	// change visibility flag
        	catalog.setVisibility(!catalog.isVisible());
        	
        	// change visibility image
        	if (catalog.isVisible()) {
        		cmdVisibility.setIcon(IconBank.getIcon(JHVIcon.VISIBLE));
        		
        		// reload solar events because image layer might have changed
        		catalog.loadCatalogEvents(currentViewOfActiveLayer, currentDateOfActiveLayer);
        	} else {
        		cmdVisibility.setIcon(IconBank.getIcon(JHVIcon.HIDDEN));
        	}
        	
        	// update displayed icons
        	if (layeredView != null && layeredView instanceof ViewListener) {
    			((ViewListener)layeredView).viewChanged(null, new ChangeEvent(new SubImageDataChangedReason(null)));
    		}
        }
        
        /**
         * Represents a panel which displays the icon of a solar event catalog.
         * 
         * @author Stephan Pagel
         * */
        private class SolarEventCatalogImagePanel extends JPanel {

        	// ///////////////////////////////////////////////////////////////////
    		// Definitions
    		// ///////////////////////////////////////////////////////////////////
        	
            private static final long serialVersionUID = 1L;
            
            // offset of image from left side
            private final int vOffset = 2;
            
            private Image img = null;
            
            // ///////////////////////////////////////////////////////////////////
    		// Definitions
    		// ///////////////////////////////////////////////////////////////////

            /**
             * Default constructor.
             * 
             * @param iconFileName
             *            File name of the icon of the solar event catalog.
             * */
			public SolarEventCatalogImagePanel(String iconFileName) {
				
				// load icon from icon bank
				img = new ImageIcon(IconBank.class.getResource(iconFileName)).getImage();
				
				// set up the general visual appearance of this component
				setLayout(null);
				
				Dimension size = new Dimension(img.getWidth(null) + vOffset, img.getHeight(null));
				setMinimumSize(size);
			    setSize(size);
			    
			    setPreferredSize(size);
            }
			
			/**
			 * Draws the icon on the component.
			 * */
			@Override
			public void paintComponent(Graphics g) {
				
				if (img != null) {
					g.drawImage(img, vOffset, (getHeight() - img.getHeight(null)) / 2, null);	
				}   
			}
        }
	}
}
