package org.helioviewer.jhv.internal_plugins.overlay.solarevents;

import java.awt.image.BufferedImage;
import java.io.IOException;

import java.util.AbstractList;
import java.util.AbstractMap;
import java.util.HashMap;

import javax.imageio.ImageIO;

import org.helioviewer.base.math.Vector2dDouble;
import org.helioviewer.base.physics.Constants;
import org.helioviewer.jhv.gui.IconBank;
import org.helioviewer.viewmodel.renderer.physical.PhysicalRenderGraphics;
import org.helioviewer.viewmodel.renderer.physical.PhysicalRenderer;

/**
 * The solar event renderer provides a possibility to draw solar events with
 * there associated icons.
 * 
 * @author Markus Langenberg
 * @author Stephan Pagel
 */
public class SolarEventRenderer implements PhysicalRenderer {
	
    // ///////////////////////////////////////////////////////////////////////////
    // Definitions
    // ///////////////////////////////////////////////////////////////////////////
    
	private AbstractMap<String, BufferedImage> eventIcons = new HashMap<String, BufferedImage>();
	
	// ///////////////////////////////////////////////////////////////////////////
    // Methods
    // ///////////////////////////////////////////////////////////////////////////
	
	/**
	 * Default constructor.
	 */
	public SolarEventRenderer() {
	    
	    // load the icons for the different event types
		try {
			eventIcons.put("CACTusService::CACTus", ImageIO.read(IconBank.class.getResource("resources/images/EventIcons/CACTusService__CACTus.png")));
			eventIcons.put("EGSOService::halpha_flares_event", ImageIO.read(IconBank.class.getResource("resources/images/EventIcons/EGSOService__halpha_flares_event.png")));
			eventIcons.put("EGSOService::hessi_flare", ImageIO.read(IconBank.class.getResource("resources/images/EventIcons/EGSOService__hessi_flare.png")));
			eventIcons.put("EITPlanningService::EITActivity", ImageIO.read(IconBank.class.getResource("resources/images/EventIcons/EITPlanningService__EITActivity.png")));
			eventIcons.put("GOESXRayService::GOESXRay", ImageIO.read(IconBank.class.getResource("resources/images/EventIcons/GOESXRayService__GOESXRay.png")));
			eventIcons.put("SEEDSService::SEEDS", ImageIO.read(IconBank.class.getResource("resources/images/EventIcons/SEEDSService__SEEDS.png")));
			eventIcons.put("VSOService::cmelist", ImageIO.read(IconBank.class.getResource("resources/images/EventIcons/VSOService__cmelist.png")));
			eventIcons.put("VSOService::noaa", ImageIO.read(IconBank.class.getResource("resources/images/EventIcons/VSOService__noaa.png")));
			eventIcons.put("VSOService::type2cme", ImageIO.read(IconBank.class.getResource("resources/images/EventIcons/VSOService__type2cme.png")));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * {@inheritDoc}
	 * 
	 * Draws all available and visible solar events with there associated icon.
	 */
	public void render(PhysicalRenderGraphics g) {
		
	    // go through all available and visible solar event catalogs
		AbstractList<SolarEventCatalog> catalogs = SolarEventCatalogs.getSingletonInstance().getVisibleCatalogs();
		
		for(SolarEventCatalog catalog : catalogs) {
			
		    // go through the event catalog and draw each solar event at the defined position with the associated icon
			AbstractList<SolarEvent> events = catalog.getEvents();
			
			for(SolarEvent event : events) {
				try {
					Vector2dDouble eventPos = new Vector2dDouble(
							Constants.SunRadius * event.getDouble("sunX"), // Offset of event in x-axis starting from the center of the sun
	        				Constants.SunRadius * event.getDouble("sunY")); // Offset of event in y-axis starting from the center of the sun (down=positive)

					g.drawImage(eventIcons.get(event.getCatalogID()), eventPos);
				} catch (SolarEvent.SolarEventException e) {}
			}
		}
	}
}
