package org.helioviewer.jhv.io;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Scanner;
import java.util.TimeZone;

import org.helioviewer.jhv.JHVGlobals;
import org.helioviewer.jhv.Settings;
import org.helioviewer.jhv.gui.ImageViewerGui;
import org.helioviewer.jhv.gui.ViewchainFactory;
import org.helioviewer.viewmodel.io.APIResponse;
import org.helioviewer.viewmodel.io.APIResponseDump;
import org.helioviewer.viewmodel.view.ImageInfoView;
import org.helioviewer.viewmodel.view.ViewHelper;

/**
 * This class provides methods to download files from a server.
 * 
 * Most of the methods only will work with the current Helioviewer server
 * because they modify links and requests that they will fit with the API.
 * 
 * @author Stephan Pagel
 */
public class APIRequestManager {
    
    /**
     * Method does remote opening. File (single frame) is opened via JPIP on
     * delphi.nascom.nasa.gov:8090.
     *
     * @param date
     *            date of the requested image.
     * @param time
     *            time if the requested image.
     * @param observation
     *            observation of the requested image.
     * @param instrument
     *            instrument of the requested image.
     * @param detector
     *            detector of the requested image.
     * @param measurement
     *            measurement of the requested image.
     * @throws IOException
     */
    public static void requestAndOpenRemoteFile(Date date, String time, 
            String observation, String instrument, String detector, String measurement) throws IOException {
        
        requestAndOpenRemoteFile(true, "", date, time, null, null, 
                observation, instrument, detector, measurement);
    }
    
    /**
     * Method does remote opening. If image series, file is downloaded. If
     * single frame, file is opened via JPIP on delphi.nascom.nasa.gov:8090.
     * 
     * @param singleImage
     *            boolean value if a single image or an image series should be
     *            loaded.
     * @param cadence
     *            cadence between two frames (if an image series is requested).
     * @param startDate
     *            start date of the requested image (series).
     * @param startTime
     *            start time of the requested image (series).
     * @param endDate
     *            end date of the requested image (series).
     * @param endTime
     *            end time of the requested image (series).
     * @param observation
     *            observation of the requested image (series).
     * @param instrument
     *            instrument of the requested image (series).
     * @param detector
     *            detector of the requested image.
     * @param measurement
     *            measurement of the requested image.
     * @throws IOException
     */
    public static void requestAndOpenRemoteFile(boolean singleImage, String cadence, Date startDate, String startTime, Date endDate, String endTime,
            String observation, String instrument, String detector, String measurement) throws IOException {
        
        if ((cadence == "" || endDate == null || endTime == null || singleImage)
                && dateIsValid(startDate) && timeIsValid(startTime)) {
            
        	newLoad(getImageURI(observation, instrument, detector, measurement, startDate, startTime)); 
                 
        } else if (dateIsValid(startDate) && timeIsValid(startTime) && dateIsValid(endDate)
                && timeIsValid(endTime) && new Integer(cadence) > 0) {
            
        	newLoad(getImageSeriesURI(observation, instrument, detector, measurement, 
                    startDate, startTime, endDate, endTime, cadence));
        }
    }
    
    /**
     * Loads the image or image series from the given URI, creates a new image
     * info view and adds it as a new layer to the view chain of the main image.
     * 
     * @param uri
     *            specifies the location of the file.
     * @return associated image info view of the given image or image series
     *         file.
     * @throws IOException
     */
    public static ImageInfoView newLoad(URI uri) throws IOException {
        if(uri == null) {
            return null;
        }

        // Load new view and assign it to view chain of Main Image
        
        ImageInfoView view = ViewHelper.loadView(uri);

        ViewchainFactory factory = new ViewchainFactory();
        factory.addLayerToViewchainMain(view, ImageViewerGui.getSingletonInstance().getMainView());
            
        return view;
    }
    
    /**
     * Loads the image or image series from the given URI, creates a new image
     * info view and adds it as a new layer to the view chain of the main image.
     * 
     * @param uri
     *            specifies the location of the file.
     * * @param downloadUTI
     *            the http uri from which the whole file can be downloaded
     * @return associated image info view of the given image or image series
     *         file.
     * @throws IOException
     */
    public static ImageInfoView newLoad(URI uri, URI downloadURI)
            throws IOException {
        if (uri == null) {
            return null;
        }

        // Load new view and assign it to view chain of Main Image

        ImageInfoView view = ViewHelper.loadView(uri, downloadURI);

        ViewchainFactory factory = new ViewchainFactory();
        factory.addLayerToViewchainMain(view, ImageViewerGui
                .getSingletonInstance().getMainView());

        return view;

    }
    
    /**
     * Method that checks whether a date is valid.
     * 
     * @param _date
     *            Date to check.
     * @return boolean value if date is valid.
     */
    public static boolean dateIsValid(Date _date) {
        return _date != null;
    }
    
    /**
     * Method that checks whether a time is valid.
     * 
     * @param _time
     *            Time to check
     * @return boolean value if date is valid.
     */
    public static boolean timeIsValid(String _time) {
        boolean check = true;
        String[] parts = _time.split(":");

        if (parts.length == 3) {
            try {
                int hours = new Integer(parts[0]);
                int minutes = new Integer(parts[1]);
                int seconds = new Integer(parts[2]);
    
                if (hours < 0 || hours >= 24)
                    check = false;
                if (minutes < 0 || minutes >= 60)
                    check = false;
                if (seconds < 0 || seconds >= 60)
                    check = false;
            } catch (NumberFormatException nfe) {
                check = false;
            } catch (Exception e) {
                e.printStackTrace();
            }
        } else
            check = false;

        return check;
    }
    
    /**
     * Sends an request to the server to compute where the nearest image is
     * located on the server. The address of the file will be returned.
     * 
     * @param observation
     *            observation of the requested image.
     * @param instrument
     *            instrument of the requested image.
     * @param detector
     *            detector of the requested image.
     * @param measurement
     *            measurement of the requested image.
     * @param startDate
     *            date of the requested image.
     * @param startTime
     *            time if the requested image.
     * @return address of the nearest image file on the server.
     * @throws MalformedURLException
     * @throws IOException
     */          
    public static URI getImageURI(String observation, String instrument, String detector, String measurement, Date startDate, String startTime) throws MalformedURLException, IOException {
        
        if (measurement.equalsIgnoreCase("white light")) {
            measurement = "white%20light";
        }
        
        if (detector.replaceAll(" ", "").length() == 0) {
            detector = instrument;
        }
        
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'");
        dateFormat.setTimeZone(TimeZone.getTimeZone("GMT"));

        String url = Settings.getSingletonInstance().getProperty("API.jp2images.path")
        + "?action=getJP2Image&observatory="
        + observation
        + "&instrument="
        + instrument
        + "&detector="
        + detector
        + "&measurement="
        + measurement
        + "&date="
        + dateFormat.format(startDate) + startTime + 'Z'
        + "&getJPIP=true";
        
        // get URL from server where nearest image is located
        try {
	        Scanner scanner = new Scanner(new URL(url).openStream());
	
	        if (scanner != null) {
	            URI address = new URI(scanner.nextLine());
	            scanner.close();
	            return address;
	        }
        } catch (IOException e) {
        	if(e instanceof UnknownHostException) {
        		throw new IOException("Unknown Host: " + e.getMessage());
        	} else {
        		throw new IOException("Error in the server communication:" + e.getMessage());
        	}
        } catch (URISyntaxException e) {
			e.printStackTrace();
		}
        
        return null;
    }
    
    /**
     * Sends an request to the server to compute where the image series is
     * located on the server. The address of the file will be returned.
     * 
     * @param observation
     *            observation of the requested image series.
     * @param instrument
     *            instrument of the requested image series.
     * @param detector
     *            detector of the requested image series.
     * @param measurement
     *            measurement of the requested image series.
     * @param startDate
     *            start date of the requested image series.
     * @param startTime
     *            start time of the requested image series.
     * @param endDate
     *            end date of the requested image series.
     * @param endTime
     *            end time of the requested image series.
     * @param cadence
     *            cadence between to images of the image series.
     * @return address of the file which represents the image series on the
     *         server.
     * @throws MalformedURLException
     * @throws IOException
     */
    public static URI getImageSeriesURI(String observation, String instrument, String detector, String measurement, Date startDate, String startTime, Date endDate, String endTime, String cadence) throws MalformedURLException, IOException {
        
    	if (measurement.equalsIgnoreCase("white light")) {
            measurement = "white%20light";
        }
        
        if (detector.replaceAll(" ", "").length() == 0) {
            detector = instrument;
        }
        
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'");
        dateFormat.setTimeZone(TimeZone.getTimeZone("GMT"));

        String url = Settings.getSingletonInstance().getProperty("API.jp2series.path")
        + "?action=getJPX&observatory="
        + observation
        + "&instrument="
        + instrument
        + "&detector="
        + detector
        + "&measurement="
        + measurement
        + "&startTime="
        + dateFormat.format(startDate) + startTime + 'Z'
        + "&endTime="
        + dateFormat.format(endDate) + endTime + 'Z'
        + "&cadence="
        + cadence
        + "&getJPIP=true&frames=true&verbose=true";
        
        // get URL from server where file with image series is located
        try {
	        Scanner scanner = new Scanner(new URL(url).openStream());
	
	        if (scanner != null) {
	        	APIResponse response = new APIResponse(scanner.nextLine());
	        	scanner.close();
	            
	        	String message = response.getString("message");
	        	
	        	if(message != null && !message.equalsIgnoreCase("null")) {
		        		
	        		String[] messageWords = message.split(" ");
		        	message = "";
		        	int lineLength = 0;
		        	for(String word : messageWords) {
		        		if(lineLength + word.length() < 70) {
		        			message += word + " ";
		        			lineLength += word.length() + 1;
		        		} else {
		        			message += "\n" + word + " ";
		        			lineLength = word.length() + 1;
		        		}
		        	}
		        	
		        	JHVGlobals.warn("Warning", message);
	        	}
	            APIResponseDump.getSingletonInstance().putResponse(response);
	            
	            return response.getURI();
	        }
        } catch (IOException e) {
        	if(e instanceof UnknownHostException) {
        		throw new IOException("Unknown Host: " + e.getMessage());
        	} else {
        		throw new IOException("Error in the server communication:" + e.getMessage());
        	}
        }
        
        return null;
    }
}
