package org.helioviewer.jhv.io;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.AbstractList;
import java.util.LinkedList;

/**
 * A simple class to process command line arguments for JHelioviewer specification.
 * 
 * @author Alen Alexanderian
 * @author Stephan Pagel
 */
public class CommandLineProcessor {
   
    // ////////////////////////////////////////////////////////////////
    // Definitions
    // ////////////////////////////////////////////////////////////////
    
    private static String[] arguments;
    private static int expectedArgumentCount = 0;
    private static String usageMessage; 

    // ////////////////////////////////////////////////////////////////
    // Methods
    // ////////////////////////////////////////////////////////////////    
    
    /**
     * Default constructor.
     * */
    public CommandLineProcessor() {}
    
    /**
     * Sets the arguments (text from command line) to this class.
     * 
     * @param args command line arguments.
     */
    public void setArguments(String[] args) {
        arguments = args;
        
        setUsageMessage("The following command-line options are available: \n\n" +
                "-jhv  jhv request(s)\n" +
                "      allows the user to pass a comma separated list if requests of following form: \n\n"    +
                "             [date;time;observatory;instrument;detector;measurement;opacity]\n\n" +
                "             opacity is an integer 1-100.\n\n"      +
                "-jpip  jpip url\n"                                                                           +
                "       allows users to pass a jpip url of a JP2 or JPX image to be opened upon program's\n"  +
                "       start.\n\n"                                                                           +
                "-download  http path \n"                                                                     +
                "       allows the users to pass the location of JP2 or JPX image, which will be \n"          +
                "       downloaded and opened when the program starts. This is specially useful \n"           +
                "       for the case of large jpx files which will be very slow to play remotely.\n"
                );
        
        setExpectedArgumentCount(0);
    }
    
    /**
     * Returns the values of the -jhv option. Each value which was comma
     * separated in the command line will be returned as an entry in the list.
     * 
     * @return values of the -jhv option.
     */
    public AbstractList<String[]> getJHVOptionValues() {
        
        AbstractList<String[]> result = new LinkedList<String[]>();
        
        String optionValue = getOption("jhv");
        
        if (!optionValue.equals("")) {
            String[] jhvRequests = optionValue.split(",");
            
            for (int i = 0; i < jhvRequests.length; i++) {
                
                try {                    
                    result.add(parseJHVRequest(jhvRequests[i]));
                } catch (IllegalArgumentException e) {}
            }
        }
        
        return result;
    }
    
    /**
     * Returns the value of the -jpip option. If an illegal value was defined
     * and no URI could be created a null value will be returned.
     * 
     * @return value of the -jpip option as URI or null if no URI could be
     *         created.
     */
    public URI getJPIPOptionValue() {
        
        // get associated value string for "jpip" option
        String jpipURL = getOption("jpip");
        
        // create URI from value string if possible
        if (!jpipURL.equals("")) {
            try {
                return new URI(jpipURL);
            } catch (URISyntaxException e) {
                e.printStackTrace();
            }
        }
        
        // no URI available
        return null;
    }
    
    /**
     * Returns the value of the -download option. If an illegal value was
     * defined and no URI could be created a null value will be returned.
     * 
     * @return value of the -download option as URI or null if no URI could be
     *         created.
     */
    public URI getDownloadOptionValue() {
        
        // get associated value for "download" option
        String address = getOption("download");
        
        // create URI from value if possible
        if (!address.equals("")) {
            try {
                return new URI(address);
            } catch (URISyntaxException e) {
                e.printStackTrace();
            }
        }
        
        // no URI available
        return null;
    }

    /**
     * Method that checks that correct number of arguments have been passed.
     * 
     * @return boolean value if the correct number of arguments have been
     *         passed.
     * */
    public boolean validateCommandLine() {
        if ((arguments == null) || (arguments.length != expectedArgumentCount)) {

            System.err.println(getUsageMessage());
            return false;
        }
        return true;
    }

    /**
     * Method that looks for an option in the command line.
     * 
     * @param param
     *            name of the option.
     * @return the values associated to the option.
     * */
    public String getOption(String param)
    {
        param = "-"+param;
        String value = "";
        
        if (arguments != null) {
            for (int i = 0; i < arguments.length; i++) {
                if (param.equals(arguments[i] )) {
                    if (arguments.length > (i + 1)) {
                        value =  arguments[i+1];
                        break;
                    }
                }
            }
        }
        
        return value;
    }

    /** 
     * Method that parses JHV data request string in the form:
     * [date;time;observatory;instrument;detector;measurement:opacity].
     * 
     * @param _data data request from command line.
     * @return values of JHV data request as an array.
     * @throws IllegalArgumentException
     */
    public String[] parseJHVRequest(String _data) throws IllegalArgumentException
    {
        String[] fields = null;
        
        if ((_data.charAt(0) != '[') || (_data.charAt(_data.length()-1) != ']'))
        {
            System.err.println(getUsageMessage());
            throw new IllegalArgumentException();
        }
        _data = _data.substring(1, _data.length()-1);

        fields = _data.split(";");
        if (fields.length != 7)
        {
            System.err.println(getUsageMessage());
            throw new IllegalArgumentException();
        }
                
        return fields;
    }

    /**
     * Returns the expected number of arguments.
     * 
     * @return expected number of arguments.
     */
    public int getExpectedArgumentCount() {
        return expectedArgumentCount;
    }

    /**
     * Returns the command line usage message.
     * 
     * @return command line usage message.
     */
    public String getUsageMessage() {
        return usageMessage;
    }

    /**
     * Sets the expected number of arguments.
     * 
     * @param i
     *            of expected arguments.
     */
    public void setExpectedArgumentCount(int i) {
        expectedArgumentCount = i;
    }

    /**
     * Sets the command line usage message.
     * 
     * @param string
     *            command line usage message.
     */
    public void setUsageMessage(String string) {
        usageMessage = string;
    }
}
