package org.helioviewer.jhv.opengl;

import java.awt.Toolkit;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.util.LinkedList;
import java.util.List;

import org.helioviewer.jhv.JHVDirectory;
import org.helioviewer.jhv.JHVGlobals;
import org.helioviewer.jhv.JHVSplashScreen;
import org.helioviewer.jhv.io.FileDownloader;

/** Class responsible for downloading and linking all OpenGL related libraries.
 * 
 * <p>Before being able to call any OpenGL function, the native OpenGL
 * libraries have to be linked. Therefore, they have to be downloaded, if
 * they are not available on the machine yet. This class performs both
 * tasks.
 * 
 * @author Markus Langenberg
 */
public class GLLoadLibraries {
	
	private static final String SERVER_PATH = "http://www.jhelioviewer.org/libjhv/jogl-libs/jogl/";
	
	private final String OS;
	private final String ARCH;
    private final Platform PLATFORM;
    
    private List<String> libs;
	
    /**
     * Default constructor.
     * 
     * Determines operating system, architecture and platform, to decide
     * which libraries are needed.
     */
	public GLLoadLibraries() {
		OS = System.getProperty("os.name").toUpperCase();
	    ARCH = System.getProperty("os.arch").toUpperCase(); 
	
	    Platform temp = Platform.UNKNOWN;
	    if(OS.indexOf("WINDOWS") != -1){
	    	if(ARCH.indexOf("64") != -1)
	    		temp = Platform.WINDOWS_AMD64;
	    	else if(ARCH.indexOf("86") != -1)
	    		temp = Platform.WINDOWS_I586;
	    }else if(OS.indexOf("LINUX") != -1){
	    	if(ARCH.indexOf("64") != -1)
	    		temp = Platform.LINUX_AMD64;
	    	else if(ARCH.indexOf("86") != -1)
	    		temp = Platform.LINUX_I586;
	    }else if(OS.indexOf("MAC OS X") != -1){
	        temp = Platform.MAC;
	    }
	    
	    if(temp == Platform.UNKNOWN) {
	        // Fail 'hard'
	        JHVGlobals.err(
	                "Cannot detect the OS/ARCH!",
	                "OS - " + OS + "\nARCH - " + ARCH,
	                true);
	    }
	    PLATFORM = temp;
	    
	    // Preparations: Load system Library
	    Toolkit.getDefaultToolkit();
	    if(PLATFORM != Platform.MAC)
	    	System.loadLibrary("jawt");
	    
	    // Build list of needed libraries
	    libs = new LinkedList<String>();
	    libs.add(System.mapLibraryName("gluegen-rt"));
	    libs.add(System.mapLibraryName("jogl"));
	    libs.add(System.mapLibraryName("jogl_awt"));
	}
	
	/** 
	 * Attempts to delete the old libraries.
	 * 
	 * This might be useful if the libraries are outdated,
	 * incomplete or corrupt.
     * 
     * @throws SecurityException If the access to delete the files is denied
     */
    public void deleteLibs() throws SecurityException{
        for(String lib : libs){
            File libFile = new File(JHVDirectory.LIBS.getPath() + JHVDirectory.FILE_SEP + lib);
            if(libFile.exists())
                libFile.delete();
        }
    }

    /**
     * Attempts to load the required libraries.
     * 
     * @throws UnsatisfiedLinkError If at least one library does not exist
     */
    public void loadLibs() throws UnsatisfiedLinkError{
    	for(String lib : libs) {
        	String libPath = JHVDirectory.LIBS.getPath() + lib;
        	
        	try {
    			File localLib = new File(libPath);
    			URL remoteLib = new URL(SERVER_PATH + PLATFORM.getServerFolder() + lib);
    			URLConnection remoteConnection = remoteLib.openConnection();
    			
    			remoteConnection.setConnectTimeout(2000);
    			int remoteLibLength = remoteConnection.getContentLength();
    			
    			if(localLib.length() != remoteLibLength && remoteLibLength > 0) {
    				throw new UnsatisfiedLinkError(libPath + " is outdated or corrupted");
    			}
    		
            } catch (MalformedURLException e) {
    			e.printStackTrace();
    		} catch (IOException e) {}
    		
        	System.load(libPath);
        }
    	com.sun.opengl.impl.NativeLibLoader.disableLoading();
    	com.sun.gluegen.runtime.NativeLibLoader.disableLoading();
    }
    
    /**
     * Returns the number of required libraries.
     * 
     * @return number of required libraries.
     */
    public int getNumberOfLibs() {
        
        // return number of libs + 1 cgc file!
        return libs.size() + 1;
    }

    /**
     * Attempts to download the required libraries from the server.
     * 
     * @param splash
     *            instance of the splash screen to retrieve the progress bar and
     *            status label.
     * @throws IOException If the server is unreachable or the files are not found
     */
    public void downloadLibs(JHVSplashScreen splash) throws IOException{
        
    	splash.setProgressText("Installing OpenGL Libraries...");
            
    	FileDownloader fileDownloader = new FileDownloader();
        for(String lib : libs){
            
            splash.nextStep();
            
            try {
				fileDownloader.get(
				        new URI(SERVER_PATH + PLATFORM.getServerFolder() + lib), 
				        new File(JHVDirectory.LIBS.getPath()), 
				        splash.getProgressBar());
			} catch (URISyntaxException e) {
				throw new IOException(e.getMessage());
			}
        }
    }
    
    /** Checks whether the Cg stand-alone compiler is present, downloads it if not.
     * 
     * @param splash
     *            instance of the splash screen to retrieve the progress bar and
     *            status label.
     * @throws IOException If the server is unreachable or the files are not found
     */
    public void checkAndDownloadCgc(JHVSplashScreen splash) throws IOException {
        
        splash.nextStep();
        
    	String cgcName = "cgc";
    	if(PLATFORM == Platform.WINDOWS_I586 || PLATFORM == Platform.WINDOWS_AMD64)
    		cgcName += ".exe";
    		
    	File cgc = new File(JHVDirectory.LIBS.getPath() + cgcName);
    	URL remoteCgc = new URL(SERVER_PATH + PLATFORM.getServerFolder() + cgcName);
    	URLConnection remoteConnection = remoteCgc.openConnection();
    	
    	remoteConnection.setConnectTimeout(2000);
		int remoteCgcLength = remoteConnection.getContentLength();
		
		if(!cgc.exists() || (cgc.length() != remoteCgcLength && remoteCgcLength > 0)) {
    		FileDownloader fileDownloader = new FileDownloader();
    		splash.setProgressText("Installing OpenGL Libraries...");
        
    		try {
				fileDownloader.get(
				        new URI(SERVER_PATH + PLATFORM.getServerFolder() + cgcName), 
				        new File(JHVDirectory.LIBS.getPath()), 
				        splash.getProgressBar());
			} catch (URISyntaxException e) {
				throw new IOException(e.getMessage());
			}
    	}
    	
    	if(PLATFORM != Platform.WINDOWS_I586 && PLATFORM != Platform.WINDOWS_AMD64) {
    		Runtime.getRuntime().exec("chmod 755 " + cgc);
    	}
    }
	
	private static enum Platform {
        WINDOWS_AMD64{
            @Override
            public String getServerFolder() {return "windows/amd64/";}
        },
        WINDOWS_I586{
            @Override
            public String getServerFolder() {return "windows/i586/";}
        },
        LINUX_AMD64{
            @Override
            public String getServerFolder() {return "linux/amd64/";}
        },
        LINUX_I586{
            @Override
            public String getServerFolder() {return "linux/i586/";}
        },
        MAC{
            @Override
            public String getServerFolder() {return "macosx/";}
        },
        UNKNOWN{
            @Override
            public String getServerFolder() {return "";}
        };

        abstract public String getServerFolder();
    }
}
