package org.helioviewer.viewmodel.imagedata;

import java.awt.Point;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferUShort;
import java.awt.image.Raster;

import org.helioviewer.viewmodel.imageformat.ImageFormat;
import org.helioviewer.viewmodel.imageformat.SingleChannelImageFormat;
import org.helioviewer.viewmodel.imagetransport.ImageTransport;
import org.helioviewer.viewmodel.imagetransport.Short16ImageTransport;

/** Representation of image data in single channel format, using 9 to 16 bits per pixel.
 * 
 * <p>Note that this is the only implementation of ImageData, which is supposed to
 * handle a variable number of bits per pixel.
 * 
 * @author Ludwig Schmidt
 * @author Markus Langenberg
 *
 */
public class SingleChannelShortImageData implements JavaBufferedImageData {

	private SingleChannelImageFormat format;
	private Short16ImageTransport imageTransport;
    private int width, height;
    private BufferedImage image = null;

    /** Constructor, given an array as data source.
     * 
     * <p>This constructor receives the raw data as a data source.
     * If the caller handles raw data as well, the use of this constructor
     * is recommended.
     * <p>The pixel data has to be given as a one-dimensional array containing
     * the pixel data line by line. Each array element represents one pixel.
     * 
     * @param newWidth width of the image
     * @param newHeight height of the image
     * @param newBitDepth number of bits per pixel
     * @param newPixelData pixel data
     */
    public SingleChannelShortImageData(int newWidth, int newHeight,
            int newBitDepth, short[] newPixelData) {
        
    	width = newWidth;
        height = newHeight;
        imageTransport = new Short16ImageTransport(newPixelData);
        format = new SingleChannelImageFormat(newBitDepth);
    }

    /** Constructor, given an BufferedImage as data source.
     * 
     * <p>This constructor receives a BufferedImage as data source.
     * If the caller operates on BufferedImages as well, the use of this
     * constructor is recommended.
     * 
     * @param newImage pixel data
     */
    public SingleChannelShortImageData(int newBitDepth, BufferedImage newImage) {
        image = newImage;
        width = newImage.getWidth();
    	height = newImage.getHeight();
    	imageTransport = new Short16ImageTransport(((DataBufferUShort) newImage.getRaster().getDataBuffer()).getData());
        format = new SingleChannelImageFormat(newBitDepth);
    }

    /**
	 * {@inheritDoc}
	 */
	public BufferedImage getBufferedImage() {
    	if(image == null) {
    		image = new BufferedImage(width, height, BufferedImage.TYPE_USHORT_GRAY);
    		DataBufferUShort dataBuffer = new DataBufferUShort(imageTransport.getShort16PixelData(), width * height);

            // create the appropriate bit mask
            int mask = 0xffffffff;
            mask = mask >>> (32 - format.getBitDepth());

            Raster raster = Raster.createPackedRaster(dataBuffer, width,
            		height, width, new int[] { mask }, new Point(0, 0));
            image.setData(raster);
    	}
        return image;
    }

	/**
	 * {@inheritDoc}
	 */
	public int getHeight() {
        return height;
    }

	/**
	 * {@inheritDoc}
	 */
	public ImageFormat getImageFormat() {
        return format;
    }

	/**
	 * {@inheritDoc}
	 */
	public ImageTransport getImageTransport() {
        return imageTransport;
    }

	/**
	 * {@inheritDoc}
	 */
	public int getWidth() {
        return width;
    }
}
