package org.helioviewer.viewmodel.metadata;

import org.helioviewer.base.math.Vector2dDouble;
import org.helioviewer.base.math.Vector2dInt;
import org.helioviewer.base.physics.Constants;
import org.helioviewer.viewmodel.view.cache.HelioviewerDateTimeCache;
import org.helioviewer.viewmodel.view.jp2view.datetime.ImmutableDateTime;

/** Implementation of MetaData representing solar images.
 * 
 * <p>This class is supposed to be for solar images.
 * Currently, it supports the observatory SOHO with its
 * instruments EIT, LASCO and MDI, as well as some
 * instruments on board of the observatory STEREO. 
 * 
 * @author Ludwig Schmidt
 *
 */
public class HelioviewerMetaData extends AbstractMetaData implements
        SunMetaData, ObserverMetaData, ImageSizeMetaData {

	private MetaDataContainer metaDataContainer;
    private String instrument;
    private String detector;
    private String measurement = " ";
    private String observatory = " ";
    private String fullName;
    private Vector2dInt pixelImageSize;
    private double solarPixelRadius = -1;
    private Vector2dDouble sunPixelPosition;

    private double meterPerPixel;
    private ImmutableDateTime time;

    /** Default constructor.
     * 
     * Tries to read all informations required.
     * 
     * @param m Meta data container serving as a base for the construction
     */
    public HelioviewerMetaData(MetaDataContainer m) {

    	metaDataContainer = m;
    	
    	if (m.get("INSTRUME") == null)
    	    return;
    	
        detector = m.get("DETECTOR");
        instrument = m.get("INSTRUME");
        double sunX = m.tryGetDouble("CRPIX1");
        double sunY = m.tryGetDouble("CRPIX2");
        sunPixelPosition = new Vector2dDouble(sunX, sunY);

        pixelImageSize = new Vector2dInt(m.getPixelWidth(), m.getPixelHeight());

        if (detector == null) {
            detector = " ";
        }
        if (instrument == null) {
            instrument = " ";
        }

        if (instrument.equals("EIT")) {
            measurement = m.get("WAVELNTH");
            if (measurement == null) {
                measurement = "" + m.tryGetInt("WAVELNTH");
            }
            solarPixelRadius = m.tryGetDouble("SOLAR_R");
            observatory = m.get("TELESCOP");
            fullName = "EIT " + measurement;
        }

        else if (detector.equals("C2") || detector.equals("C3")) {

            String measurement1 = m.get("FILTER");
            String measurement2 = m.get("POLAR");
            measurement = measurement1 + " " + measurement2;
            solarPixelRadius = m.tryGetDouble("RSUN");

            if (solarPixelRadius == 0) {
                if (detector.equals("C2")) {
                    solarPixelRadius = 80.814221;
                } else if (detector.equals("C3")) {
                    solarPixelRadius = 17.173021;
                }

            }

            observatory = m.get("TELESCOP");
            fullName = "LASCO " + detector;
        }

        else if (instrument.equals("MDI")) {
            solarPixelRadius = m.tryGetDouble("R_SUN");
            observatory = m.get("TELESCOP");
            measurement = m.get("DPC_OBSR");
            fullName = "MDI " + measurement.substring(3,6);
        }

        else if (detector.equals("COR1") || detector.equals("COR2")
                || detector.equals("EUVI")) {
            observatory = m.get("OBSRVTRY");
            measurement = m.get("WAVELNTH");
            if (measurement == null) {
                measurement = "" + m.tryGetDouble("WAVELNTH");
            }
            solarPixelRadius = m.tryGetDouble("RSUN");
            fullName = instrument + " " + detector;
        } else {
            System.out.println("this  type of FITS is not yet supportet!");
        }
        
        meterPerPixel = Constants.SunRadius / solarPixelRadius;
        
        setPhysicalLowerLeftCorner(sunPixelPosition.scale(-meterPerPixel));
        setPhysicalImageSize(new Vector2dDouble(pixelImageSize.getX()
                * meterPerPixel, pixelImageSize.getY() * meterPerPixel));

        updateDateTime();
    }
    
    /**
     * Copy constructor
     * 
     * @param original Object to copy
     */
    public HelioviewerMetaData(HelioviewerMetaData original) {
    	super(original);
    	metaDataContainer = original.metaDataContainer;
        instrument = new String(original.instrument);
        detector = new String(original.detector);
        measurement = new String(original.measurement);
        observatory = new String(original.observatory);
        fullName = new String(original.fullName);
        pixelImageSize = new Vector2dInt(original.pixelImageSize);
        solarPixelRadius = original.solarPixelRadius;
        sunPixelPosition = new Vector2dDouble(original.sunPixelPosition);
        meterPerPixel = original.meterPerPixel;
        time = new ImmutableDateTime(original.time);
    }
    
    /** 
     * {@inheritDoc}
     */
    public void updateDateTime() {
    	String observedDate = metaDataContainer.get("DATE_OBS");
    	
    	if (detector.equals("C2") || detector.equals("C3")) {
    		observedDate += "T" + metaDataContainer.get("TIME_OBS");
    	}
    	
    	time = HelioviewerDateTimeCache.parseDateTime(observedDate);
    }
    
    /** 
     * {@inheritDoc}
     */
    public void updateDateTime(ImmutableDateTime newDateTime) {
    	time = newDateTime;
    }

    /**
     * {@inheritDoc}
     */
    public String getDetector() {
        return detector;
    }

    /**
     * {@inheritDoc}
     */
    public String getInstrument() {
        return instrument;
    }

    /**
     * {@inheritDoc}
     */
    public String getMeasurement() {
        return measurement;
    }

    /**
     * {@inheritDoc}
     */
    public String getObservatory() {
        return observatory;
    }
    
    public String getFullName() {
    	return fullName;
    }

    /**
     * {@inheritDoc}
     */
    public double getSunPixelRadius() {
        return solarPixelRadius;
    }

    /**
     * {@inheritDoc}
     */
    public Vector2dDouble getSunPixelPosition() {
        return sunPixelPosition;
    }

    /**
     * {@inheritDoc}
     */
    public Vector2dInt getResolution() {
        return pixelImageSize;
    }

    /**
     * {@inheritDoc}
     */
    public double getUnitsPerPixel() {
        return meterPerPixel;
    }
    
    /**
     * {@inheritDoc}
     */
    public void updateUnitsPerPixel(float scaleFactor) {
    	meterPerPixel *= scaleFactor;
    	solarPixelRadius = Constants.SunRadius / meterPerPixel;
    	sunPixelPosition = sunPixelPosition.scale(scaleFactor);
    }

    /**
     * {@inheritDoc}
     */
    public ImmutableDateTime getDateTime() {
        return time;
    }

}
