package org.helioviewer.viewmodel.renderer;

import java.awt.image.BufferedImage;
import java.util.HashMap;

import javax.media.opengl.GL;

import org.helioviewer.viewmodel.view.opengl.GLTextureHelper;
import org.helioviewer.viewmodel.view.opengl.shader.GLScalePowerOfTwoVertexShaderProgram;
import org.helioviewer.viewmodel.view.opengl.shader.GLTextureCoordinate;

/** Class containing OpenGL specific render functions equal for both
 * renderer types.
 * 
 * This class should only be used by {@link org.helioviewer.viewmodel
 * .renderer.physical.GLPhysicalRenderGraphics} and {@link 
 * org.helioviewer.viewmodel.renderer.screen.GLScreenRenderGraphics}
 * 
 * @author Markus Langenberg
 */
public class GLCommonRenderGraphics {

	private static GLScalePowerOfTwoVertexShaderProgram scalingShader;
	private static GLTextureCoordinate texCoord = new GLTextureHelper.GLMainTextureCoordinate();
	
	private GL gl;
	private static GLTextureHelper textureHelper = new GLTextureHelper();
	private static HashMap<BufferedImage, Integer> mapImageToTexture = new HashMap<BufferedImage, Integer>();
	
	
	/** Default constructor.
	 * 
	 * <p>The caller has to provide a gl object, which
	 * can be used by this renderer.
	 * 
	 * @param _gl gl object, that should be used for drawing.
	 */
	public GLCommonRenderGraphics(GL _gl) {
		gl = _gl;
		
		if(!GLTextureHelper.textureNonPowerOfTwoAvailable() && scalingShader == null) {
			scalingShader = new GLScalePowerOfTwoVertexShaderProgram();
			scalingShader.buildStandAlone(_gl);
		}
	}
	
	/** Sets the standard texture coordinate.
	 * 
	 * This function has to be called instead of the pure
	 * OpenGL-function to be able to use {@link org.helioviewer
	 * .viewmodel.view.opengl.shader.GLScalePowerOfTwoVertexShaderProgram}.
	 *  
	 * @param x X-coordinate to set
	 * @param y Y-coordinate to set
	 */
	public void setTexCoord(float x, float y) {
		texCoord.setValue(gl, x, y);
	}
	
	/** Binds the texture corresponding to the given image.
	 * 
	 * If the image never has been used before, a new texture
	 * is created, the texture will be saved after rendering to
	 * use it again later. If more than 256 textures have been
	 * generated, the buffer is cleared.
	 * 
	 * @param image Image corresponding to the texture to bind
	 */
	public void bindImage(BufferedImage image) {
		Integer texID = mapImageToTexture.get(image);
		if(texID == null) {
			if(mapImageToTexture.size() > 256) {
				clearTextureBuffer(gl);
			}
			
			texID = textureHelper.genTextureID(gl);
			textureHelper.moveBufferedImageToGLTexture(gl, image, texID.intValue());
			mapImageToTexture.put(image, texID);
		}
		
		textureHelper.bindTexture(gl, texID.intValue());
	}
	
	/** Binds the scaling shader.
	 * 
	 * This should be done before using a texture. After using the
	 * texture, {@link #unbindScalingShader()} should be called.
	 * 
	 * <p>This function only has an effect, if non-power-of-two-
	 * textures are disabled.
	 */
	public void bindScalingShader() {
		if(GLTextureHelper.textureNonPowerOfTwoAvailable())
			return;
		
		gl.glEnable(GL.GL_VERTEX_PROGRAM_ARB);
		scalingShader.bind(gl);
	}
	
	/** Unbinds the scaling shader.
	 * 
	 * This is the corresponding call to {@link #bindScalingShader()}.
	 */
	public void unbindScalingShader() {
		gl.glDisable(GL.GL_VERTEX_PROGRAM_ARB);
	}
	
	/** Clears all saved textures.
	 * 
	 * This might be useful to clean up after switching drawing modes.
	 * 
	 * @param gl valid gl object to use.
	 */
	public static void clearTextureBuffer(GL gl) {
		for(Integer i : mapImageToTexture.values()) {
			textureHelper.delTextureID(gl, i);
		}
		mapImageToTexture.clear();
	}
	
	/**
	 * Clears the shader shared by this render graphics objects.
	 */
	public static void clearShader() {
		scalingShader = null;
	}
}
