package org.helioviewer.viewmodel.renderer.physical;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;

import org.helioviewer.base.math.Vector2dInt;
import org.helioviewer.viewmodel.view.View;

/** Implementation of ScreenRenderGraphics operating on Java BufferedImages.
 * 
 * <p>Maps all methods to corresponding methods of Graphics2D.
 * 
 * @author Stephan Pagel
 * @author Markus Langenberg
 * 
 */
public class BufferedImagePhysicalRenderGraphics extends AbstractPhysicalRenderGraphics {

	private Graphics2D graphics;
	
	/** Default constructor.
	 * 
	 * <p>The caller has to provide a Java Graphics2D object, which
	 * can be used by this renderer.
	 * 
	 * @param g Java Graphics2D object, that should be used for drawing.
	 * @param view View to access information about the physical coordinate system.
	 */
	public BufferedImagePhysicalRenderGraphics(Graphics g, View view) {
		super(view);
		graphics = (Graphics2D) g;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void drawLine(double x0, double y0, double x1, double y1) {
		Vector2dInt p0 = convertPhysicalToScreen(x0, y0);
		Vector2dInt p1 = convertPhysicalToScreen(x1, y1);
		graphics.drawLine(p0.getX(), p0.getY(), p1.getX(), p1.getY());
	}

	/**
	 * {@inheritDoc}
	 */
	public void drawRectangle(double x, double y, double width, double height) {
		Vector2dInt pos = convertPhysicalToScreen(x, y);
		Vector2dInt size = convertPhysicalToScreen(width, height);
		graphics.drawRect(pos.getX() - (size.getX() >> 1), pos.getY() - (size.getY() >> 1), size.getX(), size.getY());
	}

	/**
	 * {@inheritDoc}
	 */
	public void fillRectangle(double x, double y, double width, double height) {
		Vector2dInt pos = convertPhysicalToScreen(x, y);
		Vector2dInt size = convertPhysicalToScreen(width, height);
		graphics.fillRect(pos.getX() - (size.getX() >> 1), pos.getY() - (size.getY() >> 1), size.getX(), size.getY());
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void drawOval(double x, double y, double width, double height) {
		Vector2dInt pos = convertPhysicalToScreen(x, y);
		Vector2dInt size = convertPhysicalToScreen(width, height);
		graphics.drawOval(pos.getX() - (size.getX() >> 1), pos.getY() - (size.getY() >> 1), size.getX(), size.getY());
	}

	/**
	 * {@inheritDoc}
	 */
	public void fillOval(double x, double y, double width, double height) {
		Vector2dInt pos = convertPhysicalToScreen(x, y);
		Vector2dInt size = convertPhysicalToScreen(width, height);
		graphics.fillOval(pos.getX() - (size.getX() >> 1), pos.getY() - (size.getY() >> 1), size.getX(), size.getY());
	}

	/**
	 * {@inheritDoc}
	 */
	public void setColor(Color color) {
		graphics.setColor(color);
	}

	/**
	 * {@inheritDoc}
	 */
	public void setLineWidth(float lineWidth) {
		graphics.setStroke(new BasicStroke(lineWidth));
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void drawImage(BufferedImage image, double x, double y) {
		Vector2dInt pos = convertPhysicalToScreen(x, y);
		graphics.drawImage(image, pos.getX() - (image.getWidth() >> 1), pos.getY() - (image.getHeight() >> 1), null);
	}

	/**
	 * {@inheritDoc}
	 */
	public void drawImage(BufferedImage image, double x, double y, double width, double height) {
		Vector2dInt pos = convertPhysicalToScreen(x, y);
		Vector2dInt size = convertPhysicalToScreen(width, height);
		graphics.drawImage(image, pos.getX() - (size.getX() >> 1), pos.getY() - (size.getY() >> 1), size.getX(), size.getY(), null);	
	}

}
