package org.helioviewer.viewmodel.renderer.physical;

import java.awt.Color;
import java.awt.image.BufferedImage;

import org.helioviewer.base.math.Vector2dDouble;

/** Interface for a renderer with basic graphical drawing functions.
 *
 * <p>An implementation of this interface will provided for every
 * {@link PhysicalRenderer}, so it can draw whatever needed.
 * 
 * <p>All coordinates shall be given as physical coordinates, using
 * kilometers as unit.
 * 
 * @author Stephan Pagel
 * @author Markus Langenberg
 * */
public interface PhysicalRenderGraphics {

	/** Sets the color used for drawing lines and rectangles
	 * 
	 * @param color new color
	 */
	public void setColor(Color color);
	
	/** Sets the line width used for drawing lines and rectangles
	 * 
	 * @param lineWidth new line width
	 */
	public void setLineWidth(float lineWidth);
	
	/** Draws a line from one point to another.
	 * 
	 * @param x0 x-coordinate from starting point
	 * @param y0 y-coordinate from starting point
	 * @param x1 x-coordinate from ending point
	 * @param y1 y-coordinate from ending point
	 */
	public void drawLine(double x0, double y0, double x1, double y1);
	
	/** Draws a line from one point to another.
	 * 
	 * @param p0 starting point
	 * @param p1 ending point
	 */
	public void drawLine(Vector2dDouble p0, Vector2dDouble p1);
	
	/** Draws a rectangle
	 * 
	 * @param x x-coordinate of center
	 * @param y y-coordinate of center
	 * @param width width of the rectangle
	 * @param height height of the rectangle
	 */
	public void drawRectangle(double x, double y, double width, double height);
	
	/** Draws a rectangle
	 * 
	 * @param position of center
	 * @param size size of the rectangle
	 */
	public void drawRectangle(Vector2dDouble position, Vector2dDouble size);
	
	/** Fills a rectangle
	 * 
	 * @param x x-coordinate of center
	 * @param y y-coordinate of center
	 * @param width width of the rectangle
	 * @param height height of the rectangle
	 */
	public void fillRectangle(double x, double y, double width, double height);
	
	/** Fills a rectangle
	 * 
	 * @param position center position of the rectangle
	 * @param size size of the rectangle
	 */
	public void fillRectangle(Vector2dDouble position, Vector2dDouble size);
	
	/** Draws an oval
	 * 
	 * @param x x-coordinate of center
	 * @param y y-coordinate of center
	 * @param width width of the oval
	 * @param height height of the oval
	 */
	public void drawOval(double x, double y, double width, double height);
	
	/** Draws an oval
	 * 
	 * @param position center position
	 * @param size size of the oval
	 */
	public void drawOval(Vector2dDouble position, Vector2dDouble size);
	
	/** Fills an oval
	 * 
	 * @param x x-coordinate of center
	 * @param y y-coordinate of center
	 * @param width width of the oval
	 * @param height height of the oval
	 */
	public void fillOval(double x, double y, double width, double height);
	
	/** Fills an oval
	 * 
	 * @param position center position
	 * @param size size of the oval
	 */
	public void fillOval(Vector2dDouble position, Vector2dDouble size);
	
	/** Draws an unscaled image
	 * 
	 * @param image BufferedImage to draw
	 * @param x x-coordinate of center
	 * @param y y-coordinate of center
	 */
	public void drawImage(BufferedImage image, double x, double y);
	
	/** Draws an unscaled image
	 * 
	 * @param image BufferedImage to draw
	 * @param position center position of the image
	 */
	public void drawImage(BufferedImage image, Vector2dDouble position);
	
	/** Draws an scaled image
	 * 
	 * @param image BufferedImage to draw
	 * @param x x-coordinate of center
	 * @param y y-coordinate of center
	 * @param width width of the rectangle the image is drawn to
	 * @param height height of the rectangle the image is drawn to
	 */
	public void drawImage(BufferedImage image, double x, double y, double width, double height);
	
	/** Draws an scaled image
	 * 
	 * @param image BufferedImage to draw
	 * @param position center position of the image
	 * @param size size of the rectangle the image is drawn to
	 */
	public void drawImage(BufferedImage image, Vector2dDouble position, Vector2dDouble size);
}
