package org.helioviewer.viewmodel.renderer.screen;

import java.awt.Color;
import java.awt.image.BufferedImage;

import org.helioviewer.base.math.Vector2dInt;

/** Interface for a renderer with basic graphical drawing functions.
 * 
 * <p>An implementation of this interface will provided for every
 * {@link ScreenRenderer}, so it can draw whatever needed.
 * 
 * <p>All coordinates shall be given as screen coordinates, using
 * pixels as unit.
 * 
 * @author Stephan Pagel
 * @author Markus Langenberg
 *
 */
public interface ScreenRenderGraphics {

	/** Sets the color used for drawing lines and rectangles
	 * 
	 * @param color new color
	 */
	public void setColor(Color color);
	
	/** Sets the line width used for drawing lines and rectangles
	 * 
	 * @param lineWidth new line width
	 */
	public void setLineWidth(float lineWidth);
	
	/** Draws a line from one point to another.
	 * 
	 * @param x0 x-coordinate of starting point
	 * @param y0 y-coordinate of starting point
	 * @param x1 x-coordinate of ending point
	 * @param y1 y-coordinate of ending point
	 */
	public void drawLine(int x0, int y0, int x1, int y1);
	
	/** Draws a line from one point to another.
	 * 
	 * @param p0 starting point
	 * @param p1 ending point
	 */
	public void drawLine(Vector2dInt p0, Vector2dInt p1);
	
	/** Draws a rectangle
	 * 
	 * @param x x-coordinate of upper left corner
	 * @param y y-coordinate of upper left corner
	 * @param width width of the rectangle
	 * @param height height of the rectangle
	 */
	public void drawRectangle(int x, int y, int width, int height);
	
	/** Draws a rectangle
	 * 
	 * @param position upper left corner
	 * @param size size of the rectangle
	 */
	public void drawRectangle(Vector2dInt position, Vector2dInt size);
	
	/** Fills a rectangle
	 * 
	 * @param x x-coordinate of upper left corner
	 * @param y y-coordinate of upper left corner
	 * @param width width of the rectangle
	 * @param height height of the rectangle
	 */
	public void fillRectangle(int x, int y, int width, int height);
	
	/** Fills a rectangle
	 * 
	 * @param position upper left corner
	 * @param size size of the rectangle
	 */
	public void fillRectangle(Vector2dInt position, Vector2dInt size);
	
	/** Draws an oval
	 * 
	 * @param x x-coordinate of upper left corner
	 * @param y y-coordinate of upper left corner
	 * @param width width of the oval
	 * @param height height of the oval
	 */
	public void drawOval(int x, int y, int width, int height);
	
	/** Draws an oval
	 * 
	 * @param position upper left corner
	 * @param size size of the oval
	 */
	public void drawOval(Vector2dInt position, Vector2dInt size);
	
	/** Fills an oval
	 * 
	 * @param x x-coordinate of upper left corner
	 * @param y y-coordinate of upper left corner
	 * @param width width of the oval
	 * @param height height of the oval
	 */
	public void fillOval(int x, int y, int width, int height);
	
	/** Fills an oval
	 * 
	 * @param position upper left corner
	 * @param size size of the oval
	 */
	public void fillOval(Vector2dInt position, Vector2dInt size);
	
	/** Draws an unscaled image
	 * 
	 * @param image BufferedImage to draw
	 * @param x x-coordinate of upper left corner
	 * @param y y-coordinate of upper left corner
	 */
	public void drawImage(BufferedImage image, int x, int y);
	
	/** Draws an unscaled image
	 * 
	 * @param image BufferedImage to draw
	 * @param position upper left corner
	 */
	public void drawImage(BufferedImage image, Vector2dInt position);
	
	/** Draws an scaled image
	 * 
	 * @param image BufferedImage to draw
	 * @param x x-coordinate of upper left corner
	 * @param y y-coordinate of upper left corner
	 * @param width width of the rectangle the image is drawn to
	 * @param height height of the rectangle the image is drawn to
	 */
	public void drawImage(BufferedImage image, int x, int y, int width, int height);
	
	/** Draws an scaled image
	 * 
	 * @param image BufferedImage to draw
	 * @param position upper left corner
	 * @param size size of the rectangle the image is drawn to
	 */
	public void drawImage(BufferedImage image, Vector2dInt position, Vector2dInt size);
}
