package org.helioviewer.viewmodel.view.bufferedimage;

import java.awt.AlphaComposite;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;

import org.helioviewer.base.math.Vector2dInt;
import org.helioviewer.base.physics.Constants;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.changeevent.SubImageDataChangedReason;
import org.helioviewer.viewmodel.changeevent.ViewChainChangedReason;
import org.helioviewer.viewmodel.imagedata.ARGBInt32ImageData;
import org.helioviewer.viewmodel.imagedata.ImageData;
import org.helioviewer.viewmodel.imagedata.JavaBufferedImageData;
import org.helioviewer.viewmodel.metadata.HelioviewerMetaData;
import org.helioviewer.viewmodel.metadata.HelioviewerOcculterMetaData;
import org.helioviewer.viewmodel.view.AbstractBasicView;
import org.helioviewer.viewmodel.view.HelioviewerGeometryView;
import org.helioviewer.viewmodel.view.MetaDataView;
import org.helioviewer.viewmodel.view.RegionView;
import org.helioviewer.viewmodel.view.SubimageDataView;
import org.helioviewer.viewmodel.view.View;
import org.helioviewer.viewmodel.view.ViewHelper;
import org.helioviewer.viewmodel.viewportimagesize.StaticViewportImageSize;
import org.helioviewer.viewmodel.viewportimagesize.ViewportImageSize;

/** Implementation of HelioviewGeometryView for rendering in OpenGL mode.
 * 
 * <p>This class provides the capability to cut out the invalid areas in
 * solar images. It does so by calculating the distance from the center 
 * for every single pixel in the image. If the distance is outside the 
 * valid area of that specific image, its alpha value is set to zero,
 * otherwise it remains untouched.
 * 
 * <p>Technically, it uses the Java AlphaComposite to mask invalid areas.
 * 
 * <p>For further information about the role of the HelioviewerGeometryView
 * within the view chain, see
 * {@link org.helioviewer.viewmodel.view.HelioviewerGeometryView}
 * 
 * @author Markus Langenberg
 */
public class BufferedImageHelioviewerGeometryView extends AbstractBasicView
		implements SubimageDataView, HelioviewerGeometryView {
	
	private RegionView regionView;
    private SubimageDataView subimageDataView;
    private ImageData imageData;
    private BufferedImage maskImage;
    private HelioviewerMetaData metaData;
    
    /**
     * {@inheritDoc}
     */
    protected void setViewSpecificImplementation(View newView, ChangeEvent changeEvent) {
    	updatePrecomputed();
    	
		if (subimageDataView != null) {
            redraw();
            changeEvent.addReason(new SubImageDataChangedReason(this));
        }
	}
	
    /**
     * {@inheritDoc}
     */
    public ImageData getSubimageData() {
        return imageData;
    }

    /**
     * {@inheritDoc}
     */
    public void viewChanged(View sender, ChangeEvent aEvent) {
    	
    	if(aEvent.reasonOccurred(ViewChainChangedReason.class)) {
    		updatePrecomputed();
    	}
    	
    	if(aEvent.reasonOccurred(SubImageDataChangedReason.class)) {
    		redraw();
    	}
    	
    	notifyViewListeners(aEvent);
    }
    
    /**
     * Updates the precomputed view adapters.
     */
    private void updatePrecomputed() {
    	regionView = ViewHelper.getViewAdapter(view, RegionView.class);
		subimageDataView = ViewHelper.getViewAdapter(view, SubimageDataView.class);
		
		MetaDataView metaDataView = ViewHelper.getViewAdapter(view, MetaDataView.class);
		if(metaDataView != null && metaDataView.getMetaData() != null && metaDataView.getMetaData() instanceof HelioviewerMetaData) {
			metaData = (HelioviewerMetaData) metaDataView.getMetaData();
		}
		
		redrawMask();
    }
    
    /** Redraws the mask used for alpha masking.
     * 
     * The mask is an image containing transparent areas where the actual
     * image contains invalid areas. The two images are merged using an
     * AlphaComposite.
     */
    private void redrawMask() {
    	ImageData data = subimageDataView.getSubimageData();
    	
    	if(data instanceof JavaBufferedImageData && metaData != null) {
    		maskImage = new BufferedImage(data.getWidth(), data.getHeight(), BufferedImage.TYPE_INT_ARGB);
    		
    		//ViewportImageSize viewportImageSize = ViewHelper.calculateViewportImageSize(viewportView.getViewport(), regionView.getRegion());
    		ViewportImageSize viewportImageSize = StaticViewportImageSize.createAdaptedViewportImageSize(data.getWidth(), data.getHeight());
    		
    		Vector2dInt offset = ViewHelper.convertImageToScreenDisplacement(
        				-regionView.getRegion().getUpperLeftCorner().getX(),
        				regionView.getRegion().getUpperLeftCorner().getY(),
        				regionView.getRegion(), viewportImageSize);
    		
    		Vector2dInt radius;
    		
    		if(metaData instanceof HelioviewerOcculterMetaData) {
    			radius = ViewHelper.convertImageToScreenDisplacement(
    					((HelioviewerOcculterMetaData) metaData).getInnerPhysicalOcculterRadius() * roccInnerFactor, 
    					((HelioviewerOcculterMetaData) metaData).getOuterPhysicalOcculterRadius() * roccOuterFactor,
    					regionView.getRegion(), viewportImageSize);
    		} else {
    			radius = ViewHelper.convertImageToScreenDisplacement(
    					0, Constants.SunRadius * discFactor,
    					regionView.getRegion(), viewportImageSize);
    		}
    		
    		Graphics2D g = maskImage.createGraphics();
			g.setComposite(AlphaComposite.Clear);
			g.fillRect(0, 0, data.getWidth(), data.getHeight());
			g.setComposite(AlphaComposite.Src);
			g.fillOval(offset.getX() - radius.getY(), offset.getY() - radius.getY(), radius.getY() * 2, radius.getY() * 2);
			if(radius.getX() != 0) {
				g.setComposite(AlphaComposite.Clear);
				g.fillOval(offset.getX() - radius.getX(), offset.getY() - radius.getX(), radius.getX() * 2, radius.getX() * 2);
			}
			g.dispose();
    	}
    }
    
    /** Redraws the masked image.
     * 
     * To turn the invalid areas transparent, the image is masked with
     * a precalculated mask using AlphaComposite.
     */
    private void redraw() {
    	
    	ImageData data = subimageDataView.getSubimageData();
    	
    	if(data instanceof JavaBufferedImageData && metaData != null) {
			BufferedImage source = ((JavaBufferedImageData) data).getBufferedImage();
			BufferedImage target = new BufferedImage(source.getWidth(), source.getHeight(), BufferedImage.TYPE_INT_ARGB);
			
			redrawMask();
			
			Graphics2D g = target.createGraphics();
			g.setComposite(AlphaComposite.Clear);
			g.fillRect(0, 0, data.getWidth(), data.getHeight());
			g.setComposite(AlphaComposite.Src);
			g.drawImage(source, 0, 0, null);
			g.setComposite(AlphaComposite.DstIn);
			g.drawImage(maskImage, 0, 0, null);
			g.dispose();
			
			imageData = new ARGBInt32ImageData(target);
    	} else {
    		imageData = data;
    	}
    }
}
