package org.helioviewer.viewmodel.view.jp2view;

import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.changeevent.MetaDataChangedReason;
import org.helioviewer.viewmodel.imagedata.ImageData;
import org.helioviewer.viewmodel.metadata.ObserverMetaData;
import org.helioviewer.viewmodel.view.CachedMovieView;
import org.helioviewer.viewmodel.view.LinkedMovieManager;
import org.helioviewer.viewmodel.view.TimedMovieView;
import org.helioviewer.viewmodel.view.cache.DateTimeCache;
import org.helioviewer.viewmodel.view.cache.HelioviewerDateTimeCache;
import org.helioviewer.viewmodel.view.cache.ImageCacheStatus;
import org.helioviewer.viewmodel.view.cache.LocalImageCacheStatus;
import org.helioviewer.viewmodel.view.cache.RemoteImageCacheStatus;
import org.helioviewer.viewmodel.view.jp2view.J2KRender.RenderReasons;
import org.helioviewer.viewmodel.view.jp2view.datetime.ImmutableDateTime;
import org.helioviewer.viewmodel.view.jp2view.image.JP2ImageParameter;
import org.helioviewer.viewmodel.view.jp2view.image.SubImage;

/** Implementation of TimedMovieView for JPX files.
 * 
 * <p>This class is an extions of {@link JHVJP2View} for JPX-Files,
 * providing additional movie commands.
 * 
 * <p>For information about image series, see
 * {@link org.helioviewer.viewmodel.view.MovieView} and
 * {@link org.helioviewer.viewmodel.view.TimedMovieView}.
 *  
 * @author Markus Langenberg
 */
public class JHVJPXView extends JHVJP2View implements TimedMovieView, CachedMovieView {
	
	// Caching
	protected ImageCacheStatus imageCacheStatus;
	protected DateTimeCache dateTimeCache;
	protected int lastRenderedCompositionLayer = -1;
	
	// Linking movies
	protected LinkedMovieManager linkedMovieManager; // if the move is not linked, this has to be null
	
	/**
     * Default constructor.
     * 
     * <p>When the view is not marked as a main view, it is
     * assumed, that the view will only serve one single image
     * and will not have to perform any kind of update any more.
     * The effect of this assumption is, that the view will not
     * try to reconnect to the JPIP server when the connection
     * breaks and that there will be no other timestamps used
     * than the first one.
     * 
     * @param isMainView Whether the view is a main view or not
     */
    public JHVJPXView(boolean isMainView) {
    	super(isMainView);
    }

    /**
     * {@inheritDoc}
     */
    public void setJP2Image(JP2Image newJP2Image) {
    	if(!isMainView) {
    		super.setJP2Image(newJP2Image);
    		imageCacheStatus = ((JHVJPXView) jp2Image.getParentView()).getImageCacheStatus();
    		dateTimeCache = ((JHVJPXView) jp2Image.getParentView()).getDateTimeCache();
    		return;
    	}
    	
    	if(jp2Image != null) {
        	abolish();
        }
    	
    	jp2Image = newJP2Image;
    	
    	if(newJP2Image.isRemote()) {
    		imageCacheStatus = new RemoteImageCacheStatus(this);
    	} else {
    		imageCacheStatus = new LocalImageCacheStatus(this);
    	}
    	jp2Image.setImageCacheStatus(imageCacheStatus);
    	
    	dateTimeCache = new HelioviewerDateTimeCache(this, jp2Image);
	
    	super.setJP2Image(newJP2Image);
	
    	dateTimeCache.startParsing();
    }
    
    /**
     * {@inheritDoc}
     */
    public DateTimeCache getDateTimeCache() {
    	return dateTimeCache;
	}

    /**
     * {@inheritDoc}
     */
    public ImageCacheStatus getImageCacheStatus() {
    	return imageCacheStatus;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void setCurrentFrame(int frameNumber, ChangeEvent event) {
		
		boolean changed;
		
		if(frameNumber <= 0) {
			changed = setCurrentFrameNumber(0, event);
		} else if(frameNumber > getMaximumFrameNumber()) {
			changed = setCurrentFrameNumber(getMaximumFrameNumber(), event);
		} else {
			changed = setCurrentFrameNumber(frameNumber, event);
		}
		
		if(changed) {
			if(linkedMovieManager != null) {
				linkedMovieManager.setCurrentFrame(getCurrentFrameDateTime(), event);
			}
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void setCurrentFrame(ImmutableDateTime time, ChangeEvent event) {
		
		if(time == null)
			return;
		
		if(linkedMovieManager != null && linkedMovieManager.setCurrentFrame(time, event)) {
			return;
		}
		
		int frameNumber = -1;
		long timeMillis = time.getMillis();
		long lastDiff, currentDiff = -Long.MAX_VALUE;
		
		do {
			lastDiff = currentDiff;
			
			if(dateTimeCache.getDateTime(++frameNumber) == null) {
				return;
			}
			
			currentDiff = dateTimeCache.getDateTime(frameNumber).getMillis() - timeMillis;
		} while(currentDiff < 0 && frameNumber < jp2Image.getCompositionLayerRange().rightSide);
		
		if(-lastDiff < currentDiff) {
			setCurrentFrameNumber(frameNumber-1, event);
		} else {
			setCurrentFrameNumber(frameNumber, event);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public int getCurrentFrameNumber() {
    	return imageViewParams.compositionLayer;
    }
	
	/**
	 * {@inheritDoc}
	 */
	public int getMaximumFrameNumber() {
		return jp2Image.getCompositionLayerRange().rightSide;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public int getMaximumAccessibleFrameNumber() {
		return Math.min(dateTimeCache.getMetaStatus(), imageCacheStatus.getImageCachedPartiallyUntil());
	}
	
	/**
	 * {@inheritDoc}
	 */
	public ImmutableDateTime getCurrentFrameDateTime() {
		return dateTimeCache.getDateTime(getCurrentFrameNumber());
	}
	
	/**
	 * {@inheritDoc}
	 */
	public ImmutableDateTime getFrameDateTime(int frameNumber) {
		return dateTimeCache.getDateTime(frameNumber);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void setAnimationMode(AnimationMode mode) {
		render.setAnimationMode(mode);
	}

	/**
	 * {@inheritDoc}
	 */
	public void setDesiredRelativeSpeed(int framesPerSecond) {
		render.setMovieRelativeSpeed(framesPerSecond);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void setDesiredAbsoluteSpeed(int observationSecondsPerSecond) {
		render.setMovieAbsoluteSpeed(observationSecondsPerSecond);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void linkMovie() {
		linkedMovieManager = LinkedMovieManager.getActiveInstance();
		linkedMovieManager.linkMovie(this);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void unlinkMovie() {
		LinkedMovieManager temp = linkedMovieManager;
		linkedMovieManager = null;
		temp.unlinkMovie(this);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public float getActualFramerate() {
	    
	    if (render != null)
	        return render.getActualMovieFramerate();
	    
	    return 0;
	}

	/**
	 * {@inheritDoc}
	 */
	public void pauseMovie() {
		if(linkedMovieManager != null) {
			linkedMovieManager.pauseLinkedMovies();
		}
		
		if(render != null)
			render.setMovieMode(false);
	}

	/**
	 * {@inheritDoc}
	 */
	public void playMovie() {
		if(linkedMovieManager != null && linkedMovieManager.playLinkedMovies()) {
			return;
		}
		
		if(getMaximumFrameNumber() > 1) {
			render.setMovieMode(true);
			readerSignal.signal();
			if(readerMode != ReaderMode.ONLYFIREONCOMPLETE) {
				renderRequestedSignal.signal(RenderReasons.MOVIE_PLAY);
			}
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean isMoviePlaying() {
		if(render != null) {
			return render.isMovieMode()
			|| (linkedMovieManager != null && linkedMovieManager.isPlaying());
		}
		
		return false;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void abolish() {
		if(linkedMovieManager != null) {
			linkedMovieManager.unlinkMovie(this);
		}
		pauseMovie();
		dateTimeCache.stopParsing();
		super.abolish();
	}
	
	/**
	 * {@inheritDoc}
	 */
	void setSubimageData(ImageData newImageData, SubImage roi, int compositionLayer) {
		if(event == null)
	    	event = new ChangeEvent();
	    
		if(compositionLayer != lastRenderedCompositionLayer) {
        	lastRenderedCompositionLayer = compositionLayer;
        	
        	if(metaData instanceof ObserverMetaData) {
        		((ObserverMetaData) metaData).updateDateTime(dateTimeCache.getDateTime(compositionLayer));
        	}
        	
        	event.addReason(new MetaDataChangedReason(this, metaData));
        }
    	
	    super.setSubimageData(newImageData, roi, 0);
    }
	
	LinkedMovieManager getLinkedMovieManager() {
		return linkedMovieManager;
	}
	
	/** Internal function for setting the current frame number.
	 * 
	 * Before actually setting the new frame number, checks wheater
	 * that is necessary. If the frame number has changed, also
	 * triggers an update of the image.
	 * 
	 * @param frameNumber
	 * @return true, if the frame number has changed
	 */
	protected boolean setCurrentFrameNumber(int frameNumber, ChangeEvent event) {
		if(frameNumber != imageViewParams.compositionLayer) {
			
	        imageViewParams.compositionLayer = frameNumber;
	        
	        while(getMaximumAccessibleFrameNumber() < imageViewParams.compositionLayer) {
	        	try {
					Thread.sleep(200);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
	        }
	        
	        if(this.event == null)
				this.event = event;
	        
			readerSignal.signal();
			if(readerMode != ReaderMode.ONLYFIREONCOMPLETE ) {
				renderRequestedSignal.signal(RenderReasons.MOVIE_PLAY);
			}

	        return true;
    	}
    	return false;
	}
	
	/** Recalculates the image parameters.
     * 
     * <p>This function maps between the set of parameters used within
     * the view chain and the set of parameters used within the
     * jp2-package.
     * 
     * <p>To achieve this, calls {@link #calculateParameter(int, int)}
     * with the currently used number of quality layers and the current
     * frame number.
     * 
     * @return Set of parameters used within the jp2-package
     */
    protected JP2ImageParameter calculateParameter() {
		return calculateParameter(getCurrentNumQualityLayers(), getCurrentFrameNumber());
    }
}
