package org.helioviewer.viewmodel.view.jp2view;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import kdu_jni.Jp2_family_src;
import kdu_jni.Jp2_palette;
import kdu_jni.Jpx_codestream_source;
import kdu_jni.Jpx_source;
import kdu_jni.KduException;
import kdu_jni.Kdu_channel_mapping;
import kdu_jni.Kdu_codestream;
import kdu_jni.Kdu_coords;
import kdu_jni.Kdu_dims;
import kdu_jni.Kdu_region_compositor;
import kdu_jni.Kdu_tile;

import org.helioviewer.base.math.Interval;
import org.helioviewer.base.math.MathUtils;
import org.helioviewer.viewmodel.metadata.ImageSizeMetaData;
import org.helioviewer.viewmodel.metadata.MetaData;
import org.helioviewer.viewmodel.metadata.MetaDataConstructor;
import org.helioviewer.viewmodel.metadata.MultiFrameMetaDataContainer;
import org.helioviewer.viewmodel.view.cache.ImageCacheStatus;
import org.helioviewer.viewmodel.view.jp2view.image.ResolutionSet;
import org.helioviewer.viewmodel.view.jp2view.io.jpip.JPIPResponse;
import org.helioviewer.viewmodel.view.jp2view.io.jpip.JPIPSocket;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_KduException;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_Kdu_cache;
import org.helioviewer.viewmodel.view.jp2view.kakadu.JHV_Kdu_thread_env;
import org.helioviewer.viewmodel.view.jp2view.kakadu.KakaduUtils;
import org.w3c.dom.CharacterData;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This class can open JPEG2000 images, yeah baby! Modified to improve the JPIP
 * communication.
 * 
 * @author caplins
 * @author Benjamin Wamsler
 * @author Juan Pablo
 */
public class JP2Image implements MultiFrameMetaDataContainer {

	/** An array of the file extensions this class currently supports */
	public static final String[] SUPPORTED_EXTENSIONS = { ".JP2", ".JPX" };

	private static int numJP2Images = 0;

	/** This is the URI that uniquely identifies the image. */
	private URI uri;

	/** This is the URI from whch the whole file can be downloaded via http */
	private URI downloadURI;

	/**
	 * This is the object in which all transmitted data is stored. It has the
	 * ability to write itself to disk, and read a relevant cache file from
	 * disk.
	 */
	private JHV_Kdu_cache cache;

	/**
	 * The this extended version of Jp2_family_src can open any file conforming
	 * to the jp2 specifications (.jp2, .jpx, .mj2, etc). The reason for
	 * extending this class is that the Acquire/Release_lock() functions needed
	 * to be implemented.
	 */
	private Jp2_family_src familySrc = new Jp2_family_src();

	/** The Jpx_source object is capable of opening jp2 and jpx sources. */
	private Jpx_source jpxSrc = new Jpx_source();

	/**
	 * The compositor object takes care of all the rendering via its process
	 * function.
	 */
	private Kdu_region_compositor compositor = new Kdu_region_compositor();

	/** The range of valid quality layers for the image. */
	private Interval qLayerRange;

	/** The range of valid composition layer indices for the image. */
	private Interval layerRange;

	/** An object with all the resolution layer information. */
	private ResolutionSet resolutionSet;
	private int resolutionSetCompositionLayer = -1;

	/**
	 * This is a little tricky variable to specify that the file contains
	 * multiple frames
	 */
	private boolean multiFrame = false;

	/** cache path */
	private static File cachePath;

	private JHVJP2View parentView;
	private MetaData metaData;
	private ReentrantLock lock = new ReentrantLock();
	private int referenceCounter = 0;
	private JPIPSocket socket;

	/**
	 * The number of output components (should be the number of 8 bits
	 * channels). Currently only value of 1 and 3 are supported (corresponding
	 * to grayscale and RGB images).
	 */
	private int numComponents;

	/**
	 * The sole constructor for this class. To open an image an ImageDescriptor
	 * must be given and this should be made unique. All initialization for this
	 * object is done in the constructor or in methods called by the
	 * constructor. Either the constructor throws an exception or the image was
	 * opened successfully.
	 * 
	 * @param newUri
	 *            URI representing the location of the image
	 * @throws IOException
	 * @throws JHV_KduException
	 */
	public JP2Image(URI newUri) throws IOException, JHV_KduException {

		numJP2Images++;

		uri = newUri;
		downloadURI = newUri;
		String name = uri.getPath().toUpperCase();
		boolean supported = false;
		for (String ext : SUPPORTED_EXTENSIONS)
			if (name.endsWith(ext))
				supported = true;
		if (!supported)
			throw new JHV_KduException("File extension not supported.");

		multiFrame = name.endsWith(".JPX");

		String scheme = uri.getScheme().toUpperCase();
		if (scheme.equals("JPIP"))
			initRemote();
		else if (scheme.equals("FILE"))
			initLocal();
		else
			throw new JHV_KduException(scheme + " scheme not supported!");

		createKakaduMachinery();

		metaData = MetaDataConstructor.getMetaData(this);
	}

	public JP2Image(URI newUri, URI downloadURI) throws IOException,
			JHV_KduException {
		numJP2Images++;

		uri = newUri;
		this.downloadURI = downloadURI;
		String name = uri.getPath().toUpperCase();
		boolean supported = false;
		for (String ext : SUPPORTED_EXTENSIONS)
			if (name.endsWith(ext))
				supported = true;
		if (!supported)
			throw new JHV_KduException("File extension not supported.");

		multiFrame = name.endsWith(".JPX");

		String scheme = uri.getScheme().toUpperCase();
		if (scheme.equals("JPIP"))
			initRemote();
		else if (scheme.equals("FILE"))
			initLocal();
		else
			throw new JHV_KduException(scheme + " scheme not supported!");

		createKakaduMachinery();

		metaData = MetaDataConstructor.getMetaData(this);
	}

	/**
	 * Initializes the Jp2_family_src for a remote file. (JPIP comms happen
	 * here).
	 * 
	 * @throws JHV_KduException
	 * @throws IOException
	 */
	private void initRemote() throws JHV_KduException {
		// Creates the JPIP-socket necessary for communications
		JPIPResponse res;
		socket = new JPIPSocket();

		try {
			// Connects to the JPIP server, stores the first response in the res
			// variable
			res = (JPIPResponse) socket.connect(uri);

			// Parses the first JPIP response for the JPIP target-ID
			String jpipTargetID;

			if (res.getHeader("JPIP-tid") == null)
				throw new JHV_KduException(
						"The target id was not sent by the server");
			else
				jpipTargetID = res.getHeader("JPIP-tid");

			// Creates the cache object and adds the first response to it.
			cache = new JHV_Kdu_cache(jpipTargetID, cachePath, !multiFrame);
			cache.addJPIPResponseData(res);

			// Download the necessary initial data if there isn't any cache file
			// yet
			if ((cache.getCacheFile() == null)
					|| !cache.getCacheFile().exists())
				KakaduUtils.downloadInitialData(socket, cache);

			familySrc.Open(cache);

		} catch (IOException e) {
			throw new JHV_KduException("Error in the server communication:"
					+ System.getProperty("line.separator") + e.getMessage(), e);
		} catch (KduException e) {
			throw new JHV_KduException("Kakadu engine error opening the image",
					e);
		} finally {
			Timer timer = new Timer("WaitForCloseSocket");
			timer.schedule(new TimerTask() {

				public synchronized void run() {
					if (socket != null) {
						socket.close();
						socket = null;
					}
				}
			}, 5000);
		}
	}

	/**
	 * Initializes the Jp2_family_src for a local file.
	 * 
	 * @throws JHV_KduException
	 * @throws IOException
	 */
	private void initLocal() throws JHV_KduException, IOException {

		// Source is local so it must be a file
		File file = new File(uri);

		// Open the family source
		try {
			familySrc.Open(file.getCanonicalPath(), true);
		} catch (KduException ex) {
			throw new JHV_KduException("Failed to open familySrc", ex);
		}
	}

	/**
	 * Creates the Kakadu objects and sets all the data-members in this object.
	 * 
	 * @throws JHV_KduException
	 */
	private void createKakaduMachinery() throws JHV_KduException {

		// The amount of cache to allocate to each codestream
		final int CODESTREAM_CACHE_THRESHOLD = 1024 * 256;

		try {
			// Open the jpx source from the family source
			jpxSrc.Open(familySrc, false);

			// I don't know if I should be using the codestream in a persistent
			// mode or not...
			compositor.Create(jpxSrc, CODESTREAM_CACHE_THRESHOLD);

			// I create references here so the GC doesn't try to collect the
			// Kdu_dims obj
			Kdu_dims ref1 = new Kdu_dims(), ref2 = new Kdu_dims();

			// A layer must be added to determine the image parameters
			compositor.Add_compositing_layer(0, ref1, ref2);

			{
				// Retrieve the number of composition layers
				{
					int[] tempVar = new int[1];
					jpxSrc.Count_compositing_layers(tempVar);
					layerRange = new Interval(0, tempVar[0] - 1);
				}

				Kdu_codestream stream = compositor.Access_codestream(compositor
						.Get_next_codestream(0, false, true));

				{
					Kdu_tile tile = stream.Open_tile(new Kdu_coords());

					// Retrieve the number of quality layers.
					qLayerRange = new Interval(1, tile.Get_num_layers());

					// Cleanup
					tile.Close();
					tile = null;
				}

				// Retrieve the number of components
				{
					// Since it gets tricky here I am just grabbing a bunch of
					// values
					// and taking the max of them. It is acceptable to think
					// that an
					// image is color when its not monochromatic, but not the
					// other way
					// around... so this is just playing it safe.
					Kdu_channel_mapping cmap = new Kdu_channel_mapping();
					cmap.Configure(stream);

					Jpx_codestream_source jpxStream = jpxSrc
							.Access_codestream(0);
					Jp2_palette palette = jpxStream.Access_palette();

					int maxComponents = MathUtils.max(cmap.Get_num_channels(),
							cmap.Get_num_colour_channels(), palette
									.Get_num_luts(), stream
									.Get_num_components(true), stream
									.Get_num_components(false));

					// numComponents = maxComponents == 1 ? 1 : 3;
					numComponents = maxComponents; // With new file formats we
													// may have 2 components

					// Cleanup
					palette = null;
					jpxStream = null;

					cmap.Clear();
					cmap.Native_destroy();
					cmap = null;
				}

				// Cleanup
				stream = null;
			}

			updateResolutionSet(0);

			// Remove the layer that was added
			compositor.Remove_compositing_layer(-1, true);

		} catch (KduException ex) {
			ex.printStackTrace();
			throw new JHV_KduException("Failed to create Kakadu machinery: "
					+ ex.getMessage(), ex);
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}

	public void setParentView(JHVJP2View _parentView) {
		parentView = _parentView;
	}

	/** Returns true if the image is remote or if image is note open. */
	public boolean isRemote() {
		return cache != null;
	}

	/** Returns the multi frame flag */
	public boolean isMultiFrame() {
		return multiFrame;
	}

	public MetaData getMetaData() {
		return metaData;
	}

	public Jp2_family_src getFamilySrc() {
		return familySrc;
	}

	/**
	 * Method that executes getValueFromXML(_keyword, _box, <currentBoxNumber>).
	 * This will get the xml box from the currently shown frame
	 * 
	 * @param _keyword
	 * @param _box
	 * @throws JHV_KduException
	 */
	public String getValueFromXML(String _keyword, String _box)
			throws JHV_KduException {
		int boxNumber = 1;
		if (parentView != null && parentView.getImageViewParams() != null) {
			boxNumber = parentView.getImageViewParams().compositionLayer + 1;
		}
		return getValueFromXML(_keyword, _box, boxNumber);
	}

	/**
	 * Method that returns value of specified _keyword from specified _box.
	 * 
	 * @param _keyword
	 * @param _box
	 * @param _boxNumber
	 * @throws JHV_KduException
	 */
	public String getValueFromXML(String _keyword, String _box, int _boxNumber)
			throws JHV_KduException {
		String xml = null;

		lock.lock();
		try {
			xml = KakaduUtils.getXml(familySrc, _boxNumber);
		} catch (JHV_KduException e) {
			throw e;
		} finally {
			lock.unlock();
		}

		if (xml == null) {
			throw new JHV_KduException("No XML data present");
		} else if (!xml.contains("</meta>")) {
			throw new JHV_KduException("XML data incomplete");
		}

		xml = xml.trim().replace("$OBS", "");

		InputStream in = null;
		try {
			in = new ByteArrayInputStream(xml.getBytes("UTF-8"));
		} catch (Exception ex) {
			ex.printStackTrace();
		}

		try {
			DocumentBuilder builder = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder();
			Document doc = builder.parse(in);

			NodeList nodes = doc.getElementsByTagName("meta");
			nodes = ((Element) nodes.item(0)).getElementsByTagName(_box);
			NodeList value = ((Element) nodes.item(0))
					.getElementsByTagName(_keyword);
			Element line = (Element) value.item(0);

			if (line == null)
				return null;

			Node child = line.getFirstChild();
			if (child instanceof CharacterData) {
				CharacterData cd = (CharacterData) child;
				return cd.getData();
			}
			return null;
		} catch (Exception e) {
			throw new JHV_KduException("Failed parsing XML data", e);
		}
	}

	public URI getURI() {
		return uri;
	}

	public URI getDownlodURI() {
		return downloadURI;
	}

	/**
	 * Returns the socket, if in remote mode.
	 * 
	 * The socket is returned only one time. After calling this function for the
	 * first time, it will always return null.
	 * 
	 * @return Socket connected to the server
	 */
	public JPIPSocket getSocket() {
		if (socket == null)
			return null;

		JPIPSocket output = socket;
		socket = null;
		return output;
	}

	/** Returns the number of output components. */
	public int getNumComponents() {
		return numComponents;
	}

	/** Returns the an interval of the valid composition layer indices. */
	public Interval getCompositionLayerRange() {
		return layerRange;
	}

	/** Returns the an interval of the valid quality layer values */
	public Interval getQualityLayerRange() {
		return qLayerRange;
	}

	/**
	 * Gets the ResolutionSet object that contains the Resolution level
	 * information.
	 */
	public ResolutionSet getResolutionSet() {
		return resolutionSet;
	}

	/**
	 * {@inheritDoc}
	 */
	public String get(String key) {
		try {
			String value = getValueFromXML(key, "fits");
			return value;
		} catch (JHV_KduException e) {
			if (e.getMessage() != "No XML data present")
				e.printStackTrace();
		}
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	public String get(String key, int frameNumber) throws IOException {
		try {
			String value = getValueFromXML(key, "fits", frameNumber + 1);
			return value;
		} catch (JHV_KduException e) {
			throw new IOException(e.getMessage());
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public double tryGetDouble(String key) {

		String string = get(key);
		if (string != null) {
			return Double.parseDouble(string);
		}
		return 0.0;
	}

	/**
	 * {@inheritDoc}
	 */
	public int tryGetInt(String key) {

		String string = get(key);
		if (string != null) {
			return Integer.parseInt(string);
		}
		return 0;
	}

	/**
	 * {@inheritDoc}
	 */
	public int getPixelHeight() {
		return getResolutionSet().getResolutionLevel(0).getResolutionBounds().height;
	}

	/**
	 * {@inheritDoc}
	 */
	public int getPixelWidth() {

		return getResolutionSet().getResolutionLevel(0).getResolutionBounds().width;
	}

	public static void setCachePath(File newCachePath) {
		cachePath = newCachePath;
	}

	public static File getCachePath() {
		return cachePath;
	}

	/**
	 * Increases the reference counter.
	 * 
	 * This counter is used to count all views, which are using this JP2Image as
	 * their data source. The counter is decreased when calling
	 * {@link #abolish()}.
	 */
	public synchronized void addReference() {
		JHV_Kdu_thread_env.getSingletonInstance().updateNumThreads();
		referenceCounter++;
	}

	/**
	 * Closes the image out. Destroys all objects and performs cleanup
	 * operations. I use the 'abolish' name to distinguish it from what the
	 * Kakadu library uses.
	 */
	public synchronized void abolish() {
		referenceCounter--;

		if (referenceCounter > 0)
			return;

		if (referenceCounter < 0) {
			throw new IllegalStateException(
					"JP2Image abolished more than once: " + uri);
		}

		numJP2Images--;

		try {
			if (compositor != null) {
				compositor.Set_thread_env(null, 0);
				compositor.Remove_compositing_layer(-1, true);
				compositor.Native_destroy();
			}
			if (jpxSrc != null) {
				jpxSrc.Close();
				jpxSrc.Native_destroy();
			}
			if (familySrc != null) {
				familySrc.Close();
				familySrc.Native_destroy();
			}
			if (cache != null) {
				cache.Close();
				cache.Native_destroy();

				JHV_Kdu_cache.updateCacheDirectory(cachePath);
			}
		} catch (KduException ex) {
			ex.printStackTrace();
		} finally {
			compositor = null;
			jpxSrc = null;
			familySrc = null;
			cache = null;
		}
	}

	boolean updateResolutionSet(int compositionLayerCurrentlyInUse) {
		if (resolutionSetCompositionLayer == compositionLayerCurrentlyInUse)
			return false;

		resolutionSetCompositionLayer = compositionLayerCurrentlyInUse;

		try {
			Kdu_codestream stream = compositor.Access_codestream(compositor
					.Get_next_codestream(0, false, true));

			int maxDWT = stream.Get_min_dwt_levels();

			compositor.Set_scale(false, false, false, 1.0f);
			Kdu_dims dims = new Kdu_dims();
			if (!compositor.Get_total_composition_dims(dims))
				return false;

			Kdu_coords size = dims.Access_size();
			if (resolutionSet != null && size.Get_x() == getPixelWidth()
					&& size.Get_y() == getPixelHeight())
				return false;

			if (resolutionSet != null && metaData != null
					&& metaData instanceof ImageSizeMetaData) {
				((ImageSizeMetaData) metaData)
						.updateUnitsPerPixel(((float) getPixelWidth() / size
								.Get_x()));
			}

			resolutionSet = new ResolutionSet(maxDWT + 1);
			resolutionSet.addResolutionLevel(0, KakaduUtils
					.kdu_dimsToRect(dims));

			for (int i = 1; i <= maxDWT; i++) {
				compositor.Set_scale(false, false, false, 1.0f / (1 << i));
				dims = new Kdu_dims();
				if (!compositor.Get_total_composition_dims(dims))
					break;
				resolutionSet.addResolutionLevel(i, KakaduUtils
						.kdu_dimsToRect(dims));
			}

		} catch (KduException e) {
			e.printStackTrace();
		}

		return true;
	}

	JHVJP2View getParentView() {
		return parentView;
	}

	Lock getLock() {
		return lock;
	}

	/** Returns the cache reference */
	JHV_Kdu_cache getCacheRef() {
		return cache;
	}

	/** Sets the ImageCacheStatus */
	void setImageCacheStatus(ImageCacheStatus imageCacheStatus) {
		if (cache != null)
			cache.setImageCacheStatus(imageCacheStatus);
	}

	/** Returns the compositor reference */
	Kdu_region_compositor getCompositorRef() {
		return compositor;
	}

	/**
	 * Returns the number of JP2Image instances currently in use.
	 * 
	 * @return Number of JP2Image instances currently in use
	 */
	static int numJP2ImagesInUse() {
		return numJP2Images;
	}
}
