package org.helioviewer.webbackend;

import java.awt.image.BufferedImage;
import java.io.File;
import java.net.URI;

import javax.imageio.ImageIO;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.helioviewer.base.math.Vector2dInt;
import org.helioviewer.viewmodel.changeevent.ChangeEvent;
import org.helioviewer.viewmodel.region.Region;
import org.helioviewer.viewmodel.region.StaticRegion;
import org.helioviewer.viewmodel.view.jp2view.JHVJP2View;
import org.helioviewer.viewmodel.view.jp2view.JP2Image;
import org.helioviewer.viewmodel.viewport.StaticViewport;
import org.helioviewer.viewmodel.viewport.Viewport;


public class WebBackend {

    private static String inputFileName;
    private static String outputFileName;
    private static Viewport viewport;
    private static Region region;
    private static Vector2dInt tileOffset;
    private static Vector2dInt tileSize;

    public static void main(String[] args) {
        long totalStartTime = System.currentTimeMillis();

        long cmdLineParsingStartTime = System.currentTimeMillis();

        if (!readCmdLine(args)) {
            return;
        }

        if(tileSize == null) {
            tileSize = new Vector2dInt(viewport.getWidth(), viewport.getHeight());
            tileOffset = Vector2dInt.NULL_VECTOR;
        }

        if(tileOffset == null) {
            tileOffset = Vector2dInt.NULL_VECTOR;
        }

        long cmdLineParsingStopTime = System.currentTimeMillis();
        System.out.println("time for command line parsing: " + (cmdLineParsingStopTime - cmdLineParsingStartTime) + " ms");

        long kduSetupStartTime = System.currentTimeMillis();

        try {
            /*KakaduEngine engine = KakaduEngine.getSingletonInstance();
            engine.loadLibs();
            engine.startKduMessageSystem();*/
        } catch (Exception e) {
            System.err.println("Error while initialising Kakadu:");
            e.printStackTrace(System.out);
            return;
        }

        long kduSetupEndTime = System.currentTimeMillis();
        System.out.println("time for kdu initialization: " + (kduSetupEndTime - kduSetupStartTime) + " ms");

        JHVJP2View jp2View = null;
        try {

            long openFileStartTime = System.currentTimeMillis();

            URI uri = null;
            try {
                uri = new File(inputFileName).toURI();
            } catch (Exception e) {
                System.err.println("Error while loading the input file:");
                e.printStackTrace(System.err);
            }
            JP2Image image = new JP2Image(uri);

            long openFileEndTime = System.currentTimeMillis();
            System.out.println("time for opening the image: " + (openFileEndTime - openFileStartTime) + " ms");



            long jp2FilterStartTime = System.currentTimeMillis();


            jp2View = new JHVJP2View(true);

            if(region != null) {
                jp2View.setRegion(region, new ChangeEvent());
            }
            jp2View.setViewport(viewport, new ChangeEvent());
            jp2View.setJP2Image(image);

            long jp2FilterEndTime = System.currentTimeMillis();
            System.out.println("time for creating the jp2View: " + (jp2FilterEndTime - jp2FilterStartTime) + " ms");

        } catch (Exception e) {
            e.printStackTrace(System.err);
        }

        /*long filterStartTime = System.currentTimeMillis();

        StandardFilterView filterView = new StandardFilterView();
        filterView.setView(jp2View);


        long pluginCreationStartTime = System.currentTimeMillis();

        FilterPlugin plugin = new SOHOLUTFilterPlugin();

        long pluginCreationEndTime = System.currentTimeMillis();
        System.out.println("time for creating the plugin: " + (pluginCreationEndTime - pluginCreationStartTime) + " ms");


        filterView.setFilter(plugin.getFilter());

        long filterEndTime = System.currentTimeMillis();
        System.out.println("time for applying the filter: " + ((filterEndTime - filterStartTime) - (pluginCreationEndTime - pluginCreationStartTime)) + " ms");
         */


        long renderStartTime = System.currentTimeMillis();

        JavaGraphicsRenderableView renderableView = new BufferedImageJavaGraphicsRenderableView(jp2View);

        BufferedImage img = new BufferedImage(tileSize.getX(), tileSize.getY(), BufferedImage.TYPE_INT_RGB);
        renderableView.render(img.getGraphics(), tileOffset);

        long renderEndTime = System.currentTimeMillis();
        System.out.println("time for final rendering: " + (renderEndTime - renderStartTime) + " ms");


        long writeFileStartTime = System.currentTimeMillis();

        File file = new File(outputFileName);
        try {
            ImageIO.write(img, "jpeg", file);
        } catch (Exception e) {
            System.err.println("Error while writing the output file:");
            e.printStackTrace(System.err);
        }

        long writeFileEndTime = System.currentTimeMillis();
        System.out.println("time for writing the image: " + (writeFileEndTime - writeFileStartTime) + " ms");


        long totalEndTime = System.currentTimeMillis();
        System.out.println("total time: " + (totalEndTime - totalStartTime) + " ms");

        System.exit(0);
    }

    private static boolean readCmdLine(String[] args) {
        Options options = new Options();

        String inputFileNameOptionString = "inputFileName";
        OptionBuilder.withLongOpt(inputFileNameOptionString);
        OptionBuilder.withDescription("Name of the input image file.");
        OptionBuilder.isRequired();
        OptionBuilder.hasArg();
        OptionBuilder.withArgName("INPUTFILENAME");
        OptionBuilder.withType(String.class);
        Option inputFileNameOption = OptionBuilder.create('i');
        options.addOption(inputFileNameOption);

        String outputFileNameOptionString = "outputFileName";
        OptionBuilder.withLongOpt(outputFileNameOptionString);
        OptionBuilder.withDescription("Name of the output image file.");
        OptionBuilder.isRequired();
        OptionBuilder.hasArg();
        OptionBuilder.withArgName("OUTPUTFILENAME");
        OptionBuilder.withType(String.class);
        Option outputFileNameOption = OptionBuilder.create('o');
        options.addOption(outputFileNameOption);

        String viewportSizeOptionString = "viewport";
        OptionBuilder.withLongOpt(viewportSizeOptionString);
        OptionBuilder.withDescription("Size of the viewport the region is rendered to.");
        OptionBuilder.isRequired();
        OptionBuilder.hasArgs(2);
        OptionBuilder.withArgName("WIDTH,HEIGHT");
        OptionBuilder.withValueSeparator(',');
        Option viewportSizeOption = OptionBuilder.create('v');
        options.addOption(viewportSizeOption);

        String tileSizeOptionString = "tile-size";
        OptionBuilder.withLongOpt(tileSizeOptionString);
        OptionBuilder.withDescription("Size of the tile the viewport is embedded in.");
        OptionBuilder.hasArgs(2);
        OptionBuilder.withArgName("WIDTH,HEIGHT");
        OptionBuilder.withValueSeparator(',');
        Option tileSizeOption = OptionBuilder.create();
        options.addOption(tileSizeOption);

        String tileOffsetOptionString = "tile-offset";
        OptionBuilder.withLongOpt(tileOffsetOptionString);
        OptionBuilder.withDescription("Offset of the viewport in the tile (from upper left corner).");
        OptionBuilder.hasArgs(2);
        OptionBuilder.withArgName("X,Y");
        OptionBuilder.withValueSeparator(',');
        Option tileOffsetOption = OptionBuilder.create();
        options.addOption(tileOffsetOption);

        String regionOptionString = "region";
        OptionBuilder.withLongOpt("region");
        OptionBuilder.withDescription("Size of the region that is rendered to the viewport (in meters).");
        OptionBuilder.hasArgs(4);
        OptionBuilder.withArgName("X,Y,WIDTH,HEIGHT");
        OptionBuilder.withValueSeparator(',');
        Option regionOption = OptionBuilder.create('r');
        options.addOption(regionOption);

        CommandLine cmdLine = null;
        try {
            CommandLineParser parser = new PosixParser();
            cmdLine = parser.parse(options, args);
        } catch (Exception e) {
            e.printStackTrace(System.out);
            HelpFormatter formatter = new HelpFormatter();
            formatter.printHelp("java -jar webbackend.jar", options);
            return false;
        }

        inputFileName = cmdLine.getOptionValue(inputFileNameOptionString);
        outputFileName = cmdLine.getOptionValue(outputFileNameOptionString);
        int vx = Integer.parseInt(cmdLine.getOptionValues(viewportSizeOptionString)[0]);
        int vy = Integer.parseInt(cmdLine.getOptionValues(viewportSizeOptionString)[1]);
        viewport = StaticViewport.createAdaptedViewport(vx, vy);

        if(cmdLine.hasOption(tileSizeOptionString)) {
            int w = Integer.parseInt(cmdLine.getOptionValues(tileSizeOptionString)[0]);
            int h = Integer.parseInt(cmdLine.getOptionValues(tileSizeOptionString)[1]);
            tileSize = new Vector2dInt(w, h);
        } else {
            tileSize = null;
        }

        if(cmdLine.hasOption(tileOffsetOptionString)) {
            int x = Integer.parseInt(cmdLine.getOptionValues(tileOffsetOptionString)[0]);
            int y = Integer.parseInt(cmdLine.getOptionValues(tileOffsetOptionString)[1]);
            tileOffset = new Vector2dInt(x, y);
        } else {
            tileOffset = null;
        }

        if(cmdLine.hasOption(regionOptionString)) {
            double x = Double.parseDouble(cmdLine.getOptionValues(regionOptionString)[0]);
            double y = Double.parseDouble(cmdLine.getOptionValues(regionOptionString)[1]);
            double w = Double.parseDouble(cmdLine.getOptionValues(regionOptionString)[2]);
            double h = Double.parseDouble(cmdLine.getOptionValues(regionOptionString)[3]);
            region = StaticRegion.createAdaptedRegion(x, y, w, h);
        } else {
            region = null;
        }

        return true;
    }

}
